Package: phantasus
Title: Visual and interactive gene expression analysis
Version: 1.30.0
Authors@R: c(
    person("Maxim", "Kleverov", role = "aut"),        
    person("Daria", "Zenkova", role = "aut"),
    person("Vladislav", "Kamenev", role = "aut"),        
    person("Margarita", "Sablina", role = "ctb"),        
    person("Maxim", "Artyomov", role = "aut"),
    person("Alexey", "Sergushichev", email = "alsergbox@gmail.com", role = c("aut", "cre")))
Description: 
    Phantasus is a web-application for visual and interactive gene expression analysis.
    Phantasus is based on Morpheus – a web-based software for heatmap visualisation 
    and analysis, which was integrated with an R environment via OpenCPU API.
    Aside from basic visualization and filtering methods, R-based methods such 
    as k-means clustering, principal component analysis or differential expression analysis
    with limma package are supported.
URL: https://alserglab.wustl.edu/phantasus
BugReports: https://github.com/ctlab/phantasus/issues
Depends: R (>= 4.3)
biocViews: GeneExpression, GUI, Visualization, DataRepresentation,
        Transcriptomics, RNASeq, Microarray, Normalization, Clustering,
        DifferentialExpression, PrincipalComponent, ImmunoOncology
Imports: ggplot2, protolite, Biobase, GEOquery, Rook, htmltools,
        httpuv, jsonlite, limma, edgeR, opencpu, assertthat, methods,
        httr, rhdf5, utils, parallel, stringr, fgsea (>= 1.9.4),
        svglite, gtable, stats, Matrix, pheatmap, scales, ccaPP, grid,
        grDevices, AnnotationDbi, DESeq2, data.table, curl, apeglm,
        tidyr, config (>= 0.3.2), rhdf5client (>= 1.25.1), yaml, fs,
        phantasusLite, XML
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.2
Suggests: testthat, BiocStyle, knitr, rmarkdown, org.Hs.eg.db,
        org.Mm.eg.db
VignetteBuilder: knitr
NeedsCompilation: no
Config/build/copy-method: link
git_url: https://git.bioconductor.org/packages/phantasus
git_branch: RELEASE_3_22
git_last_commit: 7182df1
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-31
Packaged: 2025-10-31 23:27:47 UTC; biocbuild
Author: Maxim Kleverov [aut],
  Daria Zenkova [aut],
  Vladislav Kamenev [aut],
  Margarita Sablina [ctb],
  Maxim Artyomov [aut],
  Alexey Sergushichev [aut, cre]
Maintainer: Alexey Sergushichev <alsergbox@gmail.com>

# Generated by roxygen2: do not edit by hand

export(generatePreloadedSession)
export(getES)
export(getGDS)
export(getGSE)
export(getPhantasusConf)
export(read.gct)
export(reparseCachedESs)
export(servePhantasus)
export(setupPhantasus)
export(write.gct)
import(Biobase)
import(GEOquery)
import(Rook)
import(apeglm)
import(ccaPP)
import(data.table)
import(edgeR)
import(ggplot2)
import(grid)
import(gtable)
import(htmltools)
import(httpuv)
import(jsonlite)
import(limma)
import(opencpu)
import(rhdf5)
import(stats)
import(svglite)
importFrom(AnnotationDbi,columns)
importFrom(AnnotationDbi,keys)
importFrom(AnnotationDbi,loadDb)
importFrom(AnnotationDbi,mapIds)
importFrom(AnnotationDbi,species)
importFrom(fgsea,fgseaMultilevel)
importFrom(fgsea,plotEnrichment)
importFrom(ggplot2,ggtitle)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,dev.off)
importFrom(grDevices,svg)
importFrom(methods,new)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(tidyr,spread)
importFrom(utils,download.file)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,object.size)
importFrom(utils,read.table)
importFrom(utils,tail)
importFrom(utils,write.table)

Phantasus is a web tool designed for visual and
interactive gene expression analysis.
In particular, it was designed to allow to go from a typical dataset to
differential expression and downstream analysis in an easy and streamlined
manner. For that aim, Phantasus integrates an intuitive heatmap interface with
gene expression analysis tools from Bioconductor. 

Main features:
* Loading public datasets from Gene Expression Omnibus with both microarrays and RNA-seq datasets (via ARCHS4) being supported.
* Differential gene expression using `limma` or `DESeq2`.
* Publication ready plots with export to SVG: PCA plot, row profiles, box plots.
* Clustering: k-means and hierarchical.
* Gene set enrichment analysis via `fgsea` package.
* Sharing session links.

<img src="https://ctlab.github.io/phantasus-doc/images/screenshot.png" width="600px" />

## Quick start

Phantasus can be accessed online via its official mirror: <https://alserglab.wustl.edu/phantasus>.

Alternatively, phantasus can be set up locally as an R package. The latest version of Phantasus 
can be installed from GitHub using `devtools` package.

```r
devtools::install_github("ctlab/phantasus")
```

There are several system packages that have to be installed on the system. The
names of these packages will be displayed during installation. On Ubuntu one can
install them beforehand and all together using the command:

```bash
sudo apt-get install libapparmor-dev libfontconfig1-dev libcairo2-dev libcurl4-openssl-dev pandoc libtiff5-dev libfribidi-dev libharfbuzz-dev libssl-dev libxml2-dev libprotobuf-dev protobuf-compiler
```

Further, the latest version of `phantasus` depends on `rhdf5client (>= 1.25.1)` from Bioconductor 3.19, which on older systems can be more convenient to install from GitHub:

```r
devtools::install_github("vjcitn/rhdf5client")
```

Before the first run of Phantasus you need to do an initial setup. To perform interactive setup run the following commands. We recommend to use the first option for each of the questions:

```r
library(phantasus)
setupPhantasus()
```

After the setup is finished, the following command will start the application at  <http://0.0.0.0:8000>
and open it in the default browser:

```r
servePhantasus()
```

Please refer to the [documentation](https://ctlab.github.io/phantasus-doc/installation.html#using-docker) for comprehensive installation and configuration instructions. 

## Links:
* Official mirror: <https://alserglab.wustl.edu/phantasus>.
* Documentation: <https://ctlab.github.io/phantasus-doc>.
* Bioconductor package: <https://bioconductor.org/packages/phantasus>.
* Docker image: <https://hub.docker.com/r/alserglab/phantasus>.

## Citation:
* Kleverov M, Zenkova D, Kamenev V, Sablina R, Artyomov M, Sergushichev A (2022). Phantasus: web-application for visual and interactive gene expression analysis. doi:10.1101/2022.12.10.519861, https://www.biorxiv.org/content/10.1101/2022.12.10.519861. 

library(GEOquery)
library(rhdf5)

es <- getGEO("GSE99709")[[1]]

srcfile <- "./cache/mouse_matrix.h5"

samples <- h5read(srcfile, "meta/Sample_geo_accession")
genes <- as.character(h5read(srcfile, "meta/genes"))

keepGSMs <- head(intersect(es$geo_accession, samples), 3)

es <- es[, es$geo_accession %in% keepGSMs]

sampleIndexes <- match(es$geo_accession,
                       samples)

expression <- h5read(srcfile,
                     "data/expression",
                     index=list(seq_along(genes),
                                stats::na.omit(sampleIndexes)))
rownames(expression) <- genes
colnames(expression) <- colnames(es)[!is.na(sampleIndexes)]
H5close()


destfile <- "./inst/testdata/mouse_matrix.h5"

h5createFile(destfile)

h5createGroup(destfile, "data")

# h5createDataset(file=destfile,
#                 dataset="data/expression",
#                 dims=dim(expression),
#                 storage.mode="integer")

h5write(expression,
        file=destfile,
        name="data/expression",
        index=list(
            NULL,
            NULL))

h5createGroup(destfile, "meta")

# h5createDataset(file=destfile,
#                 dataset="meta/genes",
#                 dims=length(genes),
#                 storage.mode="character",
#                 size = max(nchar(genes)*2))

h5write(genes,
        file=destfile,
        name="meta/genes")

h5write(colnames(expression),
        file=destfile,
        name="meta/Sample_geo_accession")

H5close()

# serve-managed.R
#
# Managed worker script for Docker deployment of phantasus.
#
# This script is Docker/Linux operational infrastructure; it does not
# belong in the R package API.  It lives in inst/ so it is shipped with
# the package and can be located via system.file().
#
# Architecture:
#   Parent process  — runs setup once, forks N workers, monitors them.
#   Worker process  — serves HTTP via httpuv; self-exits when R's full
#                     gc fires or max age is exceeded.
#
# The parent builds the app object once before any fork.
# All workers share it copy-on-write.
#
# Configuration (environment variables):
#   PHANTASUS_MAX_WORKER_AGE_S   default 21600 (6 hours)

suppressPackageStartupMessages({
    library(phantasus)
    library(httpuv)
    library(opencpu)
    library(futile.logger)
})

# ── configuration ─────────────────────────────────────────────────────────────

`%||%` <- function(a, b) if (length(a) > 0L && !is.null(a)) a else b

ports <- unlist(getPhantasusConf("internal_ports"))

MAX_AGE <- as.numeric(
    Sys.getenv("PHANTASUS_MAX_WORKER_AGE_S", unset = "21600")
)

flog.info("ports=%s  max_age=%.0f s",
          paste(range(ports), collapse = "-"), MAX_AGE)

# ── pre-fork setup ────────────────────────────────────────────────────────────
#
# Runs exactly once in the parent.  The resulting `app` object and all
# its closures are shared read-only across workers via CoW.

host <- getPhantasusConf("host")

app <- phantasus:::buildPhantasusApp(run_worker = NULL)
gc()   # compact heap before forking so CoW pages start clean

# ── shared helpers ────────────────────────────────────────────────────────────

get_pss_mb <- function(pid) {
    path  <- sprintf("/proc/%d/smaps_rollup", as.integer(pid))
    lines <- tryCatch(readLines(path), error = function(e) character(0))
    hit   <- grep("^Pss:", lines, value = TRUE)
    if (length(hit) == 0L) {
        return(NA_real_)
    }
    as.numeric(sub("^Pss:\\s+(\\d+).*", "\\1", hit)) / 1024
}

is_alive <- function(pid) {
    # pskill(pid, 0) returns TRUE even for zombie processes.
    # Read /proc/<pid>/status and check the State field instead.
    lines <- tryCatch(
        readLines(sprintf("/proc/%d/status", as.integer(pid))),
        error = function(e) character(0)
    )
    if (length(lines) == 0L) {
        return(FALSE)
    }
    state <- grep("^State:", lines, value = TRUE)
    length(state) > 0L && !grepl("Z", state[[1L]])
}

# ── worker body ───────────────────────────────────────────────────────────────

run_managed_worker <- function(port) {
    worker_born <- proc.time()[["elapsed"]]
    max_age     <- MAX_AGE * runif(1L, min = 1, max = 2)

    # Sentinel: fires only on a full (level 2) gc.
    # Two minor gc passes age it into generation 2 before we drop the
    # reference, so the finalizer only triggers on a full collection.
    gc_fired <- FALSE
    sentinel <- new.env(parent = emptyenv())
    reg.finalizer(sentinel, function(e) { gc_fired <<- TRUE }, onexit = FALSE)
    gc(full = FALSE)
    gc(full = FALSE)
    rm(sentinel)

    server <- tryCatch(
        startServer(host, port, app),
        error = function(e) {
            flog.error("worker port=%d pid=%d FAILED to start: %s",
                       port, Sys.getpid(), conditionMessage(e))
            NULL
        }
    )

    if (is.null(server)) {
        return(invisible(NULL))
    }

    flog.info("worker port=%d pid=%d started", port, Sys.getpid())

    # Service loop: check gc sentinel and age every 5000 ms.
    reason <- "unknown"
    repeat {
        httpuv::service(5000L)

        age_s <- proc.time()[["elapsed"]] - worker_born
        if (age_s > max_age) {
            reason <- sprintf("max age (%.0f s)", age_s)
            break
        }

        if (gc_fired) {
            reason <- sprintf("gc fired (age=%.0f s)", age_s)
            break
        }
    }

    flog.info("worker port=%d pid=%d exiting: %s", port, Sys.getpid(), reason)
    stopServer(server)
    # Return rather than quit() — quit() triggers parallel's CleanupParallel
    # hook in the parent via R's atexit mechanism, setting
    # R_interrupts_pending and silently breaking the parent's monitor loop.
    return(invisible(NULL))
}

# ── manager side ──────────────────────────────────────────────────────────────

jobs      <- list()   # mcparallel job objects, keyed by as.character(pid)
job_ports <- list()   # port number, keyed by as.character(pid)

start_worker <- function(port) {
    job <- parallel::mcparallel(run_managed_worker(port), detach = FALSE)
    pid <- job$pid
    jobs[[as.character(pid)]]      <<- job
    job_ports[[as.character(pid)]] <<- port
}

# ── entry point ───────────────────────────────────────────────────────────────

flog.info("starting %d workers on ports %d-%d",
          length(ports), min(ports), max(ports))

for (p in ports) {
    start_worker(p)
}

# shutdown: SIGTERM all live workers and exit the manager.
shutdown <- function() {
    flog.fatal("manager shutting down — terminating workers")
    for (pid_char in names(jobs)) {
        if (is_alive(as.integer(pid_char))) {
            tryCatch(
                tools::pskill(as.integer(pid_char), tools::SIGTERM),
                error = function(e) NULL
            )
        }
    }
    quit(save = "no", status = 0L)
}

# Manager gc sentinel: logs when a full (level 2) gc fires in the manager.
# Reset after each detection so we catch subsequent collections too.
mgr_gc_fired <- FALSE
mgr_sentinel <- new.env(parent = emptyenv())
reg.finalizer(mgr_sentinel,
              function(e) { mgr_gc_fired <<- TRUE },
              onexit = FALSE)
gc(full = FALSE)
gc(full = FALSE)
rm(mgr_sentinel)

# Respawn loop.
#
# mccollect(wait=FALSE, timeout=300) blocks until any single worker
# finishes (or 300 s elapse), then returns whatever has completed.
#
# Distinguish SIGINT from SIGCHLD: if interrupted and no job finished,
# check whether any known pid is actually dead.  If none → genuine SIGINT
# → shut down.

repeat {
    interrupted <- FALSE
    finished <- tryCatch(
        parallel::mccollect(jobs, wait = FALSE, timeout = 300L),
        interrupt = function(e) { interrupted <<- TRUE; NULL }
    )
    if (is.null(finished)) {
        finished <- list()
    }

    mgr_pss   <- get_pss_mb(Sys.getpid()) %||% 0
    wkr_pss   <- vapply(names(job_ports), function(pid_char) {
        get_pss_mb(as.integer(pid_char)) %||% 0
    }, numeric(1L))
    total_pss <- sum(c(mgr_pss, wkr_pss), na.rm = TRUE)
    flog.info("manager loop  total_pss=%.1f MB", total_pss)

    if (mgr_gc_fired) {
        flog.warn("full gc fired in manager")
        mgr_gc_fired <- FALSE
        mgr_sentinel <- new.env(parent = emptyenv())
        reg.finalizer(mgr_sentinel,
                      function(e) { mgr_gc_fired <<- TRUE },
                      onexit = FALSE)
        gc(full = FALSE)
        gc(full = FALSE)
        rm(mgr_sentinel)
    }

    if (interrupted && length(finished) == 0L) {
        # Distinguish SIGINT (no worker actually dead) from a SIGCHLD that
        # mccollect hasn't collected yet.
        any_dead <- any(vapply(names(jobs), function(pid_char) {
            !is_alive(as.integer(pid_char))
        }, logical(1L)))

        if (!any_dead) {
            shutdown()
        }

        # A worker is dead but mccollect hasn't collected it yet.
        # Give the zombie a moment to be reaped, then retry.
        finished <- parallel::mccollect(jobs, wait = FALSE, timeout = 0.1)
        if (is.null(finished)) {
            finished <- list()
        }
    }

    for (pid_char in names(finished)) {
        port <- job_ports[[pid_char]]
        jobs[[pid_char]]      <- NULL
        job_ports[[pid_char]] <- NULL
        flog.info("manager port=%d pid=%s exited — replacing", port, pid_char)
        start_worker(port)
    }
}

#' K-means clusterisation.
#'
#' \code{performKmeans} returns a vector of corresponding clusters for
#'     each gene from a given ExpressionSet.
#'
#' @param es ExpressionSet object.
#'
#' @param k Expected number of clusters.
#'
#' @param replacena Method for replacing NA values
#'     in series matrix (mean by default)
#'
#' @return Vector of corresponding clusters, serialized to JSON.
#'
#' @import Biobase
#'
#' @examples
#' \dontrun{
#' data(es)
#' performKmeans(es, k = 2)
#' }
#' @keywords internal
performKmeans <- function(es, k,replacena = "mean") {
    assertthat::assert_that(k > 0)

    scaledExprs <- unname(exprs(es))

    naInd <- which(is.na(scaledExprs), arr.ind = TRUE)
    if (nrow(naInd) > 0) {
        replaceValues <- apply(scaledExprs, 1, replacena, na.rm=TRUE)
        scaledExprs[naInd] <- replaceValues[naInd[,1]]
    }

    scaledExprs <- t(scale(t(scaledExprs)))
    rowsToCluster <- which(!apply(is.na(scaledExprs), 1, any))

    km <- stats::kmeans(scaledExprs[rowsToCluster, ], k, iter.max = 100L)
    res <- character(nrow(scaledExprs))
    res[rowsToCluster] <- as.character(km$cluster)
    fData(es)$clusters = res
    assign("es", es, envir = parent.frame())

    return(jsonlite::toJSON(res))
}

context("Kmeans")
library(jsonlite)
library(phantasusLite)

test_that("kmeans finishes with result", {
  load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))
  expect_is(performKmeans(es, k = 10), "json")
})

test_that("kmeans works with datasets with lots of NAs", {
  es <- readGct(system.file("testdata/centers.gct", package="phantasus"))
  clusters = fromJSON(performKmeans(es, k=2))
  expect_length(clusters, nrow(es))
  expect_equal(clusters[4], "")
})

# Invalid test. Interface changed such way so no indices passed to arguments
#test_that("kmeans returns only vector for a subset", {
#    es <- read.gct(system.file("testdata/centers.gct", package="phantasus"))
#    rows <- c(0, 2, 3, 5)
#    expect_length(fromJSON(performKmeans(es, rows=rows, k = 2)), length(rows))
#})

#' Collapse dataset
#'
#' \code{collapseDataset} performs a collapse action on expression set
#'
#' @param es Expression set
#' @param isRows Work with rows. False if columns (default True - row mode)
#' @param selectOne select best match or merge duplicates
#' @param fn select/merge function
#' @param fields fields to unique on
#' @param removeEmpty remove unannotated genes
#'
#' @return Nothing. Collapsed dataset will be assigned to es in environment
#'
#' @import ccaPP
#'
#' @examples
#' \dontrun{
#' es <- getGSE('GSE53986')[[1]]
#' collapseDataset(es, isRows = TRUE, selectOne = TRUE,
#' fn = mean, fields = c('Gene ID', 'Gene symbol'))
#' }
#' @keywords internal
collapseDataset <- function (es, isRows = TRUE, selectOne = FALSE, fn, fields, removeEmpty = TRUE) {
    es <- collapseDatasetImpl(es, isRows, selectOne, fn, fields, removeEmpty)

    assign("es", es, envir = parent.frame())
}

collapseDatasetImpl <- function (es, isRows = TRUE, selectOne = FALSE, fn, fields, removeEmpty) {
    if (removeEmpty) {
        es <- stripEmpty(es, isRows, fields)
    }

    expr <- exprs(es)
    fact <- collectFactor(es, isRows, fields)
    f2 <- factor(fact, levels=unique(fact))

    if (selectOne) { #select fittest
        ranks <- apply(expr, 1, fn)
        factorFrame <- data.frame(f=f2, i=seq_along(ranks), r=ranks)
        factorFrame <- factorFrame[order(factorFrame$f, -factorFrame$r), ]
        keep <- factorFrame[!duplicated(factorFrame$f) & !is.na(factorFrame$f), ]$i
        res <- es[keep, ]
    } else { #merge duplicates
        factorFrame <- data.frame(f=f2, i=seq_len(length(f2)))
        keep <- factorFrame[!duplicated(factorFrame$f) & !is.na(factorFrame$f), ]$i
        keep <- sort(keep)

        if (isRows) {
            res <- es[keep, ]
            oldAnnotation <- fData(res)
        } else {
            expr <- t(expr)
            res <- es[, keep]
            oldAnnotation <- pData(res)
        }

        splitted <- split(seq_len(length(f2)), fact)
        zz <- lapply(seq_len(ncol(expr)), function(i) split(expr[, i], f2))
        zz <- unlist(zz, recursive = FALSE)
        collapsedExprs <- lapply(seq_len(ncol(expr)),
                            function (i) sapply(split(expr[, i], f2), fn))
        collapsedExprs <- do.call(cbind, collapsedExprs)

        if (!isRows) {
            collapsedExprs <- t(collapsedExprs)
        }

        rownames(collapsedExprs) <- rownames(res)
        colnames(collapsedExprs) <- colnames(res)
        exprs(res) <- collapsedExprs
        fields <- colnames(oldAnnotation)[which(colnames(oldAnnotation) %in% fields)]
        newAnnotaion <- oldAnnotation[, which(colnames(oldAnnotation) %in% fields), drop=FALSE]
        rownames(newAnnotaion) <- rownames(oldAnnotation)
        colnames(newAnnotaion) <- fields
        if (isRows) {
            fData(res) <- newAnnotaion
        } else {
            pData(res) <- newAnnotaion
        }
    }
    res
}

collectFactor <- function (es, isRows, fields) {
    if (!length(fields)) {
        stop('Empty fields given')
    }

    if (isRows) {
        target <- fData(es)
    } else {
        target <- pData(es)
    }


    f <- apply(target[fields], 1, paste, collapse='///')
    return(f)
}

stripEmpty <- function (es, isRows, fields) {
    target <- fData(es)
    if (!isRows) {
        target <- pData(es)
    }

    toStrip <- !apply(target[fields], 1, function (row) {
        any(is.na(row)) || any(nchar(row) == 0)
    })

    if (isRows) {
        es <- es[which(toStrip), ]
    } else {
        es <- es[, which(toStrip)]
    }

    es
}

context('Collapse dataset')
library(phantasusLite)
Sys.setenv(R_USER_CONFIG_DIR = system.file("/testdata/config", package = "phantasus"))
test_that("Collapse dataset selectOne works correctly", {
    es <- readGct(system.file("testdata/collapse_dataset_one.gct", package="phantasus"))
    newEs <- collapseDatasetImpl(es, selectOne = TRUE, fn = median, fields = c('Gene ID'), removeEmpty = FALSE)
    expect_equal(dim(exprs(newEs)), c(3,7))
    expect_equal(unname(dim(newEs)), dim(exprs(newEs)))
    expect_equal(fData(newEs)[1,][['Gene symbol']], 'Copg1')
    expect_true(all(isUnique(fData(newEs)[['Gene ID']])))
    expect_true(all(exprs(es)[5,] == exprs(newEs)[3,]))
    expect_true(all(exprs(es)[4,] == exprs(newEs)[2,]))
    expect_true(all(exprs(es)[1,] == exprs(newEs)[1,]))
})


test_that("Collapse dataset selectOne all unique", {
    es <- readGct(system.file("testdata/collapse_dataset_one.gct", package="phantasus"))
    newEs <- collapseDatasetImpl(es, selectOne = TRUE, fn = median, fields = c('id','Gene ID', 'Gene symbol'), removeEmpty = FALSE)
    expect_equal(dim(exprs(newEs)), c(5,7))
    expect_equal(unname(dim(newEs)), dim(exprs(newEs)))
    expect_true(all(exprs(es) == exprs(newEs)))
})

test_that("Collapse dataset by rows", {
    es <- readGct(system.file("testdata/collapse_dataset_one.gct", package="phantasus"))
    newEs <- collapseDatasetImpl(es, fn = median, fields = c('Gene ID'), removeEmpty = FALSE)
    expect_equal(ncol(fData(newEs)), 1)
    expect_equal(nrow(fData(newEs)), 3)
    expect_equal(dim(exprs(newEs)), c(3,7))
    expect_true(all(isUnique(fData(newEs)[['Gene ID']])))
    expect_true(all(exprs(es)[5,] == exprs(newEs)[3,]))
    expect_true(all(exprs(es)[4,] == exprs(newEs)[2,]))
    expect_equal(apply(exprs(es)[c(1,2,3),], 2, median), exprs(newEs)[1,])


    newEsMultiple <- collapseDatasetImpl(es, fn = mean, fields = c('Gene ID', 'Gene symbol'), removeEmpty = FALSE)
    expect_equal(ncol(fData(newEsMultiple)), 2)
    expect_equal(apply(exprs(es)[c(1,2,3),], 2, mean), exprs(newEsMultiple)[1,])
    expect_true(all(isUnique(paste(fData(newEsMultiple)[['Gene ID']], fData(newEsMultiple)[['Gene symbol']], sep="//r"))))
})

test_that("Collapse dataset colnames doesn't reoder", {
    es <- readGct(system.file("testdata/collapse_dataset_one.gct", package="phantasus"))
    newEs <- collapseDatasetImpl(es, fn = median, fields = c('Gene ID', 'id', 'Gene symbol'), removeEmpty = FALSE)
    expect_equal(colnames(fData(newEs)), colnames(fData(es)))

    newEs <- collapseDatasetImpl(es, fn = median, fields = c('Gene ID', 'Gene symbol'), removeEmpty = FALSE)
    expect_lt(match('Gene symbol', colnames(fData(newEs))), match('Gene ID', colnames(fData(newEs))))
})


test_that("Collapse dataset by columns", {
    es <- readGct(system.file("testdata/collapse_dataset_one.gct", package="phantasus"))
    newEs <- collapseDatasetImpl(es, isRows = FALSE, fn = median, fields = c('condition'), removeEmpty = FALSE)
    expect_equal(ncol(pData(newEs)), 1)
    expect_equal(nrow(pData(newEs)), 2)
    expect_equal(dim(exprs(newEs)), c(5,2))
    expect_true(all(isUnique(pData(newEs)[['condition']])))
    expect_equal(apply(exprs(es)[,c(1,2,3,4)], 1, median), exprs(newEs)[,1])
    expect_equal(apply(exprs(es)[,c(5,6,7)], 1, median), exprs(newEs)[,2])
})

test_that("Collapse strip unannotated rows", {
    old_conf <- Sys.getenv("R_CONFIG_ACTIVE")
    Sys.setenv(R_CONFIG_ACTIVE = "test_real_geo")
    ess <- getGSE('GSE53986')[[1]]
    unstripped <- collapseDatasetImpl(ess, selectOne = TRUE, fn = median, fields = c('Gene ID'), removeEmpty = FALSE)
    stripped <- collapseDatasetImpl(ess, selectOne = TRUE, fn = median, fields = c('Gene ID'), removeEmpty = TRUE)
    expect_equal(nrow(stripped) + 1, nrow(unstripped))
    Sys.setenv(R_CONFIG_ACTIVE = old_conf)
})

#' Constructs data frame with gene annotations and submits it into Shiny GAM web-server
#' @param es Expression set object
#' @return URL for Shiny GAM
#' @importFrom utils write.table
#' @keywords internal
shinyGAMAnalysis <- function(es) {
    de <- fData(es)

    deFile <- tempfile()
    write.table(de, deFile, sep="\t", row.names=FALSE, col.names=TRUE)

    # :ToDo: deal with bad certificates and remove ssl_verifypeer option
    r <- httr::POST(url="https://artyomovlab.wustl.edu/upload",
               body=readBin(deFile, what="raw", n=20e6),
               httr::config(ssl_verifypeer = FALSE))
    httr::stop_for_status(r)
    key <- httr::content(r, as="text", encoding="UTF-8")

    shinyGAMUrl <- sprintf("https://artyomovlab.wustl.edu/shiny/gatom/?geneDE_key=%s", key)
    return(jsonlite::toJSON(shinyGAMUrl))
}

#' Serve phantasus.
#'
#' \code{servePhantasus} starts http server handling phantasus static files
#'     and opencpu server.
#'
#' @param host Host to listen.
#'
#' @param port Port to listen.
#'
#' @param staticRoot Path to static files with phantasus.js
#'     (on local file system).
#'
#'
#' @param preloadedDir Full path to directory with preloaded files.
#'
#' @param openInBrowser Boolean value which states if application will be
#'     automatically loaded in default browser.
#'
#' @param quiet Boolean value which states whether the connection log should
#'     be hidden (default: TRUE)
#'
#' @param background Boolean value which states whether the server should be started in background (default: FALSE)
#'
#' @return A handle to the server as returned by `httpuv::startServer`
#'
#' @import opencpu
#' @import httpuv
#' @import Rook
#' @importFrom utils getFromNamespace
#' @importFrom parallel makeCluster stopCluster
#' @export
#'
#' @examples
#' \dontrun{
#' s <- servePhantasus(background=FALSE)
#' s$stop()
#' }
#'
#' httpuv::stopAllServers() # can be used if handle is lost
servePhantasus <- function(host = getPhantasusConf("host"),
                           port = getPhantasusConf("port"),
                           staticRoot = getPhantasusConf("static_root"),
                           preloadedDir = getPhantasusConf("preloaded_dir"),
                           openInBrowser = TRUE,
                           quiet=TRUE,
                           background=FALSE) {
    if (nchar(staticRoot) == 0){
        staticRoot = system.file("www/phantasus.js", package = "phantasus")
    }
    cacheDir <- normalizePath(getPhantasusConf("cache_root"))
    preloadedDir <- if (is.null(preloadedDir)){
        NULL
    } else{
        normalizePath(preloadedDir)
    }

    if (!dir.exists(cacheDir) | !areCacheFoldersValid(getPhantasusConf("cache_folders")) ){
        stopPhantasus()
    }
    options(phantasusCacheDir = cacheDir,
            phantasusPreloadedDir = preloadedDir)

    selfCheck()
    if (!opencpu:::win_or_mac()) {
        if (! "unix" %in% utils::installed.packages()) {
            if (interactive() && menu(c("Yes", "No"),
                     title= paste("Couldn't find the required `unix` package, do you want to install it?")) == "1") {
                install.packages("unix")
            } else {
                stop("Phantasus can't work without `unix` package, please install it")
            }
        }
        run_worker <- NULL
    } else {
        #### this fragment is adopted from opencpu::ocpu_start_server function
        #### https://github.com/opencpu/opencpu/blob/master/R/start.R
        #### :ToDo: remove code duplication

        # set root home for workers
        Sys.setenv("OCPU_MASTER_HOME" = opencpu:::tmp_root())
        on.exit(Sys.unsetenv("OCPU_MASTER_HOME"))

        # import
        sendCall <- getFromNamespace('sendCall', 'parallel')
        recvResult <- getFromNamespace('recvResult', 'parallel')
        preload <- "opencpu"

        # worker pool
        pool <- list()

        # add new workers if needed
        add_workers <- function(n = 1){
            if(length(pool) < 2){
                cl <- parallel::makeCluster(n)
                lapply(cl, sendCall, fun = function(){
                    lapply(preload, getNamespace)
                    options(phantasusCacheDir = cacheDir,
                            phantasusPreloadedDir = preloadedDir)
                    Sys.getpid()
                }, args = list())
                pool <<- c(pool, cl)
            }
        }

        # get a worker
        get_worker <- function(){
            if(!length(pool))
                add_workers(1)
            node <- pool[[1]]
            pool <<- pool[-1]
            pid <- recvResult(node)
            if(inherits(pid, "try-error"))
                warning("Worker preload error: ", pid, call. = FALSE, immediate. = TRUE)
            node$pid <- pid
            structure(list(node), class = c("SOCKcluster", "cluster"))
        }

        # main interface
        run_worker <- function(fun, ..., timeout = NULL){
            res <- tryCatch({
                if(length(timeout)){
                    setTimeLimit(elapsed = timeout)
                    on.exit(setTimeLimit(cpu = Inf, elapsed = Inf), add = TRUE)
                }
                cl <- get_worker()
                on.exit(kill_workers(cl), add = TRUE)
                node <- cl[[1]]
                sendCall(node, fun, list(...))
                recvResult(node)
            }, error = function(e){
                if(grepl("elapsed time limit", e$message)){
                    tools::pskill(node$pid)
                    stop(sprintf("Timeout reached: %ds (see rlimit.post in user.conf)", timeout))
                }
                stop(e)
            })
            if(inherits(res, "try-error"))
                stop(res)
            res
        }

        kill_workers <- function(cl){
            parallel::stopCluster(cl) # does not work when child is busy
        }

        add_workers(2)
        on.exit(kill_workers(structure(pool, class = c("SOCKcluster", "cluster"))), add = TRUE)
    }


    releaseJSfile <- file.path(tempdir(), "RELEASE.js")
    generateReleaseJS(releaseJSfile)

    utils::capture.output(type = "output", {
        subPathStatic <- function (targetDirectory, subPath) {
            static <- Rook::File$new(targetDirectory)

            function (env) {
                env$PATH_INFO <- unlist(strsplit(env$PATH_INFO, subPath, fixed=TRUE))[2]
                static$call(env)
            }
        }

        app <- Rook::URLMap$new(`/phantasus/ocpu` = opencpu:::rookhandler("/phantasus/ocpu", worker_cb=run_worker),
                                `/phantasus/geo` = subPathStatic(getPhantasusConf("cache_folders")$geo_path, '/phantasus/geo'),
                                `/phantasus/preloaded` = subPathStatic(cacheDir, '/phantasus/'),
                                `/phantasus/RELEASE.js` = subPathStatic(tempdir(), '/phantasus/'),
                                `/phantasus/?` = subPathStatic(staticRoot, '/phantasus/'),
                                `/?` = Rook::Redirect("/phantasus/index.html"))

        tryCatch({
            server <- startServer(host, port, app = app)
            message(sprintf(
                "Server was started with following parameters: host=%s, port=%s",
                host,
                port))
        },
        error = function(e) {
            stop(paste(e,
                       "The reason may be that requested port", port,
                       "is occupied with some other application"))
        })


        if (openInBrowser) {
            url <- sprintf("http://%s:%s", host, port)
            utils::browseURL(url)
            message(paste(url, "have been opened in your default browser.\n",
                          "If nothing happened, check your 'browser'",
                          "option with getOption('browser')",
                          "or open the address manually."))
        }

        if (background) {
            return(server)
        }

        on.exit(stopServer(server))
        service(0)

    }, split=!quiet)

    return(server)
}

probeDataset <- function (es, query) {
    exprsIndices <- query$exprs;
    fDataQuery <- query$fData;
    pDataQuery <- query$pData;
    response <- list()
    response[['dims']] <- dim(exprs(es))
    response[['fvarLabels']] <- colnames(fData(es))
    response[['varLabels']] <- colnames(pData(es))
    response[['probe']] <- exprs(es)[exprsIndices]
    response[['fdata']] <- list()
    response[['pdata']] <- list()
    if (!is.null(nrow(fDataQuery))) {
        for(i in seq_len(nrow(fDataQuery))) {
            row <- fDataQuery[i,]
            response[['fdata']][[row$name]] <- fData(es)[[row$name]][row$indices[[1]]]
        }
    }
    if (!is.null(nrow(pDataQuery))) {
        for(i in seq_len(nrow(pDataQuery))) {
            row <- pDataQuery[i,]
            response[['pdata']][[row$name]] <- pData(es)[[row$name]][row$indices[[1]]]
        }
    }
    jsonlite::toJSON(response)
}

rootFn <- c('loadGEO', 'loadPreloaded', 'createES')
complexArgumentLimitBytes <- 400

#' Reproduce session in R code
#'
#' @param sessionName String, OCPU session name
#' @param leaf Boolean, is it leaf (default = F)
#' @param step Integer, step of recursion (default = 0)
#' @param savedEnv Environment, where to store complex arguments (default = new.env())
#'
#' @return JSON with R code
#' @importFrom utils object.size
#' @examples
#' \dontrun{
#'   setwd(tempdir())
#'   reproduceInR('x039f1672026678');
#' }
#' @keywords internal
reproduceInR <- function (sessionName, leaf = T, step = 0, savedEnv = new.env()) {
    ocpuRoot <- strsplit(getwd(), 'ocpu-temp')[[1]][1]
    sessionPath <- paste(ocpuRoot, 'ocpu-store', sessionName, sep=.Platform$file.sep)
    RDataPath <- paste(sessionPath, '.RData', sep=.Platform$file.sep)
    REvalPath <- paste(sessionPath, '.REval', sep=.Platform$file.sep)

    rds <- readRDS(REvalPath)
    env <- new.env()
    loadedVariables <- load(RDataPath, envir = env)
    parsed <- parse(text=rds[[1]]$src)
    fn <- parsed[[1]][[1]]
    complexArguments <- "";
    for(i in 2:length(parsed[[1]])) {
        argumentName <- toString(parsed[[1]][[i]])
        if (argumentName %in% loadedVariables) {
            argumentValue <- env[[argumentName]]
            if (object.size(argumentValue) > complexArgumentLimitBytes) {
                rawArgument <- paste(argumentName, '_', step, sep='')
                savedEnv[[rawArgument]] <- argumentValue
            } else {
                rawArgument <- paste(deparse(argumentValue), collapse = '\n')
            }

            complexArgument <- paste(argumentName, ' <- ', rawArgument)
            complexArguments <- paste(complexArguments, complexArgument, sep='\n')
        }
    }

    if ('es' %in% names(parsed[[1]])) {
        parsedFnCall <- parsed[[1]]
        parsedFnCall$es <- parsedFnCall$es[[3]]
        rawFnCall <- paste(deparse(parsedFnCall), collapse = '')
        myHistory <- paste(complexArguments, rawFnCall, sep="\n")
    } else {
        myHistory <- paste(complexArguments, rds[[1]]$src, sep='\n')
    }

    if (!is.element(toString(fn), rootFn)) {
        parent <- parsed[[1]]$es[[2]]
        parentVar <- parsed[[1]]$es[[3]]
        parentHistory <- reproduceInR(parent, leaf = F, step = step + 1, savedEnv = savedEnv)
        myHistory <- paste(parentHistory, myHistory, sep='\n')
    }

    if (leaf) {
        assign('env', savedEnv, envir = parent.frame())
        myHistory <- paste('library(phantasus)', 'load(\'~/Downloads/env.rda\') # Please note', myHistory, sep = '\n')
        return(jsonlite::toJSON(myHistory))
    }
    myHistory
}

#'
#' @import Biobase
#' @import limma
#' @import edgeR
#'
tmmNormalization <- function(es, fieldName, logratioTrim, sumTrim) {
    if (fieldName != "None"){
        fieldValues <- es[[fieldName]]
    } else {
        fieldValues <- NULL
    }
    es.copy <- es
    count_factors <- DGEList(counts = exprs(es.copy), group = fieldValues)
    count_factors <- calcNormFactors(count_factors, logratioTrim = as.numeric(logratioTrim), sumTrim = as.numeric(sumTrim))
    exprs(es.copy) <- cpm(count_factors)

    assign("es", es.copy, envir = parent.frame())
    f <- tempfile(pattern = "norm_counts", tmpdir = getwd(), fileext = ".bin")
    data <- as.matrix(exprs(es.copy))
    colnames(data) <- NULL
    row.names(data) <- NULL
    rownames <- rownames(es.copy)

    writeBin(protolite::serialize_pb(list(data = data, colMetaNames = varLabels(es.copy))), f)
    return( jsonlite::toJSON(f))
}


#'
#' @import Biobase
#' @import limma
#' @import edgeR
#'
voomNormalization <- function(es, designData, filterByExp = FALSE){
    designMatrix <- getDesignMatrix(designData)
    es.copy <- es
    keep <- TRUE
    if (filterByExp){
        keep <- filterByExpr(exprs(es.copy), designMatrix)
    }
    es.copy <- es.copy[keep,]
    voom_counts <- voom(counts = exprs(es.copy), design = designMatrix)

    exprs(es.copy) <- voom_counts$E
    assayDataElement(es.copy, "weights") <- voom_counts$weights

    assign("es", es.copy, envir = parent.frame())
    f <- tempfile(pattern = "voom_counts", tmpdir = getwd(), fileext = ".bin")
    data <- as.matrix(exprs(es.copy))
    colnames(data) <- NULL
    row.names(data) <- NULL

    writeBin(protolite::serialize_pb(list(data = data, keep = which(keep)-1)), f)
    return( jsonlite::toJSON(f))
}

PROTOBUF_LAYOUT_VERSION = c(0x00, 0x02)

getIndicesVector <- function(current, neededLength) {
    if (length(current) == 0) {
        current <- 0:(neededLength - 1)
    }
    current + 1
}


#' Reads ExpressionSet from a GCT file.
#' Function is deprecated, please use phantasusLite:::readGct() instead
#'
#' @param ... parameters for phantasusLite:::readGct() call
#'
#' @return ExpressionSet object
#' @export
read.gct <- function(...) {
    warning("phantasus::read.gct() function is deprecated, please use phantasusLite:::readGct() instead")
    phantasusLite::readGct(...)
}

read.tsv <- function(file, header = TRUE, sep = "\t", quote = "",
                        comment.char = "",
                        check.names = FALSE, ...) {
    args <- list(...)
    res <- utils::read.table(file, header = header, sep = sep, quote = quote,
                    comment.char = comment.char, check.names = check.names,
                    stringsAsFactors = FALSE,
                    ...)
    if ( (!"row.names" %in% names(args)) && (colnames(res)[1] == "") ) {
        rownames(res) <- res[, 1]
        res[[1]] <- NULL
    }
    res
}

#' Saves ExpressionSet to a GCT file (version 1.3).
#' Function is deprecated, please use phantasusLite:::writeGct() instead
#' @param ... parameters for phantasusLite:::writeGct() call
#' @return Result of the closing file (as in `close()` function`)
#' @import Biobase
#' @export
write.gct <- function(...) {
    warning("phantasus::write.gct() function is deprecated, please use phantasusLite:::writeGct() instead")
    phantasusLite::writeGct(...)
}


makeAnnotated <- function(data) {
    meta <- data.frame(labelDescription = colnames(data))
    rownames(meta) <- colnames(data)

    methods::new("AnnotatedDataFrame", data = data, varMeta = meta)
}

take <- function(x, n) {
    sapply(x, function(x) {
        x[[n]]
    })
}

writeToList <- function(es) {
    data <- as.matrix(exprs(es))
    colnames(data) <- NULL
    row.names(data) <- NULL

    pdata <- pData(es)
    row.names(pdata) <- NULL
    pdata <- as.list(pdata)

    rownames <- rownames(es)

    fdata <- fData(es)
    row.names(fdata) <- NULL
    fdata <- as.list(fdata)


    ed <- experimentData(es)
    experimentList <- as.list(expinfo(ed))
    experimentList$other <- as.list(ed@other)
    experimentList$pubMedIds <- pubMedIds(ed)

    res <- list(data = data, pdata = pdata, fdata = fdata,
                rownames = rownames,
                colMetaNames = varLabels(es),
                rowMetaNames = fvarLabels(es),
                experimentData = experimentList)
    res
}

#' Update archs4 files.
#'
#' Download archs4 or archs4zoo counts in \code{cacheDir}. If directory does not exists function makes nothing and produce corresponding warnings.
#' @importFrom utils download.file
#' @param cacheDir file path to \bold{archs4} cache directory
#' @param organism vector which determines organisms to download: human, mouse, zoo or all as default.
#' Also can be a genus. Possible genus: \enumerate{ \item drosophila \item gallus \item bos \item caenorhabditis
#' \item danio \item rattus \item saccharomyces \item arabidopsis}
#' @param force logical value which let function replace current files
#' @keywords internal
updateARCHS4 <- function (cacheDir = file.path(getPhantasusConf("cache_folders")$rnaseq_counts, "archs4"), organism = c("all"), force = FALSE){
    zoo_dict <- c("drosophila" = "Drosophila_melanogaster",
                 "bos" = "Bos_taurus",
                 "caenorhabditis" = "Caenorhabditis_elegans",
                 "danio" = "Danio_rerio",
                 "gallus" = "Gallus_gallus",
                 "rattus" = "Rattus_norvegicus",
                 "saccharomyces" = "Saccharomyces_cerevisiae",
                 "arabidopsis" = "Arabidopsis_thaliana"
               )
    organism <- tolower(organism)
    if (any(organism %in% c("all","human")) && (force || !file.exists(file.path(cacheDir, "human_gene_v2.2.h5")))) {
        download.file(url = "https://s3.dev.maayanlab.cloud/archs4/files/human_gene_v2.2.h5",
                      destfile = file.path(cacheDir, "human_gene_v2.2.h5"),
                      mode = "wb")
    }
    if (any(organism  %in% c("all","mouse")) && (force || !file.exists(file.path(cacheDir, "mouse_gene_v2.2.h5")))) {
        download.file(url = "https://s3.dev.maayanlab.cloud/archs4/files/mouse_gene_v2.2.h5",
                      destfile = file.path(cacheDir, "mouse_gene_v2.2.h5"),
                      mode = "wb")
    }
    for (sp in names(zoo_dict)){
        if (any(organism %in% c("all","zoo", sp))) {
            sp_file_name <- paste0(zoo_dict[sp], "_genecount_v1.h5")
            sp_file_path <-  file.path(cacheDir, sp_file_name)
            if (force || !file.exists(sp_file_path)) {
                download.file(url = paste0("https://s3.amazonaws.com/mssm-archs4-zoo/", sp_file_name),
                              destfile = sp_file_path,
                              mode = "wb")
            }
        }
    }
}

#' Update ARCHS4 meta files
#'
#' Creates \code{meta.txt} file, which describes typical archs4 and archs4Zoo files.
#' @param archDir path to directory with arch4 .h5 files.
#' @details This function produces very specific "hardcoded" \code{meta.txt} file for arch4 and archs4ZOO counts collections.
#' See \code{\link{validateCountsCollection}} for more common information and \code{meta.txt}  file structure
#' @seealso \code{\link{validateCountsCollection}}
#' @import data.table
#' @keywords internal
updateARCHS4meta <- function(archDir = file.path(getPhantasusConf("cache_folders")$rnaseq_counts, "archs4")){
    archs4files <- list.files(archDir, pattern = '\\.h5$')
    DT_meta <- data.frame(matrix(ncol = 5, nrow = length(archs4files), dimnames = list(NULL, c("file_name", "sample_id", "sample_dim",	"gene_id", "genes_annot"))))
    DT_meta$file_name <- archs4files
    genus <- tolower(sapply(strsplit(x = archs4files, split = "_"), function(x) x[1]))
    for (i_file in seq_along(archs4files)) {
        cur_file <- file.path(archDir, archs4files[i_file])
        h5f <- H5Fopen(cur_file, flags = "H5F_ACC_RDONLY")
        arch_version <- if (H5Lexists(h5f, "info/version")) {
            h5read(h5f, "info/version")
        } else {
            h5read(h5f, "meta/info/version")
        }
        gene_id_type <- NA
        if (genus[i_file] %in% c("human", "mouse")) {
            gene_id_type <- "Gene Symbol"
        } else if (genus[i_file] %in% c("rattus", "bos", "gallus", "danio")) {
            gene_id_type <- "ENSEMBLID"
        } else if (genus[i_file] %in% c("arabidopsis")) {
            gene_id_type<- "TAIR id"
        } else if (genus[i_file] %in% c("saccharomyces")) {
            gene_id_type <- "ORF id"
        } else if (genus[i_file] %in% c("caenorhabditis")) {
            gene_id_type <- "WormBase id"
        } else if (genus[i_file] %in% c("drosophila")) {
            gene_id_type <- "FlyBase id"
        } else{
            gene_id_type<- "gene"
        }
        if (is.na(arch_version)) {
            arch_version <- "8"
        }
        if (arch_version == "1"){
            DT_meta$sample_id[i_file] <-  "/meta/Sample_geo_accession"
            DT_meta$gene_id[i_file] <-  paste(gene_id_type, "/meta/genes", sep = ":")
            DT_meta$sample_dim[i_file] = "columns"
            annot_str <- ""
            DT_meta$genes_annot[i_file] = annot_str
        }
        if (arch_version %in% c("7", "8")){
            DT_meta$sample_id[i_file] <-  "/meta/Sample_geo_accession"
            DT_meta$gene_id[i_file] <-  paste(gene_id_type, "/meta/genes", sep = ":")
            DT_meta$sample_dim[i_file] = "columns"
            annot_str <- "Gene ID:/meta/gene_entrezid"
            if (!toupper(gene_id_type) == "GENE SYMBOL") {
              annot_str <- paste(annot_str, "Gene symbol:/meta/genes", sep = ";")
            }
            if (!toupper(gene_id_type) == "ENSEMBLID") {
                annot_str <- paste(annot_str, "ENSEMBLID:/meta/gene_ensemblid", sep = ";")
            }
            DT_meta$genes_annot[i_file] = annot_str
        }
        if (arch_version %in% c("9", "10", "11")) {
            DT_meta$sample_id[i_file] <- "/meta/samples/geo_accession"
            DT_meta$gene_id[i_file] <-  paste(gene_id_type, "/meta/genes/genes", sep = ":")
            DT_meta$sample_dim[i_file] = "rows"
            annot_str <- ""
            if (!toupper(gene_id_type) == "GENE SYMBOL") {
                annot_str <-  "Gene symbol:/meta/genes/gene_symbol"
            }
            if (!toupper(gene_id_type) == "ENSEMBLID") {
                annot_str <- paste(annot_str, "ENSEMBLID:/meta/genes/ensembl_gene_id", sep = ";")
            }
            DT_meta$genes_annot[i_file] = trimws(annot_str, which = "left", whitespace = ";")
        }
        if(arch_version %in% c("2.2" , "2.2.1")){
            gene_id_type <- "ENSEMBLID"
            DT_meta$sample_id[i_file] <- "/meta/samples/geo_accession"
            DT_meta$gene_id[i_file] <-  paste(gene_id_type, "/meta/genes/ensembl_gene_id", sep = ":")
            DT_meta$sample_dim[i_file] = "rows"
            annot_str <- "Gene symbol:/meta/genes/symbol"
            DT_meta$genes_annot[i_file] = annot_str
        }

        H5Fclose(h5f)
    }

    write.table(x = DT_meta,
                file = file.path(archDir, "meta.txt"),
                sep = "\t",
                col.names = TRUE, row.names = FALSE,
                quote = FALSE)
}
#' Update DEE2 meta files
#'
#' Creates \code{meta.txt} file, which describes typical dee2 files.
#' @param destDir path to directory with DEE2 .h5 files.
#' @details This function produces very specific "hardcoded" \code{meta.txt} file for dee2 counts colletction.
#' See \code{\link{validateCountsCollection}} for more common information and \code{meta.txt}  file structure
#' @seealso \code{\link{validateCountsCollection}}
#' @import data.table
#' @keywords internal
updateDEE2meta <- function(destDir = file.path(getPhantasusConf("cache_folders")$rnaseq_counts, "dee2")){
    dee2files <- list.files(destDir, pattern = '\\.h5$')
    DT_meta <- data.frame(matrix(ncol = 5, nrow = length(dee2files), dimnames = list(NULL, c("file_name", "sample_id", "sample_dim",	"gene_id", "genes_annot"))))
    DT_meta$file_name <- dee2files
    DT_meta$sample_dim <- "rows"
    DT_meta$sample_id <- "/meta/samples/geo_accession"
    genus <- sapply(strsplit(x = dee2files, split = "_"), function(x) x[1])
    for (i_file in 1:length(dee2files)) {
        if (genus[i_file] %in% c("hsapiens", "mmusculus", "drerio", "rattus")) {
            gene_id_type <- "ENSEMBLID"
        } else if (genus[i_file] %in% c("athaliana", "ecoli")) {
            gene_id_type <- "Locus tag"
        } else if (genus[i_file] %in% c("scerevisiae")) {
            gene_id_type <- "Yeast id"
        } else if (genus[i_file] %in% c("dmelanogaster")){
            gene_id_type <- "FlyBase id"
        } else if (genus[i_file] %in% c("celegans")) {
            gene_id_type <- "WormBase id"
        } else{
            gene_id_type <- "gene"
        }
        DT_meta$gene_id[i_file] <- paste(gene_id_type, "/meta/genes/ensembl_gene_id", sep = ":")

        DT_meta$genes_annot[i_file]  <- "Gene symbol:/meta/genes/gene_symbol"
    }
    write.table(x = DT_meta, file = file.path(destDir, "meta.txt"), sep = "\t", col.names = TRUE, row.names = FALSE, quote = FALSE)
}

stopPhantasus <- function(){
    stop("Phantasus is not configured. Run phantasus::setupPhantasus() to complete setup.")
}

areCacheFoldersValid <- function(cache_folders){
    for (folder in cache_folders[names(cache_folders) != "rnaseq_counts"]) {
        folder <- normalizePath(folder)
        if (!dir.exists(folder) || !rw_dir_check(folder)){
            return(FALSE)
        }

    }
    counts_source <- getPhantasusConf("cache_folders")$rnaseq_counts

    if (nchar(counts_source) == 0){
        return(FALSE)
    }
    valid_hsds <- tryCatch(expr = isHSDS(getPhantasusConf("cache_folders")$rnaseq_counts),
                           error = function(e) {FALSE})
    options(PhantasusUseHSDS = valid_hsds)
    if (is.null(valid_hsds)){

        if (!dir.exists(counts_source)){
            return(FALSE)
        } else {
            return(rw_dir_check(counts_source))
        }
    }
    return(valid_hsds)
}
rw_dir_check <- function(dir_name){
    if (file.access(dir_name, mode = 2) != 0){
        stop(paste("!Bad configuration:", dir_name , " is not writable"))
    }
    if (file.access(dir_name, mode = 4) != 0){
        stop(paste("!Bad configuration:", dir_name , " is not readable"))
    }
    return(TRUE)
}
selfCheck <- function(verbose = FALSE) {
    cacheDir = getPhantasusConf("cache_root")
    preloadedDir = getPhantasusConf("preloaded_dir")
    if (!is.null(preloadedDir) && dir.exists(preloadedDir)) {
        preloadedFiles <- list.files(preloadedDir, pattern = "\\.(gct|rda)$")
        message(paste(length(preloadedFiles), 'preloaded datasets are available'))
        if (verbose) {
            message(paste0(preloadedFiles,  collapse = " "))
            message(" ")
        }
    } else {
        message('Preloaded dir is not set')
    }
    use_hsds <- getOption("PhantasusUseHSDS")
    h5counts <- list()
    if (is.null(use_hsds)){
        counts_dir <- getPhantasusConf("cache_folders")$rnaseq_counts
        if (!isCountsPriorityValid(counts_dir) || isCountsMetaOld(counts_dir)){
            message("!! RNA-seq counts was not installed properly.")
            stopPhantasus()
        }
        h5counts <- list.files(counts_dir, recursive = TRUE,
                               pattern = '\\.h5$')
    } else if (use_hsds == TRUE) {
        if (nchar(system.file(package = "phantasusLite")) == 0){
            stopPhantasus()
        }
        h5counts <- phantasusLite::getHSDSFileList(getPhantasusConf("cache_folders")$rnaseq_counts)
    }
    archs4Files <- list.files(file.path(cacheDir, "archs4"),
                              pattern = '\\.h5$')
    if (length(h5counts)) {
        message(paste(length(h5counts), 'counts files are available'))
        if (verbose) {
            message(paste0(h5counts, collapse = " "))
            message(" ")
        }
        if (length((archs4Files))) {
            message("!!! deprecated archs4-based directory architecture exists. It will be ignored and RNA-seq will load only from counts directory.")
        }
    } else {
        if (length(archs4Files)) {
            message("!!! archs4-based directory architecture is obsolete and will dissapear in future release")
            message(paste(length(archs4Files), 'archs4 files are available'))
            if (verbose) {
                message(paste0(archs4Files, collapse = " "))
                message(" ")
            }
        } else {
            message('!!! No counts provided RNA-seq will load without matrices')
        }
    }
    annotDir <- getPhantasusConf("cache_folders")$annot_db
    dbFiles <- list.files(annotDir, pattern = '\\.sqlite$')
    if (length(dbFiles)) {
        message(paste(length(dbFiles), 'annotationDb are available'))
        if (verbose) {
            message(paste0(dbFiles, collapse = " "))
            message(" ")
        }
    } else {
        message('!!! No annotationDb provided')
    }

    fgseaDir <-  getPhantasusConf("cache_folders")$fgsea_pathways
    fgseaFiles <- list.files(fgseaDir, '\\.rds$', full.names = FALSE)
    if (length(fgseaFiles)) {
        message(paste(length(fgseaFiles), 'fgsea tables are available'))
        if (verbose) {
            message(paste0(fgseaFiles, collapse = " "))
            message(" ")
        }
    } else {
        message('!!! No fgsea tables provided')
    }
}

#' check if url  responding as HSDS server
#' TRUE - hsds
#' FALSE - web link but not working
#' NULL - not web link
#' @param url URL to check
#' @keywords internal
isHSDS <- function(url){
    if(!grepl(pattern = "^http(s)?://", x = url)){
        return(NULL)
    }
    req <- httr::GET(url)
    if( req$headers$server == "Highly Scalable Data Service (HSDS)"){
        return(TRUE)
    }
    return(FALSE)
}

safeDownload <- function(url, dir, filename, ...) {
    dest <- file.path(dir, filename)
    if (file.exists(dest)) {
        return()
    }

    tempDest <- tempfile(paste0(filename, ".load"), tmpdir = dir)
    curl::curl_download(url, destfile = tempDest, ...)
    file.rename(tempDest, dest)
}

isValidExperimentID <- function(name) {
    grepl("^((GSE|GDS)[0-9]+(-GPL[0-9]+)?)|(GPL[0-9]+)$", name, ignore.case = TRUE)
}


getGEODir <- function(name, destdir = file.path(tempdir(), "geo")) {
    if (!isValidExperimentID(name)) {
        stop(name, " does not look like a valid GEO Series ID")
    }
    type <- substr(name, 1, 3)
    GEO <- unlist(strsplit(name, "-"))[1]
    stub <- gsub("\\d{1,3}$", "nnn", GEO, perl = TRUE)
    gdsDirPath <- "%s/datasets/%s/%s/soft"
    gseDirPath <- "%s/series/%s/%s/matrix"
    gplDirPath <- "%s/platforms/%s/%s/soft"
    if (type == 'GSE') {
        fullGEODirPath <- file.path(sprintf(gseDirPath, destdir, stub, GEO))
    } else if (type == "GDS") {
        fullGEODirPath <- file.path(sprintf(gdsDirPath, destdir, stub, GEO))
    } else if (type == "GPL") {
        fullGEODirPath <- file.path(sprintf(gplDirPath, destdir, stub, GEO))
    } else {
        stop("Unsupported GEO type: ", type)
    }
    dir.create(fullGEODirPath, showWarnings = FALSE, recursive = TRUE)

    fullGEODirPath
}

getBriefData <- function(name, destdir = tempdir()) {
    GEO <- unlist(strsplit(name, "-"))[1]
    GEOdir <- dirname(getGEODir(GEO, destdir))
    briefFile <- file.path(GEOdir, 'brief')

    if (file.exists(briefFile)) {
        message('Using cached brief file: ', briefFile)
    } else {
        url <- sprintf("https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=%s&targ=self&form=text&view=brief", GEO)
        message('Trying ', url)
        resp <- httr::GET(url)
        text <- httr::content(resp, "text", "UTF-8")
        check <- grep('Could not', text)
        if (length(check) || httr::status_code(resp) != 200) {
            message('No such dataset: ', name)
            unlink(GEOdir, recursive = TRUE, force = TRUE)
            stop('Failed to download brief data on: ', GEO, '. No such dataset')
        } else {
            writeLines(text, briefFile)
            message('Stored brief data of ', GEO, ' at ', briefFile)
        }
    }
    parsedList <- parseBriefData(readLines(briefFile))
    if (length(parsedList) == 0) {
        file.remove(briefFile)
        stop('Failed to parse brief data on: ', GEO, '. Empty list')
    }

    return (parsedList)
}

parseBriefData <- function(txt) {
    tmp <- txt[grep("!\\w*?_", txt)]
    tmp <- gsub("!\\w*?_",'', tmp)
    first.eq <- regexpr(' = ', tmp)
    tmp <- cbind(substring(tmp, first = 1, last = first.eq - 1),
                 substring(tmp, first = first.eq + 3))
    tmp <- tmp[tmp[,1] != "",]
    header <- split(tmp[,2],tmp[,1])
    return(header)
}

#' Create meta-data for single counts collection
#'
#' Creates a part of counts collections meta-data
#' @param counts_dir path to directory with count collections
#' @param collection_name name of collection and collection's directory
#' @param verbose logical value which determines a content of  the output.
#' @details Function assumes that \code{collection_name} contains \code{meta.txt} which is valid (in sence of \code{\link{validateCountsCollection}}).
#' For each row in \code{meta.txt} function reads specified \code{sample_id} dataset and writes every sample id to the resulting \code{data.table} with source file name and collection name.
#'
#' @return \code{data.table} with meta-data or nothing if \code{destdir} does not exist or does not contain files.
#' @seealso {\code{\link{validateCountsCollection}},\code{\link{getCountsMetaPart}}}
#' @examples
#' \dontrun{
#'     collDir <- "/path/to/my/collection"
#'     valid_collection = validateCountsCollection(collectionDir = collDir, verbose = TRUE)
#'     if (valid_collection){
#'         metaPart = getCountsMetaPart(destdir = collDir, verbose = TRUE)
#'      }
#'  }
#'
#'
#' @import data.table
#' @keywords internal
getCountsMetaPart <- function(counts_dir, collection_name, verbose){
    destdir <- file.path(counts_dir, collection_name)
    if (!dir.exists(destdir)) {
        return()
    }
    DT_h5_meta <- data.table()
    h5_files <- list.files(destdir, "\\.h5$", full.names = FALSE)
    if (!length(h5_files)) {
        return()
    }
    h5_meta <- fread(file.path(destdir, "meta.txt"), index = "file_name")
    subparts <- list()
    for (input_file in h5_files) {
        if (input_file %in% h5_meta$file_name) {
            full_name <- file.path(destdir, input_file)
            relative_path <- file.path(collection_name, input_file )
            h5f <- H5Fopen(full_name, flags = "H5F_ACC_RDONLY")
            h5_part <- data.table(accession = h5read(h5f, h5_meta[file_name == input_file, ]$sample_id),
                                 file = relative_path,
                                 collection_type = collection_name)
            H5Fclose(h5f)
            subparts <- c(subparts, list(h5_part))
            if (verbose) {
                message("added ", nrow(h5_part), " samples from ", file.path(destdir, input_file))
            }
        } else {
            if (verbose) {
                message("!! ", file.path(destdir, input_file), " is ignored")
            }
        }

    }
    DT_h5_meta <- rbindlist(subparts)
    return(DT_h5_meta)
}

isCountsMetaOld <- function( counts_dir = getPhantasusConf("cache_folders")$rnaseq_counts){
    meta_name <- file.path(counts_dir, "meta.rda")
    h5_files <- list.files(file.path(counts_dir), "\\.h5$", full.names = TRUE, recursive = TRUE)
    meta_time <- as.numeric(file.mtime(meta_name))
    h5_mtime <- max(unlist(lapply(h5_files, file.mtime)))
    list_dirs <-  list.dirs(counts_dir, full.names = FALSE, recursive = TRUE)
    dirs_mtime <- lapply(file.path(counts_dir, list_dirs[-1]), file.mtime)
    if (length(dirs_mtime) > 0) {
        dir_mtime <- max(unlist(dirs_mtime))
    } else {
        dir_mtime <- -Inf
    }
    if (file.exists(meta_name) && meta_time > h5_mtime && meta_time > dir_mtime) {
        return(FALSE)
    }
    if (!file.exists(meta_name) && length(h5_files) == 0){
        return(FALSE)
    }
    return(TRUE)
}

isCountsPriorityValid <- function(counts_dir =  getPhantasusConf("cache_folders")$rnaseq_counts){
    priority_file <- file.path(counts_dir, "counts_priority.txt")
    if (!file.exists(priority_file)){
        h5_files <- list.files(file.path(counts_dir), "\\.h5$", full.names = TRUE, recursive = TRUE)
        if (length(h5_files) > 0){
            return(FALSE)
        } else {
            return(TRUE)
        }
    }
    list_dirs <-  list.dirs(counts_dir, full.names = FALSE, recursive = TRUE)
    list_dirs <- c(".", list_dirs[-1])
    priority <- fread(priority_file)
    if (!(setequal(priority$directory,list_dirs) && length(unique(priority$priority)) == length(priority$priority))) {
        return (FALSE)
    }
    return(TRUE)
}

#' Update meta-data for counts collections
#'
#' Creates \code{meta.rda} file which contain information about all samples in all collections.
#' Also function checks \code{priority.txt} file. This file is used to manage collections with the same samples.
#' @param counts_dir path to counts cache directory
#' @param verbose logical value which determines a content of  the output.
#' @param force logical value wich lets function replace existing \code{meta.rda} file
#' @details  First of all function checks validity of \code{priority.txt} file. \bold{Every} Collection should have \bold{unique} priority.
#' If \code{priority.txt} is not valid function creates new one, setting priorities for each subdirectory(=collection) equal to order in \code{list.dir} output.
#'
#' Function updates \code{meta.rda} if this file is older than at least one \code{.h5} file in counts files.
#' \code{meta.rda} is \code{data.table} which is a result of union \code{data.table}s produced by \code{\link{getCountsMetaPart}} for each collection
#' @seealso {\code{\link{validateCountsCollection}},\code{\link{updateCountsMeta}}}
#' @import data.table
#' @keywords internal
updateCountsMeta <- function(counts_dir =  getPhantasusConf("cache_folders")$rnaseq_counts, force = FALSE, verbose = FALSE){
    if (!dir.exists(counts_dir)) {
        message(paste0('Counts directory ', counts_dir, " does not extist" ))
        return()
    }
    meta_name <- file.path(counts_dir, "meta.rda")
    h5_files <- list.files(file.path(counts_dir), "\\.h5$", full.names = TRUE, recursive = TRUE)
    list_dirs <-  list.dirs(counts_dir, full.names = FALSE, recursive = TRUE)
    list_dirs <- c(".", list_dirs[-1])
    priority_file <- file.path(counts_dir, "counts_priority.txt")

    need_create <- FALSE
    if (!isCountsPriorityValid(counts_dir)){
        message(paste0("!!! Counts priority file is invalid or missed. Create default one."))
        need_create <- TRUE
    }
    if (need_create) {
        priority <- data.table(directory = list_dirs, priority = seq_along(list_dirs))
        write.table(x = priority, file = priority_file, sep = "\t", eol = "\n", row.names = FALSE, col.names = TRUE, quote = FALSE)
    }
    if (!length(h5_files)) {
      return()
    }
    if (!force && !isCountsMetaOld(counts_dir)) {
        return()
    }
    if (file.exists(meta_name)) {
      unlink(meta_name)
    }
    DT_counts_meta <- data.table(matrix(ncol = 3, nrow = 0, dimnames = list( NULL, c("accession", "file", "collection_type"))))
    for (cur_dir in list_dirs) {
        dir_path <- file.path(counts_dir, cur_dir)
        dir_path <- sub(pattern = "/./?$", replacement = "", x =  dir_path)
        cur_files <- list.files(path = dir_path, pattern = "\\.h5", recursive = FALSE )
        if (length(cur_files) == 0) {
          next
        }
        if (!file.exists(file.path(dir_path, "meta.txt"))) {
            if (startsWith(x = tolower(basename(dir_path)), prefix =  "archs4")) {
              updateARCHS4meta(archDir = dir_path)
            } else if (startsWith(x = tolower(basename(dir_path)), prefix = "dee2")) {
              updateDEE2meta(destDir = dir_path)
            }
        }
        message(paste0('Populating ', cur_dir , ' counts meta' ))
        if (!validateCountsCollection(collectionDir = dir_path, verbose = verbose)) {
            message(paste0("!! files in ", cur_dir , " are ignored because there is not correct meta file in this directory."))
            next
        }
        DT_part <- getCountsMetaPart(counts_dir = counts_dir, collection_name = cur_dir, verbose = verbose)
        if (length(DT_part)) {
            DT_counts_meta <- rbindlist(l = list(DT_counts_meta, DT_part))
        }
        rm(DT_part)
    }
    save(DT_counts_meta, file = meta_name, eval.promises = TRUE)
    rm(DT_counts_meta)
}

#' Check a counts collection
#'
#' Function checks existing  and  structure of \code{meta.txt} file in specified counts folder.Also it checks accessibility of specified datasets in  corresponding \code{.h5} files.
#' @param collectionDir path to directory with collection
#' @param verbose logical value which determines a content of  the output.
#' @details  \code{collectionDir} should contain a bunch of \code{.h5} files
#'  and a single \code{meta.txt}.  \code{meta.txt} is \code{.tsv}-like file where for each \code{.h5} exists a row wit columns:
#' \describe{ \item{file_name}{ name of \code{.h5} file in \code{collectionDir}.}
#'            \item{sample_id}{name of dataset in \code{file_name} which contains sample IDs (sample_geo_accession for example).}
#'            \item{sample_dim}{which dimension of the expression matrix in \code{file_name} corresponds to samples. Should be one of \code{c("rows", "columns")} }
#'            \item{gene_id}{name of dataset in \code{file_name} which contains ids for genes and the "meaning" for that ids( column name in result ES). For correct work this dataset should contain unique values. Example: ENSEMBLID:/meta/genes/ensembl_gene_id}
#'            \item{genes_annot}{Names of datasets and their meanings to extract gene-related metadata from \code{file_name}. Can be empty or \code{gene_id}-like values separated with semicolon(;).}
#'            }
#'
#' @import data.table
#' @import rhdf5
#' @keywords internal
validateCountsCollection <- function(collectionDir, verbose=FALSE){
    if (!file.exists(file.path(collectionDir, "meta.txt"))) {
        if (verbose) {
            message(paste0("metafile does not exist in ",  file.path(collectionDir)))
        }
        return(FALSE)
    }

    h5_meta <- fread(file.path(collectionDir, "meta.txt"), index = "file_name")
    for (input_file  in h5_meta$file_name) {
        full_path <- file.path(collectionDir, input_file)
        cur_meta <- h5_meta[file_name == input_file, ]
        if (nrow(cur_meta) > 1) {
            if (verbose) {
                message(paste0("two or more rows in meta file for ", full_path ))
            }
            return(FALSE)
        }
        if (!file.exists(full_path)){
            message("Skipping absent file ", full_path)
            next
        }
        h5f <- H5Fopen(full_path, flags = "H5F_ACC_RDONLY")

        tryCatch({
            is_sample_valid <- H5Lexists(h5f, name =  cur_meta$sample_id)

            if(!is_sample_valid){
                if (verbose) {
                    message(paste0("can't read sample_id in ", full_path ))
                }
                return(FALSE)
            }
            gene_id <- strsplit(cur_meta$gene_id, split = ":")[[1]]
            if (length(gene_id) != 2){
                if (verbose){
                    message(paste0("wrong gene id format for", full_path))
                }
                return(FALSE)
            }
            gene_ids <- if (H5Lexists(h5f, name = gene_id[2])) {
                h5read(h5f, name = gene_id[2])
            } else NULL

            if (length(gene_ids) == 0) {
                if (verbose) {
                    message(paste("can't read gene_id in ", full_path))
                }
                return(FALSE)
            }
            if (length(gene_ids)  != length(unique(gene_ids))) {
                if (verbose) {
                    message(paste("Non-unique gene ids in file", full_path))
                }
                return(FALSE)
            }
        }, finally = {
            H5Fclose(h5f)
        })
    }
    return(TRUE)
}


checkBinValidity <- function(filePath, valid_from) {
  if (!file.exists(filePath)) {
    return(FALSE)
  }

  if (file.info(filePath)$ctime < valid_from) {
    return(FALSE)
  }

  raw_proto_version <- as.raw(PROTOBUF_LAYOUT_VERSION)
  bin_ref <- protolite::serialize_pb(list(raw_proto_version))
  file_head <- readBin(con = filePath,what = raw(), n = length(bin_ref))

  if (!all(bin_ref == file_head)) {
    return(FALSE)
  }

  return(TRUE)
}

getDesignMatrix <- function(designData){
    designMatrix <- do.call(cbind, designData)
    designRownames <- designMatrix[,"id"]

    designMatrix <- subset(designMatrix, select = -c(id))
    designMatrix <- apply(X = designMatrix, FUN = as.numeric, MARGIN =c(2))
    colnames(designMatrix) [colnames(designMatrix) == "intercept"] <- "(Intercept)"
    rownames(designMatrix) <- designRownames
    if(qr(designMatrix)$rank < ncol(designMatrix)){
        stop("Error: redundancy of model parameters appears. Try to exclude nested factors.")
    }
    return(designMatrix)
}

phantasusVersion <- function() {
  jsonlite::toJSON(as.character(utils::packageVersion("phantasus")))
}

generateReleaseJS <- function(destfile, build="none") {
    lines <- c(sprintf("window.PHANTASUS_VERSION='%s';", as.character(utils::packageVersion("phantasus"))),
      sprintf("window.PHANTASUS_BUILD='%s';", build))
    writeLines(lines, destfile)

}

context('counts files')
library(rhdf5)
Sys.setenv(R_USER_CONFIG_DIR = system.file("/testdata/config", package = "phantasus"))
old_conf <- Sys.getenv("R_CONFIG_ACTIVE")
Sys.setenv(R_CONFIG_ACTIVE = "default")
test_that("getCountsMetaPart provides correct part of meta.rda", {
    counts_dir <- system.file("testdata/counts", package = "phantasus")
    if (counts_dir == "") {
        stop("test directory for counts does not exist")
    }
    arch_dir <- file.path(counts_dir, "archs4")
    metaPart <- phantasus:::getCountsMetaPart(counts_dir = counts_dir, collection_name = "archs4",verbose = FALSE)
    meta_file <- file.path(arch_dir, "meta.txt")
    h5_meta <- fread(meta_file, index = "file_name")
    files_in_txt <- file.path(arch_dir, h5_meta$file_name)
    files_in_meta <- file.path( counts_dir,unique(metaPart$file))
    expect_setequal(files_in_meta, files_in_txt)
    samples_in_meta <- sapply(files_in_txt, function(file) {
        sample_id <- h5_meta[file_name == basename(file)]$sample_id
        gsm_in_file <- h5read(file = file, name = sample_id)
        all(gsm_in_file %in% metaPart$accession)
    })
    expect_equal(sum(samples_in_meta), length(h5_meta$file_name))
})


test_that("updateCountsMeta works with empty directories", {
    destdir = file.path(tempdir(), "empty_dir")
    dir.create(destdir)
    updateCountsMeta(counts_dir = destdir)
    meta_file <- file.path(destdir, "meta.rda")
    expect_false(file.exists(meta_file))
    unlink(destdir,recursive = TRUE)
})
test_that("updateCountsMeta works with priorities", {
    real_path <- system.file("testdata/counts/archs4", package = "phantasus")
    if (real_path == "") {
        stop("test directory for counts does not exist")
    }
    destdir <- file.path(tempdir(), "test_counts")
    dir.create(destdir, recursive = TRUE)
    tmp_path <- file.path(destdir, "archs4")
    dir.create(tmp_path)
    files <- list.files(real_path, pattern = "(.+\\.h5)|(meta.txt)", full.names = TRUE)
    file.copy(files, tmp_path)

     # without priority file
    updateCountsMeta(counts_dir = destdir)
    priority_file <- file.path(destdir, "counts_priority.txt")
    meta_file <- file.path(destdir, "meta.rda")
    expect_true(file.exists(priority_file))
    expect_true(file.exists(meta_file))
    priority <- fread(priority_file)
    expect_equal(length(priority),2)
    in_file <- priority$directory
    real_dirs <- c(".", list.dirs(destdir, full.names = FALSE, recursive =  TRUE)[-1])
    expect_true(all(in_file %in% real_dirs))

    # with prioroty file
    before_time <- file.info(priority_file)$mtime
    file.remove(meta_file)
    updateCountsMeta(counts_dir = destdir)
    expect_true(file.exists(priority_file))
    after_time <- file.info(priority_file)$mtime
    expect_equal(before_time, after_time)
    expect_true(file.exists(meta_file))

    unlink(destdir, recursive = TRUE)
})

test_that("loadCounts returns result", {
    destdir = system.file("testdata/", package = "phantasus")
    counts_dir = system.file(file.path("testdata", "counts"), package = "phantasus")
    if (counts_dir == "") {
        stop("test directory for counts does not exist")
    }
    tmp_counts <- getPhantasusConf("cache_folders")$rnaseq_counts
    if (!dir.exists(tmp_counts)){
        dir.create(tmp_counts)
    }
    dir.create(file.path(tmp_counts, "archs4"))
    count_files <- list.files(counts_dir, recursive = TRUE, full.names = FALSE)
    tmp_count_files <- file.path(tmp_counts, count_files )
    file.copy(file.path(counts_dir, count_files), tmp_count_files)
    real_wd <- getwd()
    setwd(system.file(".", package = "phantasus"))
    test_gse <- "GSE107746"
    ess <- getGSE(test_gse, mirrorPath = "https://ftp.ncbi.nlm.nih.gov")
    expect_gt(nrow(ess[[1]]), 0)
    expect_gt(ncol(ess[[1]]), 0)
    setwd(real_wd)
    unlink(tmp_counts, recursive = TRUE)

})

test_that("updateCountsMeta create good meta files for archs4", {
    destdir = system.file("testdata/", package = "phantasus")
    counts_dir = system.file(file.path("testdata", "counts"), package = "phantasus")
    if (counts_dir == "") {
        stop("test directory for counts does not exist")
    }
    tmp_counts <- getPhantasusConf("cache_folders")$rnaseq_counts
    if (!dir.exists(tmp_counts)){
        dir.create(tmp_counts, recursive = TRUE )
    }
    dir.create(file.path(tmp_counts, "archs4"))
    count_files <- list.files(counts_dir, recursive = TRUE,  pattern = ".h5$")
    tmp_count_files <- file.path(tmp_counts, count_files )
    file.copy(file.path(counts_dir, count_files), tmp_count_files)
    updateCountsMeta(counts_dir = tmp_counts)
    meta_file <- file.path(tmp_counts, "meta.rda")
    priority_file <- file.path(tmp_counts, "counts_priority.txt")
    arch_meta_file <- file.path(tmp_counts, "archs4", "meta.txt")
    expect_true(file.exists(meta_file))
    expect_true(file.exists(priority_file))
    expect_true(file.exists(arch_meta_file))
    arch_meta <- fread(arch_meta_file)
    expect_equal(nrow(arch_meta), length(count_files))
    expect_true(validateCountsCollection(file.path(tmp_counts, "archs4")))
    unlink(tmp_counts, recursive = TRUE)
})
Sys.setenv(R_CONFIG_ACTIVE = old_conf)

#' Calculate column averages in row groups
#'
#' @param m matrix n x m
#' @param groups vector of size n of numbers from 1 to k
#' @return matrix k*m of column averages  by groups
#' @keywords internal
colMeansByGroups <- function(m, groups) {
    x <- Matrix::sparseMatrix(j=seq_along(groups),
                 i=groups,
                 x=rep(1, length(groups)))
    z <- x %*% m
    res <- sweep(z, 1, table(groups), "/")
}




rasterizeHeatmap <- function(m, palette=palette, maxDimensions=c(2500, 1000)) {
    mr <- round((m - min(m))/(max(m)-min(m)) * (length(palette) - 1) + 1)
    cap <- matrix(palette[as.matrix(mr)], nrow=nrow(m))

    repN <- floor(maxDimensions/dim(m))


    cap1 <- matrix(rep(t(cap), each=repN[2]), nrow=nrow(cap), byrow = TRUE)
    cap2 <- matrix(rep(cap1, each=repN[1]), ncol=ncol(cap1), byrow = FALSE)

    # cap2[repN[1], repN[2]] == cap[1,1]
    # cap2[repN[1]+1, repN[2]] == cap[2,1]
    # cap2[repN[1], repN[2]+1] == cap[1,2]
    # cap2[repN[1]+1, repN[2]+1] == cap[2,2]

    res <- rasterGrob(cap2,
                      width=unit(1, "npc"), height=unit(1, "npc"),
                      interpolate = FALSE)
    res
}

#' Returns path to an svg file with enrichment plot
#' @param es ExpressionSet object.
#' @param rankBy name of the numeric column used for gene ranking
#' @param selectedGenes indexes of selected genes (starting from one, in the order of fData)
#' @param width width of the image (in inches)
#' @param height height of the image (in inches)
#' @param vertical whether to use vertical orientation (default: FALSE)
#' @param addHeatmap whether to add an expression heatmap, sorted by rankBy (default: FALSE)
#' @param showAnnotation a name of column annotation to add to the heatmap, default: NULL (no annotation)
#' @param pallete a vector of colors to draw heatmap
#' @param annotationColors a list of colors to use in annotation
#' @return path to an svg file
#' @importFrom fgsea plotEnrichment fgseaMultilevel
#' @importFrom ggplot2 ggtitle
#' @importFrom grDevices colorRampPalette dev.off svg
#' @importFrom utils head
#' @import grid
#' @import gtable
#' @import svglite
#' @import stats
#' @keywords internal
gseaPlot <- function(es, rankBy, selectedGenes, width, height,
                     vertical=FALSE,
                     addHeatmap=FALSE,
                     showAnnotation=NULL,
                     annotationColors=NULL,
                     pallete=c("blue", "white", "red")) {
    fullPalette <- colorRampPalette(pallete)(50)
    featureData <- fData(es)
    colnames(featureData) <- fvarLabels(es)
    eps <- 1e-10

    ranks <- featureData[, rankBy]
    if (!is.numeric(ranks)) {
        ranks <- as.numeric(ranks)
    }
    names(ranks) <- as.character(seq_along(ranks))

    pathway <- as.character(selectedGenes)


    fgseaRes <- fgseaMultilevel(list(p=pathway), ranks, sampleSize = 101, nproc=1, eps = eps/2)


    pvalString <- if (fgseaRes$pval < eps) {
        sprintf("<%.2g", eps)
    } else {
        sprintf("\u2248%.2g", fgseaRes$pval)
    }

    labelString <- sprintf("p-value%s, NES=%.2f", pvalString, fgseaRes$NES)


    p <- plotEnrichment(pathway, ranks) + ggtitle(NULL, subtitle=labelString)
    if (vertical) {
        p <- p +
            scale_x_reverse(limits=c(length(ranks) + 1, -1), expand=c(0, 0)) +
            coord_flip() +
            NULL
    } else {
        p <- p + scale_x_continuous(limits=c(-1, length(ranks) + 1), expand=c(0, 0))
    }
    p <- p + theme(plot.margin = unit(c(0.5, 0.5, 0.5, 0.5), "cm"))
    enrichmentGrob <- ggplotGrob(p)

    if (addHeatmap && ncol(es) > 1) {
        # preparing heatmap
        mat <- exprs(es)[order(ranks, decreasing = TRUE), ]
        mat <- t(apply(mat, 1, scales::rescale))
        grouping <- ceiling(seq_len(nrow(mat)) / nrow(mat) * 1000)
        if (nrow(mat) <= length(unique(grouping))) {
            aggr <- mat
        } else {
            aggr <- colMeansByGroups(mat, grouping)
        }

        annotation_col <- NULL
        annotation_colors <- NULL
        if (!is.null(showAnnotation)) {
            values <- es[[showAnnotation]]
            annotation_col <- data.frame(row.names=colnames(es),
                                    setNames(list(factor(values,
                                                            levels=unique(values))),
                                             showAnnotation))

            if (!is.null(annotationColors)) {
                annotation_colors <- list()
                annotation_colors[[showAnnotation]] <- as.character(annotationColors)
                names(annotation_colors[[showAnnotation]]) <- names(annotationColors)
            }
        }


        # adding column to the left

        # 4 -> 1.5
        # 6 -> 2
        # 20+ -> 3

        heatmapWidth <- max(1, 3 - 6/ncol(aggr))

        if (vertical) {
            heatmapWidth <- min(heatmapWidth, width/4)
            ph <- pheatmap::pheatmap(aggr,
                                     cluster_rows = FALSE, cluster_cols = FALSE,
                                     show_rownames = FALSE, show_colnames = FALSE,
                                     color=fullPalette,
                                     annotation_col = annotation_col,
                                     annotation_colors = annotation_colors,
                                     legend = FALSE,
                                     silent = TRUE)

            heatmapGrobs <- ph$gtable
            hgMatrix <- rasterizeHeatmap(aggr,
                                         palette=fullPalette,
                                         maxDimensions=c(2500, 1000))

            panel_id <- enrichmentGrob$layout[enrichmentGrob$layout$name == "panel",c("t","l")]

            enrichmentGrob <- gtable_add_cols(enrichmentGrob, unit(heatmapWidth, "inch"), 0)
            enrichmentGrob <- gtable_add_grob(enrichmentGrob, hgMatrix,
                                              t = panel_id$t, l = 1, name="matrix")

            if (!is.null(showAnnotation)) {
                hgColAnnotation <- heatmapGrobs$grobs[[head(which(heatmapGrobs$layout$name == "col_annotation"), 1)]]
                hgAnnotationLegend <- heatmapGrobs$grobs[[head(which(heatmapGrobs$layout$name == "annotation_legend"), 1)]]
                hgLegend <- gtable_filter(heatmapGrobs, "annotation_legend", fixed=TRUE, trim=TRUE)

                enrichmentGrob <- gtable_add_grob(enrichmentGrob, hgColAnnotation,
                                                  t = 1, l = 1, b=panel_id$t-1, name="col_annotation")


                enrichmentGrob <- gtable_add_cols(enrichmentGrob, gtable_width(hgLegend), 0)
                enrichmentGrob <- gtable_add_grob(enrichmentGrob, hgAnnotationLegend,
                                                  t = panel_id$t, l = 1, name="annotation_legend")

            }

        } else {
            # horizontal
            heatmapWidth <- min(heatmapWidth, height/4)
            ph <- pheatmap::pheatmap(Matrix::t(aggr),
                                     cluster_rows = FALSE, cluster_cols = FALSE,
                                     show_rownames = FALSE, show_colnames = FALSE,
                                     color=fullPalette,
                                     annotation_row = annotation_col,
                                     annotation_colors = annotation_colors,
                                     legend = FALSE,
                                     silent = TRUE)

            heatmapGrobs <- ph$gtable
            hgMatrix <- rasterizeHeatmap(Matrix::t(aggr),
                                         palette=fullPalette,
                                         maxDimensions=c(1000, 2500))

            panel_id <- enrichmentGrob$layout[enrichmentGrob$layout$name == "panel",c("t","l")]


            enrichmentGrob <- gtable_add_rows(enrichmentGrob, unit(heatmapWidth, "inch"), 0)
            enrichmentGrob <- gtable_add_grob(enrichmentGrob, hgMatrix,
                                              t = 1, l=panel_id$l, name="matrix")

            if (!is.null(showAnnotation)) {
                hgColAnnotation <- heatmapGrobs$grobs[[head(which(heatmapGrobs$layout$name == "row_annotation"), 1)]]
                hgAnnotationLegend <- heatmapGrobs$grobs[[head(which(heatmapGrobs$layout$name == "annotation_legend"), 1)]]
                hgLegend <- gtable_filter(heatmapGrobs, "annotation_legend", fixed=TRUE, trim=TRUE)

                enrichmentGrob <- gtable_add_grob(enrichmentGrob, hgColAnnotation,
                                                  t = 1, l = 1, b=1, r=panel_id$l-1, name="col_annotation")


                enrichmentGrob <- gtable_add_cols(enrichmentGrob, gtable_width(hgLegend), 0)
                enrichmentGrob <- gtable_add_grob(enrichmentGrob, hgAnnotationLegend,
                                                  t = 1, l = 1, name="annotation_legend")

            }

        }





    }

    f <- tempfile(pattern = "enrichment", tmpdir = getwd(), fileext = ".svg")
    svg(f, width=width, height=height)
    grid.draw(enrichmentGrob)
    dev.off()
    # ggsave(p, filename = f, width=width, height=height)
    jsonlite::toJSON(f)
}

context("GSEA plot")
library(jsonlite)
library(Biobase)

test_that("gseaPlot works", {
    load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))

    set.seed(42)
    fData(es)$t <- rnorm(nrow(es))

    f <- fromJSON(gseaPlot(es, rankBy = "t", selectedGenes = sample.int(nrow(es), 10),
             width=6, height=4))
    expect_true(file.exists(f))

    f <- fromJSON(gseaPlot(es, rankBy = "t", selectedGenes = sample.int(nrow(es), 10),
             width=6, height=4, vertical = TRUE))
    expect_true(file.exists(f))

    f <- fromJSON(gseaPlot(es, rankBy = "t", selectedGenes = sample.int(nrow(es), 10),
             width=6, height=4, vertical = TRUE,
             addHeatmap = TRUE, showAnnotation = "time"))
    expect_true(file.exists(f))

    f <- fromJSON(gseaPlot(es, rankBy = "t", selectedGenes = sample.int(nrow(es), 10),
             width=6, height=4, vertical = FALSE,
             addHeatmap = TRUE, showAnnotation = "time"))
    expect_true(file.exists(f))

    f <- fromJSON(gseaPlot(es, rankBy = "t", selectedGenes = sample.int(nrow(es), 10),
                           width=6, height=4, vertical = FALSE,
                           addHeatmap = TRUE))
    expect_true(file.exists(f))
})

test_that("gseaPlot with heatmap works for single column", {
    load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))
    es <- es[, 1]
    set.seed(42)
    fData(es)$t <- rnorm(nrow(es))

    f <- fromJSON(gseaPlot(es, rankBy = "t", selectedGenes = sample.int(nrow(es), 10),
                           width=6, height=4, addHeatmap = TRUE))
    expect_true(file.exists(f))
})

sessionExists <- function(sessionName) {
    ocpuRoot <- strsplit(getwd(), 'ocpu-temp')[[1]][1]
    sessionPath <- file.path(ocpuRoot, 'ocpu-store', sessionName)

    savedPath <- file.path(sessionPath, 'sess.bin')
    RDataPath <- file.path(sessionPath, '.RData')

    if (file.exists(savedPath)) {
        return(jsonlite::toJSON(list(result = TRUE), auto_unbox = TRUE))
    }

    #-----------------LEGACY----------------------------------- DELETE AFTER 3 JUNE 2019

    if (file.exists(RDataPath)) {
        env <- new.env()
        load(RDataPath, envir = env)
        validForLoading = !is.null(env$es) && !is.null(attr(env$es, 'published'))

        return (jsonlite::toJSON(list(result=validForLoading), auto_unbox = TRUE))
    }

    return (jsonlite::toJSON(list(result=FALSE), auto_unbox = TRUE))
}

publishSession <- function (sessionName, datasetName = sessionName, heatmapJson = NULL) {
    ocpuRoot <- strsplit(getwd(), 'ocpu-temp')[[1]][1]
    sessionPath <- file.path(ocpuRoot, 'ocpu-store', sessionName)

    RDataPath <- paste(sessionPath, '.RData', sep=.Platform$file.sep)
    if (file.exists(RDataPath)) {
        env <- new.env()
        load(RDataPath, envir = env)

        if (!is.null(env$es)) { # this session is for actual dataset
            result <- list()
            datasetName <- ifelse(nchar(datasetName) > 0, datasetName, sessionName)
            result[[datasetName]] <- writeToList(env$es)
            binaryFile <- file.path(getwd(), 'sess.bin')

            writeBin(protolite::serialize_pb(list(layout_version = as.raw(PROTOBUF_LAYOUT_VERSION), ess = result)), binaryFile)
            assign("es", env$es, envir = parent.frame())
            if (!is.null(heatmapJson)) {
                heatmapPath <- file.path(getwd(), 'heatmap.json')
                write(jsonlite::toJSON(heatmapJson, auto_unbox = TRUE, null = 'null'), heatmapPath)
            }

            return (jsonlite::toJSON(basename(binaryFile)))
        }
    }

    stop('Invalid session')
}

loadSession <- function (sessionName) {
    ocpuRoot <- strsplit(getwd(), 'ocpu-temp')[[1]][1]
    sessionPath <- file.path(ocpuRoot, 'ocpu-store', sessionName)

    savedPath <- paste(sessionPath, 'sess.bin', sep=.Platform$file.sep)
    if (file.exists(savedPath)) {
        return (jsonlite::toJSON(basename(savedPath)))
    }

    #---------------------LEGACY------------------------------------- DELETE AFTER 3 JUNE 2019

    RDataPath <- paste(sessionPath, '.RData', sep=.Platform$file.sep)
    if (file.exists(RDataPath)) {
        savedEnv <- load(RDataPath)

        if (!("es" %in% savedEnv) || is.null(attr(es, 'published'))) {
            stop('Invalid session key')
        }

        result <- list()
        result[[sessionName]] <- writeToList(es)
        writeBin(protolite::serialize_pb(result), savedPath)
        return (jsonlite::toJSON(basename(savedPath)))
    }

    stop('Invalid session key')
}

heatmapSettings <- function (sessionName, isTempSession = TRUE) {

    if (isTempSession){
        ocpuRoot <- strsplit(getwd(), 'ocpu-temp')[[1]][1]
        sessionPath <- file.path(ocpuRoot, 'ocpu-store', sessionName)
        heatmapPath <- file.path(sessionPath, 'heatmap.json')
    }
    else{
        preloadedDir <- getOption("phantasusPreloadedDir")
        heatmapPath <- file.path(preloadedDir, paste0(sessionName,'.json'))
    }
    if (file.exists(heatmapPath)) {
        heatmap <- fromJSON(heatmapPath)
        return (jsonlite::toJSON(list(result=heatmap), auto_unbox = TRUE, null = 'null'))
    }

    return (jsonlite::toJSON(list(result=FALSE), auto_unbox = TRUE))
}

availableFGSEADatabases <- function () {
    dbDir <- getPhantasusConf("cache_folders")$fgsea_pathways
    if (is.null(dbDir)){
        stop("Current Phantasus configuration doesn't support FGSEA methods. 'fgsea_pathways' setting for cache folders is missed.")
    }
    if (! dir.exists(dbDir)){
        stop("Current Phantasus configuration doesn't support FGSEA methods. FGSEA pathways folder does not exist.")
    }
    metaFile <- file.path(dbDir, 'fgsea.txt')

    if (!file.exists(metaFile)){
        stop("FGSEA pathways folder was not properly configured. meta file is missed.")
    }
    meta = read.table(metaFile, header = TRUE, sep = '\t')

    return (jsonlite::toJSON(meta))
}

FGSEAmeta <- function (dbDir) {
    if (!dir.exists(dbDir)) {
        return()
    }

    metaFile <- file.path(dbDir, 'fgsea.txt')
    columnFiles <- list.files(dbDir, '*.rds', full.names = FALSE)
    message('Populating FGSEA pathway meta')
    if (file.exists(metaFile)) {
        meta = read.table(metaFile, header = TRUE, sep = '\t')
    } else {
        meta = data.frame(matrix(ncol=2,nrow=0), stringsAsFactors = FALSE)
    }
    colnames(meta) <- c('FILE', 'HINT')

    for(file in columnFiles) {
        if (nrow(meta[meta$FILE == file,]) == 0) {
            meta[nrow(meta) + 1, ] <- c(file, file)
        }
    }

    write.table(meta, file=metaFile, quote=FALSE, sep='\t', row.names = FALSE)
}

performFGSEA <- function (dbName, ranks) {
    dbDir <- getPhantasusConf("cache_folders")$fgsea_pathways
    dbFile <- file.path(dbDir, dbName)
    if (!file.exists(dbFile)) {
        stop('Invalid DB name supplied')
    }

    db <- readRDS(dbFile)
    pathways <- lapply(lapply(split(db, db$pathName), '[[', 'geneID'), as.character)
    rranks <- ranks$ranks
    names(rranks) <- ranks$genes

    assign('db', db, envir = parent.frame())

    return (jsonlite::toJSON(fgseaMultilevel(pathways,
                                             rranks,
                                             sampleSize = 101,
                                             eps = 1e-10,
                                             minSize = 15,
                                             maxSize = 500,
                                             nproc = 1)
                             , digits = NA
                             , na="string"))
}

queryPathway <- function (dbName, pathwayName) {
    dbDir <- getPhantasusConf("cache_folders")$fgsea_pathways
    dbFile <- file.path(dbDir, dbName)
    if (!file.exists(dbFile)) {
        stop('Invalid DB name supplied')
    }

    db <- readRDS(dbFile)
    filteredDb <- db[db$pathName == pathwayName,]

    jsonlite::toJSON(list(geneID = filteredDb$geneID, geneSymbol = filteredDb$geneSymbol))
}

#' Subsets es, if rows or columns are not specified, all are retained
#' @param es ExpressionSet object.#'
#' @param columns List of specified columns' indices (optional), indices start from 0#'
#' @param rows List of specified rows' indices (optional), indices start from 0
#' @return new expression set `es`
#'
#' @keywords internal
subsetES <- function(es, columns = c(), rows=c()) {
    rows <- getIndicesVector(rows, nrow(exprs(es)))
    columns <- getIndicesVector(columns, ncol(exprs(es)))

    es <- es[rows, columns]
    assign("es", es, envir = parent.frame())
}

renameColumn <- function (es, isFeature = TRUE, oldName, newName) {
    if (isFeature) {
        fData(es)[[newName]] <- fData(es)[[oldName]]
        fData(es)[[oldName]] <- NULL
    } else {
        phenoData(es)[[newName]] <- phenoData(es)[[oldName]]
        phenoData(es)[[oldName]] <- NULL
    }

    assign("es", es, envir = parent.frame())
}

#' Map indexes using Annotation DB
#'
#' \code{convertByAnnotationDB} function returns \code{keyType} ids from \code{dbName} mapped to \code{columnName} in \code{es}.
#'
#' @param es source ExpressionSet
#'
#' @param dbName name of AnnotationDB file
#'
#' @param columnName name of column in featureData of source ExpressionSet
#'
#' @param columnType Type of indexes in columnName
#'
#' @param keyType Type of mapped indexes
#'
#' @param otherOptions additional parameters for conversion. Currently only named boolean deleteDotVersion is not ignored.
#'
#' @return JSON object with a vector of converted IDs
#' @importFrom AnnotationDbi loadDb mapIds
#'
#' @keywords internal
convertByAnnotationDB <- function (es, dbName, columnName, columnType, keyType, otherOptions) {
    annotDir <- getPhantasusConf("cache_folders")$annot_db
    dbPath <- file.path(annotDir, dbName)
    if (!file.exists(dbPath)) {
        stop('Invalid database specified')
    }

    dbFile <- loadDb(dbPath)

    inputData <- fData(es)[[columnName]]
    if ("deleteDotVersion" %in% names(otherOptions)){
        if(otherOptions$deleteDotVersion){
            inputData <- sub(pattern = "\\.[0-9]+$", replacement = "", x = inputData)
        }
    }
    convertedData <- mapIds(dbFile,
                           keys = inputData,
                           keytype = columnType,
                           column = keyType,
                           multiVals = function (x) { paste0(x, collapse="///")})

    convertedData[ convertedData == 'NA' ] <- NA # AnnotationDB can produce 'NA' as string and <NA> as NA. Confusing
    fData(es)[[keyType]] <- convertedData
    assign("es", es, envir = parent.frame())

    return(jsonlite::toJSON(convertedData))
}

#' Get meta list for annotationDB files
#'
#' \code{queryAnnotationDBMeta} Function reads txt meta files for provided
#' sqlite annotation databases.
#'
#' @return meta info in JSON
#' @importFrom utils read.table
#' @examples
#' \dontrun{
#' queryAnnotationDBMeta()
#' }
#' @keywords internal
queryAnnotationDBMeta <- function () {
    annotDir <- getPhantasusConf("cache_folders")$annot_db
    if (is.null(annotDir)){
        stop("Current Phantasus configuration doesn't support annotation databases. 'annot_db' setting for cache folders is missed.")
    }
    if (!dir.exists(annotDir)){
        stop("Current Phantasus configuration doesn't support annotation databases. AnnotationDB folder does not exist.")
    }
    columnFiles <- list.files(annotDir, '.selected_fields.txt', full.names = TRUE)
    dbFiles <- list.files(annotDir, '*.sqlite$', full.names = TRUE)
    if (length(dbFiles) != length(columnFiles)){
        stop("AnnotationDB folder was not properly configured. Some meta files are missed.")
    }
    metaList <- list()
    for(columnFile in columnFiles) {
        columnsTSV <- read.table(file = columnFile, sep = '\t', header = TRUE, skip = 1)
        columns <- apply(columnsTSV, 1, paste, collapse = " - ")

        orgName = unlist(strsplit(basename(columnFile), ".selected_fields.txt"))[1]

        metaList[[orgName]] = list(species = readLines(columnFile, n = 1), columns = columns)
    }

    return(jsonlite::toJSON(metaList))
}

#' Create meta file for AnnotationDB
#'
#' \code{annotationDBMeta} function creates txt files containing meta information of
#' provided sqlite files for AnnotationDB.
#'
#' @param annotDir path to folder with annotationDB sqlite files
#'
#' @return nothing
#'
#' @importFrom AnnotationDbi keys species columns loadDb
#'
#' @examples
#' \dontrun{
#' annotationDBMeta('/var/phantasus/cache')
#' }
#' @keywords internal
annotationDBMeta <- function (annotDir) {
    if (!dir.exists(annotDir)) {
        return()
    }

    message('Populating AnnotationDB meta')
    dbFiles <- list.files(annotDir, '*.sqlite$', full.names = TRUE)
    for (dbFile in dbFiles) {
        db <- loadDb(dbFile)
        columnFile <- paste(dbFile, ".selected_fields.txt", sep = "")
        if (!file.exists(columnFile)) {
            columnsDB <- columns(db)
            columns <- sapply(columnsDB, function (column) {
                hint <- paste(head(keys(db, keytype=column), n = 2), collapse=";")
                return(paste(column, hint, sep=" - "))
            })

            humanMeta <- t(data.frame(strsplit(columns, " - ", fixed=TRUE)))
            cat(AnnotationDbi::species(db), "\n", file=columnFile)
            suppressWarnings(write.table(humanMeta, file=columnFile, quote=FALSE, sep='\t',  row.names=FALSE, col.names=c('FIELD', 'HINT'), append=TRUE))
        }
    }
}

context("Convert by AnnotationDb")
library(Biobase)
library(data.table)
library(phantasusLite)

# Write a temp config with annot_db as a literal path so that
# config::get() can resolve it without the devtools system.file shim.
local({
    annotDir <- system.file("testdata/annotationdb/", package = "phantasus")
    cfgDir <- file.path(tempdir(), "phantasus-annottest-config", "R", "phantasus")
    dir.create(cfgDir, recursive = TRUE, showWarnings = FALSE)
    writeLines(paste0(
        "default:\n",
        "  host: \"0.0.0.0\"\n",
        "  preloaded_dir: NULL\n",
        "  static_root: \"",
            system.file("www/phantasus.js", package = "phantasus"), "\"\n",
        "  cache_root: \"", tempdir(), "\"\n",
        "  cache_folders:\n",
        "    geo_path: \"", file.path(tempdir(), "geo"), "\"\n",
        "    annot_db: \"", annotDir, "\"\n",
        "    fgsea_pathways: \"", file.path(tempdir(), "fgsea"), "\"\n",
        "    rnaseq_counts: \"", file.path(tempdir(), "counts"), "\"\n"
    ), file.path(cfgDir, "user.conf"))
    Sys.setenv(R_USER_CONFIG_DIR =
                   file.path(tempdir(), "phantasus-annottest-config"))
    Sys.setenv(R_CONFIG_ACTIVE = "default")
})
test_that("AnnotationbyDb works with delete version", {
    test_file <- system.file("testdata/counts_versioned_ids.gct", package="phantasus")
    if (test_file == ""){
        stop("test counts file doesn't exists")
    }
    es <- readGct(test_file)
    dbName <-"sample_mouse_db.sqlite"
    columnName <- "id"
    columnType <- "ENSEMBL"
    keyType <- "SYMBOL"
    otherOptions <- list(deleteDotVersion = FALSE)
    expect_error(convertByAnnotationDB(es, dbName, columnName, columnType, keyType, otherOptions))
    otherOptions <- list(deleteDotVersion = TRUE)
    symbols <- jsonlite::fromJSON(convertByAnnotationDB(es, dbName, columnName, columnType, keyType, otherOptions))
    expect_gt(sum(!is.na(symbols)),0)

})

#' Create calculated annotation
#'
#' \code{calculatedAnnotation} adds a column calculated by operation
#'
#' @param es ExpressionSet object.
#'
#' @param operation Name of the operation to perform calculation
#'
#' @param columns List of specified columns' indices (optional),
#' indices start from 0#'
#'
#' @param rows List of specified rows' indices (optional), indices start from 0
#'
#' @param name Name of the new annotation
#'
#' @param isColumns Apply fn to columns
#'
#' @return Nothing. Annotated dataset will be assigned to es in environment
#'
#' @import Biobase
#' @keywords internal
calculatedAnnotation <- function (es, operation, rows = c(),
                                  columns = c(), isColumns = FALSE, name = NULL) {
    rows <- getIndicesVector(rows, nrow(exprs(es)))
    columns <- getIndicesVector(columns, ncol(exprs(es)))

    fn <- tolower(operation)
    if (!isColumns) {
        fData(es)[[name]] <- NA
        fData(es)[[name]][rows] <- apply(exprs(es[rows,columns]), 1, fn)
    } else {
        phenoData(es)[[name]] <- NA
        phenoData(es)[[name]][columns] <- apply(exprs(es[rows,columns]), 2, fn)
    }

    assign("es", es, envir = parent.frame())

}

#' Setup phantasus.
#' Read user config file ( or create default one) and fill \code{cache_root} using sources in \code{file}.
#' @param setup_name  name of config from \code{file}. If unset or not existed, "default".
#' @param file Location of the setup.yml file with setup parameters. If not existed use file from package
#' @export
setupPhantasus <- function(
        setup_name = "default",
        file = confFile("setup.yml")){
    setup_config <- getSetupConf(file, setup_name)
    user_conf_file <- confFile("user.conf")
    cache_root = ""
    if (!file.exists(user_conf_file)){
        default_cache <- tools::R_user_dir("phantasus", which = "cache")
        if (interactive()){
            message("Specify the cache root for Phantasus. \n This folder will contain all Phantasus-related files: \n - GEO cache files \n - Annotation databases \n - FGSEA pathways \n - RNA-seq counts")
            cache_root <- readline(paste0("Enter path or leave blank for default (", default_cache, "):"))
        }
        if (is.null(cache_root) || is.na(cache_root) || (nchar(cache_root)==0)){
            cache_root <- default_cache
        }
        user_conf <- get_user_conf(cache_root = cache_root, setup_config)
        dir.create(dirname(user_conf_file), showWarnings = FALSE, recursive = TRUE)
        message(paste("Create user configuration file:", user_conf_file))
        cat(user_conf_to_yaml(user_conf), file = user_conf_file)

    } else{
        message(paste("Use existed configuration file:", user_conf_file))
    }
    user_conf <- getPhantasusConf()
    Sys.setenv(R_CONFIG_ACTIVE = "default")

    user_conf <- getPhantasusConf()
    message(paste("Current configuration: ", Sys.getenv("R_CONFIG_ACTIVE")))
    default_timeout <- getOption("timeout")
    options(timeout = max(1000, default_timeout))

    configureGEOcache(user_conf = user_conf )
    configureAnnotDB(user_conf = user_conf, setup_config = setup_config )
    configureFGSEA(user_conf = user_conf, setup_config = setup_config )
    configureRnaseqCounts(user_conf = user_conf, setup_config = setup_config )
    if (!opencpu:::win_or_mac() && ! "unix" %in% utils::installed.packages()) {
        if (menu(c("Yes", "No"),
                 title= paste("Couldn't find the required `unix` package, do you want to install it?")) == "1") {
            install.packages("unix")
        }
    }

    options(timeout =  default_timeout)
    if (file.exists(user_conf_file)){
        message(paste("! Successfully setup configuration file:", user_conf_file, "!"))
    }
}

confFile <- function(name){
    return(file.path(tools::R_user_dir(package = "phantasus", which = "config"), name ))

}

getSetupConf <- function(file, setup_name){
    if (!file.exists(file)){
        message("Use default setup parameters")
        file <- system.file("configs/R/phantasus/setup.yml", package = "phantasus")
    }
    setup_config <-  config::get(   config = setup_name,
                                    file = file,
                                    use_parent = FALSE
    )
    return(setup_config)
}

#' Read Phantasus Config
#'
#' @param value Value to retrieve from the config file.
#' @param configName R_CONFIG_ACTIVE value. If unset, "default".
#' @param file Location of the config file
#' @export
getPhantasusConf <-  function(
        value = NULL,
        configName = Sys.getenv("R_CONFIG_ACTIVE"),
        # Modify this if your config file is somewhere else
        file = file.path(tools::R_user_dir(package = "phantasus", which = "config"), "user.conf")
) {
    if (!file.exists(file)){
        stopPhantasus()
    }
    config::get(
        value = value,
        config = configName,
        file = file,
        use_parent = FALSE
    )

}

configureGEOcache <- function(user_conf){
    message("Configure GEO cache folder...")
    geoPath <- user_conf$cache_folders$geo_path
    if (dir.exists(geoPath) && rw_dir_check(geoPath)){
        message("! GEO cache folder exists and is treated as already configured !")
        return()
    }
    unsafe_dir_create(geoPath)
}


configureAnnotDB <- function(user_conf, setup_config){
    message("Configure AnnotationDb...")
    local_path <- user_conf$cache_folders$annot_db

    if (dir.exists(local_path) && rw_dir_check(local_path)){
        message("! AnnotationDB folder exists and is treated as already configured !")
        return()
    }
    menu_choices = c()
    actions <- c()
    dbFiles <- list.files(local_path, recursive = FALSE)
    mm_pkg <- system.file(package='org.Mm.eg.db')
    hs_pkg <- system.file(package='org.Hs.eg.db')

    if (nchar(mm_pkg) || nchar(hs_pkg)){
        message("Found local annotation databases:")
        message(paste ("\t" , mm_pkg))
        message(paste ("\t" , hs_pkg))
        menu_choices = c(menu_choices, "Use local databases listed above")
        actions = c(actions, function(){
            if (nchar(mm_pkg)){
                mm_res <- file.copy(org.Mm.eg.db::org.Mm.eg_dbfile(), file.path(local_path, "org.Mm.eg.sqlite"))
                if (mm_res) {
                    message(paste("! Use copy of :", org.Mm.eg.db::org.Mm.eg_dbfile()))
                }
            }
            if (nchar(hs_pkg)){
                hs_res <- file.copy(org.Hs.eg.db::org.Hs.eg_dbfile(), file.path(local_path, "org.Hs.eg.sqlite"))
                if (hs_res) {
                    message(paste("! Use copy of :", org.Hs.eg.db::org.Hs.eg_dbfile()))
                }
            }
        })
    }
    if (!length(setup_config$annot_db) == 0){
        file_index <- getFileIndexDF(base_url = setup_config$annot_db, pattern = "txt|sqlite" )
        total_size <- round(sum(as.numeric(fs::fs_bytes(file_index$size))) / 2^20, digits = 0)
        menu_choices <- c(menu_choices,  paste("Download files from the ctlab Phantasus mirror (~", total_size, "MB)"))
        actions <- c(actions, function(){
            loadAllFiles(url = setup_config$annot_db,
                         file_df = file_index,
                         destdir = local_path)
        } )
    }
    menu_choices = c(menu_choices, "Keep empty")
    actions <- c(actions, function() {message("Kept empty")})
    if ((!interactive()) || length(menu_choices) == 1){
        menu_res <- 1
    } else{
        menu_res <- utils::menu(choices = menu_choices, graphics = FALSE, title = "Choose how to set up the AnnotationDB folder: ")
        if (menu_res == 0){
            message("Canceled")
            return()
        }
    }
    unsafe_dir_create(local_path)
    actions[[menu_res]]()
    annotationDBMeta(getPhantasusConf("cache_folders")$annot_db)
}


configureFGSEA <- function(user_conf, setup_config){
    message("Configure FGSEA pathways...")
    local_path <- user_conf$cache_folders$fgsea_pathways
    if (dir.exists(local_path) && rw_dir_check(local_path)){
        message("! FGSEA folder exists and is treated as already configured !")
        return()
    }

    menu_choices = c()
    actions <- c()
    if (length(setup_config$fgsea_pathways) != 0){
        message("External source for FGSEA pathways is scpecified.")
        file_index <- getFileIndexDF(base_url = setup_config$fgsea_pathways, pattern = "txt|rds" )
        total_size <- round(sum(as.numeric(fs::fs_bytes(file_index$size))) / 2^20, digits = 2)
        menu_choices <- c(menu_choices, paste0("Download files from the ctlab Phantasus mirror (~", total_size, "MB)"))

        actions <- c(actions, function(){
            loadAllFiles(url = setup_config$fgsea_pathways,
                         destdir = local_path,
                         file_df = file_index)
        } )
    }
    menu_choices = c(menu_choices,"Keep empty")
    actions <- c(actions, function() {message("Kept empty")})
    if ((!interactive()) || length(menu_choices) == 1){
        menu_res <- 1
    } else{
        menu_res <- utils::menu(choices = menu_choices, graphics = FALSE, title = "Choose how to set up the FGSEA pathways folder: ")
        if (menu_res == 0){
            message("Canceled")
            return()
        }
    }
    unsafe_dir_create(local_path)
    actions[[menu_res]]()
    FGSEAmeta(getPhantasusConf("cache_folders")$fgsea_pathways)
}

configureRnaseqCounts <- function(user_conf, setup_config){
    message("Configure RNA-seq counts...")
    selected_path <- user_conf$cache_folders$rnaseq_counts
    options(PhantasusUseHSD = NULL)
    if (is.null(selected_path) || is.na(selected_path) ||  (nchar(selected_path) == 0)){
        stop("!Configuration failed: RNA-seq counts path should be specified !")
        return()
    }
    if (dir.exists(selected_path) && rw_dir_check(selected_path)){
        message("! RNA-seq counts folder exists. Update meta... !")
        updateCountsMeta(counts_dir =  selected_path, force = FALSE, verbose = FALSE)
        message("! RNA-seq counts configured !")
        return()
    }
    hsds <- isHSDS(selected_path)
    if (is.null(hsds)){
        message("! RNA-seq count path looks like local path !")
        unsafe_dir_create(selected_path)
        message("RNA-seq counts configured")
        return()
    }
    message("! RNA-seq counts path looks like web URL !")
    if( !isHSDS(selected_path)){
        message("Resource doesn't respond as HSDS server")
        stop("Configuration failed: RNA-seq count path should be local diractory or HSDS server")
    }
    message(paste(selected_path, "response as HSDS server"))
    phantasus_lite <- system.file(package='phantasusLite')
    if (nchar(phantasus_lite) == 0){
        message("phantasus-lite package is not installed. HSDS will be ignored")
        stop("Configuration failed: Install phantasusLite package to use HSDS server")
    }
    options(PhantasusUseHSDS = TRUE)
    message("HSDS server will be used as source of RNA-seq counts")
}

getFileIndexDF <- function(base_url, pattern, location = "/"){
    req <- httr::GET(paste0(base_url, location))
    text <- httr::content(req, "text", encoding = "UTF-8")
    index_df <-  data.table(na.omit(XML::readHTMLTable(text, skip = c(1,2), as.data.frame = TRUE, trim = TRUE )[[1]]))
    file_df <- index_df[grepl(pattern = pattern, x = index_df$Name) & index_df$Size != "-", .(file = paste0(location, Name), date = `Last modified`, size = Size)]

    for (dir in index_df[index_df$Size == "-",]$Name){
        local_dir <- paste0(location, dir)
        file_df <- rbind(file_df, getFileIndexDF(base_url, pattern, local_dir ))
    }
    return(file_df)
}

loadAllFiles <- function(url, file_df, destdir ){
    message(paste("Trying download from", url))
    url <- trimws(url, which = "right", whitespace = "/")
    for (file in file_df$file) {
        target_file <- file.path(destdir,file)
        dir.create(dirname(target_file), recursive = TRUE, showWarnings = FALSE)
        safeDownload(
            url = paste0(url, file),
            filename = basename(target_file),
            dir = dirname(target_file),
            quiet = FALSE
        )
    }
}


get_user_conf <- function( cache_root, setup_config){

    cache_folders = list ()
    cache_folders$geo_path <- "file.path(cache_root, 'geo/')"
    attr(cache_folders$geo_path, "R") = TRUE
    cache_folders$annot_db <- "file.path(cache_root, 'annotationdb/')"
    attr(cache_folders$annot_db, "R") = TRUE
    cache_folders$fgsea_pathways <- "file.path(cache_root, 'fgsea_pathways/')"
    attr(cache_folders$fgsea_pathways, "R") = TRUE

    if ( is.null(setup_config$rnaseq_counts) || length(setup_config$rnaseq_counts) == 0){
        cache_folders$rnaseq_counts <- "file.path(cache_root, 'counts/')"
        attr(cache_folders$rnaseq_counts, "R") <- TRUE
    } else{
        cache_folders$rnaseq_counts <- setup_config$rnaseq_counts
        attr(cache_folders$rnaseq_counts, "R") <- FALSE
    }
    static_root <- "system.file('www/phantasus.js', package = 'phantasus')"
    attr(static_root, "R") <- TRUE
    geo_mirrors <- if (length(setup_config$geo_mirrors) == 0){
                       list(true_geo = "https://ftp.ncbi.nlm.nih.gov")
                    } else {
                        setup_config$geo_mirrors
                    }

    return(
          list(
                host = "0.0.0.0",
                port = "8000",
                preloaded_dir = NULL,
                static_root = static_root,
                cache_root = cache_root,
                cache_folders = cache_folders,
                geo_mirrors = geo_mirrors
            )
    )

}


user_conf_to_yaml = function(user_config){
    for (i in seq_along(user_config$cache_folders)){
        if (!is.null(attr(user_config$cache_folders[[i]], "R")) && attr(user_config$cache_folders[[i]], "R") ){
            attr(user_config$cache_folders[[i]], "tag") <- "!expr"
        }
    }
    if (!is.null(attr(user_config$static_root, "R")) && attr(user_config$static_root, "R") ){
        attr(user_config$static_root, "tag") <- "!expr"
    }

    for (i in seq_along(user_config$geo_mirrors)) {
        attr(user_config$geo_mirrors[[i]], "quoted") <- TRUE
    }
    attr(user_config$geo_mirrors, "quoted") <- TRUE
    attr(user_config$cache_root, "quoted") <- TRUE
    attr(user_config$host,"quoted") <- TRUE
    class(user_config$port) = 'verbatim'
    user_config <- list(default = user_config)
    yaml::as.yaml(user_config, handlers = list(
        logical = function(x){
            result <- ifelse(x, "TRUE", "FALSE")
            class(result) <- "verbatim"
            return(result)
        },
        "NULL" = function(x){
            result = "NULL"
            class(result) <- "verbatim"
            return(result)
        }
    ) )
}

#' Creates default docker conf file
#' Function creates default docker user configuration file based on provided \code{setup_file}
#' or on default parameters if \code{setup_file} doesn't exist. If \code{user_conf_file} exists function does nothing.
#' @param setup_file  name of config from \code{file}. If unset or not existed, "default".
#' @param user_conf_file Location of the setup.yml file with setup parameters. If not existed use file from package
#' @keywords internal
createDockerConf <- function( setup_file = confFile("setup.yml"), user_conf_file = confFile("user.conf")){
    if (!file.exists(user_conf_file)){
        setup_conf <-  getSetupConf(setup_file, "default")
        user_conf <- get_user_conf(cache_root = "/var/phantasus/cache", setup_config = setup_conf)
        user_conf$preloaded_dir <- "/var/phantasus/preloaded"
        dir.create(dirname(user_conf_file), showWarnings = FALSE, recursive = TRUE)
        message(paste("Create user configuration file:", user_conf_file))
        cat(user_conf_to_yaml(user_conf), file = user_conf_file)
    }
}

unsafe_dir_create <- function(dir_name,  recursive = TRUE){
    message(paste("try to create:", dir_name))
    configured <- dir.create(dir_name, recursive = recursive, showWarnings = FALSE)
    if (!configured){
        stop(paste("Configuration failed: can't create", dir_name))
    }
    rw_dir_check(dir_name)
}

#' Example dataset
#'
#' Small slice from GSE27112-GPL6103 for
#'     runnable examples.
#'
#' @docType data
#'
#' @usage data(es)
#'
#' @examples
#' \dontrun{
#' data(es)
#' performKmeans(es, k = 2)
#' }
#'
"es"

#' Load GEO Dataset.
#'
#' \code{loadGEO} returns the file with serialized ExpressionSet using
#'     ProtoBuf, parsed from data downloaded from GEO by identifier.
#'
#' @param name String, containing GEO identifier of the dataset.
#'     It should start with 'GSE' or 'GDS' and can include exact GPL
#'     to annotate dataset, separated with dash ('-') from the identifier.
#'
#' @param type Type of the dataset: 'GSE' or 'GDS'. If not specified,
#'     the function will take first three letters
#'     of \code{name} variable as type.
#'
#' @return File with ProtoBuf-serialized ExpressionSet-s
#'     that were downloaded by this identifier.
#'     For GSE-datasets there can be multiple annotations, so in file will be a
#'     list mapping name with GPL to ExpressionSet.
#'
#' @examples
#' \dontrun{
#'     loadGEO("GSE27112")
#'     loadGEO("GDS4922")
#' }
#'
#' @importFrom utils tail
#' @import Biobase
#' @import GEOquery
#' @keywords internal
loadGEO <- function(name, type = NA) {
    cacheDir <- getPhantasusConf("cache_folders")$geo_path

    if (!dir.exists(cacheDir)) {
        dir.create(cacheDir)
    }

    mirrorPath <- getPhantasusConf("geo_mirrors")


    geoDir <- getGEODir(name, cacheDir)
    binaryName <- paste0(name, '.bin')
    filePath <- file.path(geoDir, binaryName)
    urls <- c()


    ess <- getES(name, type, destdir = cacheDir, mirrorPath = mirrorPath)

    files <- list()
    assign("es", utils::tail(ess, n = 1)[[1]], envir = parent.frame())

    current_rda <- file.path(geoDir, paste0(name, ".rda"))
    bin_is_valid <- checkBinValidity(filePath, file.info(current_rda)$ctime)
    if (!bin_is_valid) {
        for (i in seq_along(ess)) {
            seriesName <- if (!grepl(pattern = "-", name) && length(ess) > 1)
                paste0(name, "-", annotation(ess[[i]])) else name
            files[[seriesName]] <- writeToList(ess[[i]])
        }
        tempBinFile <- tempfile(paste0(binaryName, ".binsave"), tmpdir = geoDir)
        protolite::serialize_pb(list(layout_version = as.raw(PROTOBUF_LAYOUT_VERSION), ess = files), tempBinFile)
        message('Saved binary file: ', tempBinFile)
        file.rename(tempBinFile, filePath)
    }


    urls <- c(urls, paste0("/geo/",unlist(strsplit(filePath, cacheDir, fixed = TRUE))[2]))
    jsonlite::toJSON(urls)
}

#' Load ExpressionSet from GEO Datasets
#'
#'\code{getGDS} return the ExpressionSet object corresponding
#'     to GEO Dataset identifier.
#'
#' @param name String, containing GEO identifier of the dataset.
#'     It should start with 'GSE' or 'GDS' and can include exact GPL
#'     to annotate dataset, separated with dash ('-') from the identifier.
#'
#' @param destdir Directory for caching loaded Series and GPL
#'     files from GEO database.
#'
#' @param mirrorPath URL string which specifies the source of matrices.
#'
#' @return ExpressionSet object wrapped in list, that was available by given
#'     in \code{name} variable GEO identifier.
#'
#' @examples
#' \dontrun{
#'    getGDS('GDS4922', destdir = tempdir(), mirrorPath = "https://ftp.ncbi.nlm.nih.gov")
#' }
#' @export
getGDS <- function(name, destdir = getPhantasusConf("cache_folders")$geo_path,
                   mirrorPath = getPhantasusConf("geo_mirrors")) {
    stub <- gsub("\\d{1,3}$", "nnn", name, perl = TRUE)
    filename <- sprintf("%s.soft.gz", name)
    gdsurl <- "%s/geo/datasets/%s/%s/soft/%s"

    fullGEODirPath <- getGEODir(name, destdir)
    destfile <- file.path(fullGEODirPath, filename)

    infile <- file.exists(destfile)

    # Several GEO mirrors supported, but use only one for now.
    mirrorPath <- head(mirrorPath,1)

    if (!infile) {
        tempDestFile <- tempfile(paste0(filename, ".load"), tmpdir=fullGEODirPath)
        for (url in sprintf(gdsurl, mirrorPath,
                            stub, name, filename)){
            tryCatch({
                utils::download.file(url,
                                     destfile = tempDestFile,
                                     method="libcurl")
                file.rename(tempDestFile, destfile)
                infile <- TRUE
            },
            error = function(e) {
                file.remove(tempDestFile)
            },
            warning = function(w) {
                file.remove(tempDestFile)
            })
            if (file.exists(destfile)){
                break
            }
        }

    } else {
        message(paste("Loading from locally found file", destfile))
    }

    l <- suppressWarnings(getGEO(GEO = name, destdir = fullGEODirPath, AnnotGPL = TRUE))

    # extracting all useful information on dataset
    table <- methods::slot(l, "dataTable")


    # extracting table ID_REF | IDENTIFIER/SAMPLE | SAMPLE1 | ...
    data <- Table(table)
    columnsMeta <- Columns(table)  # phenoData
    sampleNames <- as.vector(columnsMeta[["sample"]])
    rownames <- as.vector(data[["ID_REF"]])
    symbol <- as.vector(data[["IDENTIFIER"]])

    data <- data[, sampleNames]  # expression data
    exprs <- as.matrix(data)
    row.names(exprs) <- rownames

    row.names(columnsMeta) <- sampleNames
    pData <- AnnotatedDataFrame(data.frame(columnsMeta, check.names = FALSE))

    fData <- data.frame(id=rownames, symbol=symbol, row.names = rownames, stringsAsFactors = FALSE)
    fData <- AnnotatedDataFrame(fData)

    mabstract=ifelse(is.null(Meta(l)$description),"",Meta(l)$description)
    mpubmedids=ifelse(is.null(Meta(l)$pubmed_id),"",Meta(l)$pubmed_id)
    mtitle=ifelse(is.null(Meta(l)$title),"",Meta(l)$title)

    list(ExpressionSet(assayData = exprs,
                       phenoData = pData,
                       featureData = fData,
                       experimentData=new("MIAME",
                                          abstract=mabstract,
                                          title=mtitle,
                                          pubMedIds=mpubmedids,
                                          other=Meta(l))))
}


#' Returns list of ARCHS4 hdf5 files with expression data
#' @param cacheDir base directory for cache
#' @return list of .h5 files
#' @keywords internal
getArchs4Files <- function(cacheDir) {
    list.files(paste(file.path(cacheDir), 'archs4', sep = .Platform$file.sep), '\\.h5$', full.names = TRUE)
}


#' Loads expression data from ARCHS4 count files.
#' Only sapmles with counted expression are kept.
#' If es already containts expression data it is returned as is.
#' @param es ExpressionSet from GEO to check for expression in ARCHS4
#' @param archs4_files list of available .h5 files from ARCHS4 project
#' @return either original es or an ExpressionSet with loaded count data from ARCHS4
#' @keywords internal
loadFromARCHS4 <- function(es, archs4_files) {

    if (nrow(es) > 0 ) {
        return(es)
    }
    for (destfile in archs4_files) {

        samples <- h5read(destfile, "meta/Sample_geo_accession")
        sampleIndexes <- match(es$geo_accession,
                               samples)

        if (sum(!is.na(sampleIndexes)) == 0) {
            # no needed samples in this file
            H5close()
            next
        }
           genes <- as.character(h5read(destfile, "meta/genes"))
        h5Indexes = list(seq_len(length(genes)),
                         stats::na.omit(sampleIndexes))

        expression <- h5read(destfile,
                             "data/expression",
                             index=h5Indexes)
        rownames(expression) <- genes
        colnames(expression) <- colnames(es)[!is.na(sampleIndexes)]

        es2 <- ExpressionSet(assayData = expression,
                             phenoData = phenoData(es[, !is.na(sampleIndexes)]),
                             annotation = annotation(es),
                             experimentData = experimentData(es)
                            )

        tryCatch({

          fData(es2) <- cbind(fData(es2), "ENSEMBLID"=as.character(h5read(destfile, "meta/gene_ensemblid")))
          }, error=function (e) {})

        tryCatch({
          fData(es2) <- cbind(fData(es2), "Gene ID"=as.character(h5read(destfile, "meta/gene_entrezid")))
        }, error=function (e) {})

        fData(es2) <- cbind(fData(es2), "Gene symbol"=rownames(es2))
        H5close()
        return(es2)
    }

    return(es)
}

#' Loads expression data from .h5 count files.
#' Only samples with counted expression are kept.
#' If es already containts expression data it is returned as is.
#' @param es ExpressionSet from GEO to check for expression in ARCHS4/dee2 or other h5 files
#' @param counts_dir directory with  .h5 files  collections. There must be meta.rda file
#' in counts_dir and each collection's sub directory must have meta.txt file with description.
#' Also \code{counts_dir} must contain \code{counts_priority.txt} file.
#' @return either original es or an ExpressionSet with loaded count data from ARCHS4
#' @import data.table
#' @keywords internal
loadCounts <- function(es, counts_dir) {
    if (!file.exists(file.path(counts_dir, "counts_priority.txt"))) {
      return(es)
    }
    priority <- fread(file.path(counts_dir, "counts_priority.txt"))[, .(directory), keyby = priority]$directory
    if (nrow(es) > 0) {
        return(es)
    }
    load(paste(counts_dir,"meta.rda",sep = "/"))
    sample_amount <- DT_counts_meta[accession %in% es$geo_accession,
                                   .(.N),
                                   by = list(file, type_fac = factor(x = collection_type, levels = priority))]
    if (nrow(sample_amount) == 0) {
        return(es)
    }
    setorderv(x = sample_amount,cols = c("N","type_fac"),order = c(-1,1))
    destfile <- sample_amount[,.SD[1]]$file
    destfile <- file.path(counts_dir, destfile)
    h5_meta <- fread(file.path(dirname(destfile), "meta.txt"), index = "file_name")[file_name == basename(destfile)]
    h5f <- H5Fopen(destfile, flags = "H5F_ACC_RDONLY")
    samples <- h5read(h5f, h5_meta$sample_id)
    sampleIndexes <- match(es$geo_accession,
                           samples)

    gene_id <- strsplit(h5_meta$gene_id, split = ":")[[1]]
    genes <- as.character(h5read(h5f,gene_id[2]))
    h5Indexes = list(stats::na.omit(sampleIndexes),
                     seq_len(length(genes)))
    expression <- NULL
    if (h5_meta$sample_dim == "rows"){
        expression <- h5read(h5f,
                             "data/expression",
                             index = h5Indexes)
        expression <- t(expression)
    } else {
        expression <- h5read(h5f,
                             "data/expression",
                             index = rev(h5Indexes))
    }
    rownames(expression) <- genes
    colnames(expression) <- colnames(es)[!is.na(sampleIndexes)]
    es2 <- ExpressionSet(assayData = expression,
                         phenoData = phenoData(es[, !is.na(sampleIndexes)]),
                         annotation = annotation(es),
                         experimentData = experimentData(es)
     )
    genes_annot <- strsplit(h5_meta$genes_annot, split = ";")[[1]]
    genes_annot <- unlist( lapply(strsplit(genes_annot, split = ":"), function(annot){
        setNames(annot[2], annot[1])
    }))

    genes_annot <- lapply(genes_annot, function(annot){
      tryCatch({
          as.character(h5read(h5f, annot))
      }, error = function(e) {})
    })
    genes_annot [[gene_id[1]]] <- rownames(es2)
    genes_annot <- genes_annot[!unlist(lapply(genes_annot, is.null))]
    fData(es2) <- cbind(fData(es2), genes_annot )
    H5Fclose(h5f)
    return(es2)
}

filterFeatureAnnotations <- function(es) {
    fvarsToKeep <- c()
    if ("Gene symbol" %in% fvarLabels(es)) {
        fvarsToKeep <- c(fvarsToKeep, "Gene symbol")
    } else {
        fvarsToKeep <- c(fvarsToKeep, grep("symbol",
                                           fvarLabels(es),
                                           ignore.case = TRUE,
                                           value = TRUE))
    }

    if ("Gene ID" %in% fvarLabels(es)) {
        fvarsToKeep <- c(fvarsToKeep, "Gene ID")
    } else if ("ENTREZ_GENE_ID" %in%  fvarLabels(es)){
        fvarsToKeep <- c(fvarsToKeep, "ENTREZ_GENE_ID")
    }else if ("ID" %in% fvarLabels(es)) {
        fvarsToKeep <- c(fvarsToKeep, "ID")
    } else {
        fvarsToKeep <- c(fvarsToKeep, grep("entrez",
                                           fvarLabels(es),
                                           ignore.case = TRUE,
                                           value = TRUE))
    }

    if (length(setdiff(fvarsToKeep, "ID")) == 0) {
        fvarsToKeep <- c(fvarsToKeep,
                         setdiff(colnames(featureData(es)), fvarsToKeep))
    }


    featureData(es) <- featureData(es)[, fvarsToKeep]

    if (!any(sapply(fData(es),
                    function(x) identical(rownames(es), as.character(x))
                    ))) {
        fData(es) <- cbind("id"=rownames(es), fData(es))
    }

    es
}
#' @importFrom tidyr spread
filterPhenoAnnotations <- function(es) {
    phenoData(es) <- phenoData(es)[,
                                   grepl("characteristics",
                                         varLabels(es),
                                         ignore.case = TRUE) |
                                       (varLabels(es) %in% c("title",
                                                             "id",
                                                             "geo_accession"
                                       ))]

    chr <- varLabels(es)[grepl("characteristics",
                               varLabels(es),
                               ignore.case = TRUE)]

    parsePData <- function(old.phenodata) {
        old.pdata <- pData(old.phenodata)
        labels <- varLabels(old.phenodata)

        new.pdata <- as.data.frame(matrix(NA, nrow = nrow(old.pdata), ncol = 0))
        rownames(new.pdata) <- rownames(old.pdata)
        sample_key_value = as.data.table(matrix(NA, nrow = 0, ncol =4))
        colnames(sample_key_value) <-c("sample","key","value", "channel")
        for (i in seq_len(ncol(old.pdata))) {
            channel <- sub(pattern = "characteristics_([^.]+)(\\..+)?",replacement = "\\1",x =  names(old.pdata)[i])
            splitted <- strsplit(as.vector(old.pdata[[i]]), ':')
            lengths <- sapply(splitted, length)
            if (any(lengths != 2 & lengths != 0)) {
                new.pdata[[labels[i]]] <- old.pdata[[i]]
            } else {
                zeros <- lengths == 0
                keys <- trimws(take(splitted[!zeros], 1))
                values <- trimws(take(splitted[!zeros], 2))
                samples <- rownames(old.pdata)[!zeros]
                sample_key_value <- rbind(sample_key_value, data.frame(sample = samples, key = keys, value = values, channel = channel))
            }
        }
        if (length(unique(sample_key_value$channel)) >1){
            sample_key_value$key <- paste(sample_key_value$key, sample_key_value$channel, sep = "_")

        }
        sample_key_value$channel <- NULL
        chr_df <- spread(unique(sample_key_value), key, value, fill= NA)
        rownames(chr_df) <- chr_df$sample
        chr_df$sample <- NULL
        if (nrow(chr_df)>0) {
            new.pdata <- cbind(new.pdata, chr_df)
        }
        new.pdata
    }

    if (ncol(es) > 0 && length(chr)>0) {
        new.pdata <-  parsePData(phenoData(es)[, chr])
        pData(es)[, chr] <-NULL
        pData(es)[,colnames(new.pdata)] <- new.pdata
    }

    es
}


#' Load ExpressionSet from GEO Series
#'
#'\code{getGSE} return the ExpressionSet object(s) corresponding
#'     to GEO Series Identifier.
#'
#' @param name String, containing GEO identifier of the dataset.
#'     It should start with 'GSE' or 'GDS' and can include exact GPL
#'     to annotate dataset, separated with dash ('-') from the identifier.
#'
#' @param destdir Directory for caching loaded Series and GPL
#'     files from GEO database.
#'
#' @param mirrorPath URL string which specifies the source of matrices.
#'
#' @return List of ExpressionSet objects, that were available by given
#'     in \code{name} variable GEO identifier.
#'
#' @examples
#' \dontrun{
#'     getGSE('GSE14308', destdir = 'cache')
#'     getGSE('GSE27112')
#'     getGSE('GSE53986')
#' }
#'
#' @export
#' @import rhdf5
getGSE <- function(name, destdir = getPhantasusConf("cache_folders")$geo_path,
                   mirrorPath = getPhantasusConf("geo_mirrors")) {
    if (!isValidExperimentID(name)) {
      stop(name, " does not look like a valid GEO Series ID")
    }
    GEO <- unlist(strsplit(name, "-"))[1]

    stub <- gsub("\\d{1,3}$", "nnn", GEO, perl = TRUE)
    filename <- sprintf("%s_series_matrix.txt.gz", name)
    gseurl <- "%s/geo/series/%s/%s/matrix/%s"

    fullGEODirPath <- getGEODir(name, destdir)
    destfile <- file.path(fullGEODirPath, filename)

    if (!checkGSEType(name, destdir)) {
      stop('Currently unsupported experiment type')
    }

    infile <- file.exists(destfile)

    # Several GEO mirrors supported, but use only one for now.
    mirrorPath <- head(mirrorPath,1)
    if (!infile) {
        tempDestFile <- tempfile(paste0(filename, ".load"), tmpdir=fullGEODirPath)
        for (url in  sprintf(gseurl, mirrorPath,
                             stub, GEO, filename)){
            tryCatch({
                # TODO: should safeDownload() be used here?
                # it doesn't include deletion of temp dest file though, and doesn't suppress errors/warnings
                utils::download.file(url,
                                     destfile = tempDestFile,
                                     method="libcurl")
                file.rename(tempDestFile, destfile)
                infile <- TRUE
            },
            error = function(e) {
                if (file.exists(tempDestFile)) {
                    file.remove(tempDestFile)
                }

            },
            warning = function(w) {
                if (file.exists(tempDestFile)) {
                    file.remove(tempDestFile)
                }
            })
            if(file.exists(destfile)){
                break
            }
        }

    } else {
        message(paste("Loading from locally found file", destfile))
    }

    if (infile && file.size(destfile) > 0) {
        ess <- list(suppressWarnings(getGEO(filename = destfile, destdir = fullGEODirPath, getGPL = FALSE, AnnotGPL = FALSE)))
        for (i in seq_len(length(ess))) {
            ess[[i]] <- annotateFeatureData(ess[[i]], destdir)
        }
    } else {
        gpls <- fromJSON(checkGPLs(name))
        if (length(gpls) == 0) {
            stop(paste("Dataset", name, "not found"))
        }
        if (length(gpls) == 1 && gpls == name) {
            stop(paste("Can't download dataset ", name))

        }
        ess <- list()
        for (i in 1:length(gpls)) {
          ess[[gpls[[i]]]] <- getGSE(gpls[[i]], destdir = destdir, mirrorPath = mirrorPath)[[1]]
        }
        return(ess)
    }

    ess <- lapply(ess, filterFeatureAnnotations)
    archs4_files <- getArchs4Files(destdir)
    useHSDS <- getOption("PhantasusUseHSDS")
    counts_path <- getPhantasusConf("cache_folders")$rnaseq_counts
    if (is.null(useHSDS)){
        if (dir.exists(counts_path)){
            ess <- lapply(ess, loadCounts, counts_dir = counts_path)
        }
    } else{
        if (useHSDS == TRUE){
            ess <- lapply(ess, phantasusLite::loadCountsFromHSDS, url = counts_path)
        }
    }
    if (length(archs4_files) > 0)  {
        ess <- lapply(ess, loadFromARCHS4, archs4_files=archs4_files)
    }
    ess <- lapply(ess, filterPhenoAnnotations)

    ess <- lapply(ess, inferCondition)

    ess


}

#' Load ExpressionSet by GEO identifier
#'
#'\code{getES} return the ExpressionSet object(s) corresponding
#'     to GEO identifier.
#'
#' @param name String, containing GEO identifier of the dataset.
#'     It should start with 'GSE' or 'GDS' and can include exact GPL
#'     to annotate dataset, separated with dash ('-') from the identifier.
#'
#' @param type Type of the dataset: 'GSE' or 'GDS'. If not specified,
#'     the function will take first three letters
#'     of \code{name} variable as type.
#'
#' @param destdir Directory for caching loaded Series and GPL
#'     files from GEO database.
#'
#' @param mirrorPath URL string which specifies the source of matrices.
#'
#' @return List of ExpressionSet objects, that were available by given
#'     in \code{name} variable GEO identifier.
#'
#' @examples
#' \dontrun{
#'     getES('GSE14308', type = 'GSE', destdir = 'cache')
#'     getES('GSE27112')
#'     getES('GDS4922')
#' }
#'
#' @export
getES <- function(name, type = NA, destdir = getPhantasusConf("cache_folders")$geo_path,
                  mirrorPath = getPhantasusConf("geo_mirrors")) {
    if (is.na(type)) {
        type <- substr(name, 1, 3)
    }
    geoDir <- getGEODir(name, destdir)
    possibly.cached <- file.path(geoDir, paste0(name, ".rda"))
    if (file.exists(possibly.cached)) {
        load(possibly.cached)
        message(paste("Loaded from locally cached parsed file", possibly.cached))
    } else {
        if (type == "GSE") {
            res <- getGSE(name, destdir, mirrorPath)
        } else if (type == "GDS") {
            res <- getGDS(name, destdir, mirrorPath)
        } else {
            stop("Incorrect name or type of the dataset")
        }
        if (length(res) > 1) {
            for (i in 1:length(res)) {
                ess <- list(res[[i]])
                destfile <- file.path(geoDir,
                                      paste0(name,
                                             "-",
                                             annotation(res[[i]]),
                                             ".rda"))
                message(paste("Cached dataset to ", destfile))
                save(ess, file = destfile)
            }
        }
        ess <- res
        destfile <- file.path(geoDir, paste0(name, ".rda"))
        message(paste("Cached dataset to ", destfile))
        save(ess, file = destfile)
    }
    ess
}


listCachedESs <- function(destdir) {
    res <- list.files(destdir, pattern=".*\\.rda$", recursive = TRUE)
    res <- grep("\\.gz\\.rda$", res, invert = TRUE, value = TRUE)
    res <- res[grep("^(GSE|GDS)", basename(res), value = FALSE)]
    res <- sub("\\.rda$", "", res)
    res
}

#' Reparse cached expression sets from GEO.
#'
#' The function should be used on phantasus version updates that change
#' behavior of loading datasets from GEO. It finds all the datasets
#' that were cached and runs `getES` for them again. The function
#' uses cached Series and other files from GEO.
#'
#' @param destdir Directory used for caching loaded Series files from GEO database.
#'
#' @param mirrorPath URL string which specifies the source of matrices.
#'
#' @return vector of previously cached GSE IDs
#'
#' @examples
#' reparseCachedESs(destdir=tempdir(), "https://ftp.ncbi.nlm.nih.gov")
#'
#' @export
reparseCachedESs <- function(destdir,
                                mirrorPath = getPhantasusConf("geo_mirrors")) {
    toReparse <- listCachedESs(destdir)

    for (path in toReparse) {
        name <- basename(path)
        geoDir <- getGEODir(name, destdir)
        message(paste0("Reparsing dataset ", name))
        destfile <- file.path(geoDir, paste0(name, ".rda"))
        bakfile <- paste0(destfile, ".bak")
        binfile <- file.path(geoDir, paste0(name, ".bin"))
        if (file.exists(binfile)) {
            file.remove(binfile)
        }

        tryCatch({
            if (file.exists(destfile)) {
                file.rename(destfile, bakfile)
            }

            getES(name, destdir = destdir, mirrorPath = mirrorPath)
            if (file.exists(bakfile)) {
                file.remove(bakfile)
            }

        }, error = function(e) {
            message(paste0("Error occured while reparsing, old file stored as ",
                           bakfile))
        })
    }
    return(basename(toReparse))
}


#' Check possible annotations for GEO Dataset.
#'
#' \code{checkGPLs} returns GPL-names for
#'     the specified GEO identifier.
#'
#' @param name String, containing GEO identifier of the dataset.
#'
#' @return Vector of filenames serialized in JSON format.
#'     If there is only one GPL for that dataset, the function will
#'     return \code{name}.
#'
#' @examples
#' \dontrun{
#' checkGPLs('GSE27112')
#' checkGPLs('GSE14308')
#' }
#' @keywords internal
checkGPLsFallback <- function(name) {
    spl <- unlist(strsplit(name, "-", fixed=TRUE))
    if (length(spl) == 2) {
        gpls <- jsonlite::fromJSON(checkGPLs(spl[1]))
        gpls <- intersect(gpls, name)
        return(jsonlite::toJSON(gpls))
    }

    mirrorPath <-  getPhantasusConf("geo_mirrors")

    cacheDir <- getPhantasusConf("cache_folders")$geo_path


    type <- substr(name, 1, 3)
    assertthat::assert_that( (type == "GDS" || type == "GSE")
                            && nchar(name) >= 4)

    stub <- gsub("\\d{1,3}$", "nnn", name, perl = TRUE)
    gdsurl <- "%s/geo/%s/%s/%s/"

    url <- sprintf(gdsurl, mirrorPath,
                   if (type == "GDS") "datasets" else "series", stub, name)

    cachePath <- file.path(cacheDir,
                        if (type == "GDS") "datasets" else "series",
                        stub,
                        name,
                        "gpls")

    dir.create(dirname(cachePath), recursive = TRUE, showWarnings = FALSE)

    if (file.exists(cachePath)) {
        gpls <- readLines(cachePath)
        if (length(gpls) != 0) {
            return(jsonlite::toJSON(gpls))
        }
    }

    if (type != "GDS") {
        url <- paste0(url, "matrix/")
    }

    gpls <- c()

    tryCatch({
        for (mirror in url){
            resp <- httr::GET(mirror)
            if (httr::status_code(resp) != 404) {
                break
            }
        }
        if (httr::status_code(resp) == 404) {
            warning("No such dataset")
            return(jsonlite::toJSON(c()))
        } else {
            if (type == "GDS") {
                gpls <- c(name)
            } else {
                con <- rawConnection(resp$content)
                file.names <- GEOquery:::getDirListing(con)
                close(con)

                file.names <- file.names[grepl(pattern = paste0("^", name),
                                            x = file.names)]

                file.names <- unlist(lapply(file.names, function(x) {
                    paste0(substr(x, 1, regexpr("_", x) - 1))
                }))
                if (length(file.names) == 1) {
                    file.names <- c(name)
                }
                gpls <- file.names
            }

            writeLines(gpls, cachePath)
            return(jsonlite::toJSON(gpls))
        }
    },
    error = function(e) {
        message("Problems establishing connection.")
        return(jsonlite::toJSON(c()))
    })
}

checkGPLs <- function(name) {
  spl <- unlist(strsplit(name, "-", fixed=TRUE))
  GEO <- spl[1]
  if (length(spl) == 2) {
    gpls <- jsonlite::fromJSON(checkGPLs(spl[1]))
    gpls <- intersect(gpls, name)
    return(jsonlite::toJSON(gpls))
  }

  cacheDir <- getPhantasusConf("cache_folders")$geo_path

  GEOdir <- dirname(getGEODir(GEO, cacheDir))
  GPLCacheFile <- file.path(GEOdir, 'gpls')
  if (file.exists(GPLCacheFile)) {
    gpls <- readLines(GPLCacheFile)
    if (length(gpls) != 0) {
      return(jsonlite::toJSON(gpls))
    }
  }

  tryCatch({
    briefData <- getBriefData(name, cacheDir)
    platforms <- briefData$platform_id

    if (length(platforms) < 1) {
      stop('No platforms in brief data')
    }

    if (length(platforms) == 1) {
      gpls <- c(name)
    }

    if (length(platforms) >= 2) {
      gpls <- unlist(lapply(platforms, function (platform) {
        paste(spl[1], platform, sep='-')
      }))
    }

    writeLines(gpls, GPLCacheFile)
    return(jsonlite::toJSON(gpls))
  }, error=function(e) {
    return(checkGPLsFallback(name))
  })
}

#' Checks GSE to be supported
#' @param name GSE id, with optional GPL specification
#' @param destDir path to cache directory
#' @param combine function on how to combine results, when multiple platforms are present
#' @return logical vector if the dataset is supported or not
#' @keywords internal
checkGSEType <- function (name, destDir, combine=any) {
  spl <- unlist(strsplit(name, "-", fixed=TRUE))
  GEO <- spl[1]

  briefData <- getBriefData(name, destDir)

  gpls <- spl[2]
  if (is.na(gpls)) {
      gpls <- briefData$platform_id
  }

  gplsOK <- sapply(gpls, function(gpl) {
      gplBrief <- getBriefData(gpl, destdir=destDir)
      return(as.numeric(gplBrief$data_row_count) <= 100000)
  })

  return(combine(gplsOK))
}

removeRepeatWords <- function(titles) {
    titles_without_repeat_words <- titles
    repeat_words <- regmatches(titles, regexpr("(?![-+}\\]\\)])(\\W|_)*\\w*$", titles, ignore.case = TRUE, perl = TRUE))
    lsuff <- lcSuffix(repeat_words, ignore.case = TRUE)
    if ((all(stringr::str_length(lsuff) == stringr::str_length(repeat_words))) & (all(sub("(\\W*|_)*\\w*$", "", titles, ignore.case = TRUE, perl = TRUE) != ""))) {
        titles_without_repeat_words <- sub("(?![-+}\\]\\)])(\\W|_)*\\w*$", "", titles, ignore.case = TRUE, perl = TRUE)
    }
    if (all(grepl("-$", titles_without_repeat_words, ignore.case = TRUE))) titles_without_repeat_words <- sub("-$", "", titles_without_repeat_words, ignore.case = TRUE, perl = TRUE)
    return(titles_without_repeat_words)
}

inferConditionImpl <- function(gse_titles) {
    inferCondition <- gse_titles
    rep_num <- NULL
    if ((length(inferCondition) > 40) | (length(inferCondition) < 3))
    {
        return(list())
    } else if (! all(grepl("[A-z]", inferCondition, ignore.case = TRUE)))
    {
        return(list())
    } else if (all(grepl(" vs[. ]{1}", inferCondition)))
    {
        return(list())
    } else
    {
        lsuff <- lcSuffix(inferCondition)
        suff_length <- stringr::str_length(lsuff)
        if (suff_length > 1) inferCondition <- stringr::str_sub(inferCondition, 1, -suff_length-1)
        if (all(grepl("((bio[A-z]*)|(tech[A-z]*))?[ .#_-]*((rep.*)|(set.*)|(sample)|(exp.*)|(case))[ .#_-]*\\d+", inferCondition, ignore.case = TRUE)))
        {
            sample <- regmatches(inferCondition, regexpr("((bio[A-z]*)|(tech[A-z]*))?[ .#_-]*((rep.*)|(set.*)|(sample)|(exp.*)|(case))[ .#_-]*\\d+", inferCondition, ignore.case = TRUE))
            rep_num <- regmatches(sample, regexpr("\\d+$", sample))
            inferCondition <- sub("(?![-+}\\]\\)])(\\W|_)*((bio[A-z]*)|(tech[A-z]*))?[ .#_-]*((rep.*)|(set.*)|(sample)|(exp.*)|(case))[ .#_-]*\\d+(\\W|_)*", "", inferCondition, ignore.case = TRUE, perl = TRUE)

        }
        else if (all(grepl("((bio[A-z]*)|(tech[A-z]*))?[ .#_-]*((rep.*)|(set.*)|(sample)|(exp.*)|(case))[ .#_-]*[A-z]$", inferCondition, ignore.case = TRUE)))
        {
            sample <- regmatches(inferCondition, regexpr("((bio[A-z]*)|(tech[A-z]*))?[ .#_-]*((rep.*)|(set.*)|(sample)|(exp.*)|(case))[ .#_-]*[A-z]$", inferCondition, ignore.case = TRUE))
            rep_num <- regmatches(sample, regexpr("[A-z]$", sample))
            inferCondition <- sub("(?![-+}\\]\\)])(\\W|_)*((bio[A-z]*)|(tech[A-z]*))?[ .#_-]*((rep.*)|(set.*)|(sample)|(exp.*)|(case))[ .#_-]*[A-z]$", "", inferCondition, ignore.case = TRUE, perl = TRUE)

        }
        else if (length(unique(sub("[- \\.#_]*\\d+$", "", inferCondition))) == 1) {
            return(list())
        }
        else if (all(grepl("[- \\.#_]*\\d+$", inferCondition)) & (length(unique(regmatches(inferCondition, regexpr("\\d+$", inferCondition)))) > 1))
        {
            sample <- sub("\\d+$", "", inferCondition)
            if (length(unique(regmatches(sample, regexpr("[- \\.#_]$", sample)))) > 1)
            {
                return(list())
            }
            else
            {
                rep_num <- regmatches(inferCondition, regexpr("\\d+$", inferCondition))
                inferCondition <- removeRepeatWords(sub("[ \\.#_]*\\d+$", "", inferCondition))

            }
        }
        else return(list())
    }

    if ((length(rep_num) > 1) & (length(unique(inferCondition)) < length(unique(gse_titles))) & (length(unique(inferCondition)) > 1))
        return(list(condition=inferCondition, replicate=rep_num))
    else return(list())
}

inferCondition <- function(es) {
    newAnnot <- inferConditionImpl(es$title)
    if (length(newAnnot) == 2) {
        pData(es)$condition <- newAnnot$condition
        pData(es)$replicate <- newAnnot$replicate
    }
    es
}


downloadGPL <- function (GPL, destdir = getPhantasusConf("cache_folders")$geo_path) {
  GPL <- toupper(GPL)
  stub = gsub('\\d{1,3}$','nnn',GPL,perl=TRUE)
  GPLDirPath <- '%s/platforms/%s/%s/annot'
  fullGPLDirPath <- file.path(sprintf(GPLDirPath, destdir, stub, GPL))

  cachedFile <- file.path(fullGPLDirPath, paste0(GPL, ".annot.gz"))
  cachedSoft <- file.path(fullGPLDirPath, paste0(GPL, ".soft"))
  cachedSoftGz <- file.path(fullGPLDirPath, paste0(GPL, ".soft.gz"))
  if (file.exists(cachedFile)) {
    if (file.size(cachedFile) == 0) {
      file.remove(cachedFile)
      message(cachedFile, ' size is 0')
    } else {
      return (cachedFile)
    }
  } else if (file.exists(cachedSoft)) {
    if (file.size(cachedSoft) == 0) {
      file.remove(cachedSoft)
      message(cachedSoft, ' size is 0')
    } else {
      return (cachedSoft)
    }
  } else if (file.exists(cachedSoftGz)) {
    if (file.size(cachedSoftGz) == 0) {
      file.remove(cachedSoftGz)
      message(cachedSoftGz, ' size is 0')
    } else {
      return (cachedSoftGz)
    }
  }

  mirrorPath <- getPhantasusConf("geo_mirrors")
  annotPath <- paste0(mirrorPath,  '/geo/platforms/%s/%s/annot/%s')
  annotURL <- sprintf(annotPath,stub,GPL,paste0(GPL,'.annot.gz'))
  dir.create(fullGPLDirPath, showWarnings = FALSE, recursive = TRUE)
  targetFile <- ''

  for (url in annotURL){
      req <- httr::HEAD(url)
      if (httr::status_code(req) != 404) {
          # annot available
          tmp <- tempfile(pattern=paste0(GPL, ".annot.gz"), tmpdir=fullGPLDirPath)
          tryCatch({
              download.file(url, tmp)
          }, error=function (e) {
              unlink(tmp)
              stop('Could not download GPL ', GPL, e)
          })

          file.copy(tmp, cachedFile)
          unlink(tmp)
          targetFile <- cachedFile
      }
      if (targetFile != ''){
          break
      }
  }

  if (targetFile == ''){
    # need submitter
    tmp <- tempfile(pattern=paste0(GPL, ".soft"), tmpdir=fullGPLDirPath)
    apiURL <- "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi"
    submitterURL <- paste(apiURL,'?targ=self&acc=',GPL,'&form=text&view=data',sep='')
    tryCatch({
      download.file(submitterURL, tmp, headers = c("accept-encoding" = "gzip"))
    }, error=function (e) {
      unlink(tmp)
      stop('Could not download GPL ', GPL, e)
    })

    file.rename(tmp, cachedSoftGz)
    targetFile <- cachedSoftGz
  }

  if (file.size(targetFile) == 0) {
    file.remove(targetFile)
    stop('Could not download GPL ', GPL, '. Zero file')
  }

  con <- file(targetFile, 'r')
  first_hundred_lines <- readLines(con, n=100)
  close(con)
  found<-grep('^\\^(PLATFORM|ANNOTATION)', first_hundred_lines, ignore.case = TRUE)
  if (!length(found)) {
    file.remove(targetFile)
    stop('Could not download GPL ', GPL, '.  Possible NCBI problems')
  }


  return(targetFile)
}

getGPLAnnotation <- function (GPL, destdir = getPhantasusConf("cache_folders")$geo_path) {
    filename <- downloadGPL(GPL, destdir)
    # ret <- parseGEO(filename)

    # TODO: workaround for https://github.com/seandavi/GEOquery/pull/120,
    txt <- data.table::fread(filename,sep="")[[1]]
    if (length(txt) == 0) {
      txt <- c("table_begin", "table_end")
    }
    ret <- GEOquery:::.parseGPLTxt(txt)
    # end of workaround
    return(ret)
}

annotateFeatureData <- function (es, destdir = getPhantasusConf("cache_folders")$geo_path) {
    platform <- as.character(es$platform_id)[1]
    platformParsed <- getGPLAnnotation(platform, destdir)

    #https://github.com/seandavi/GEOquery/blob/master/R/parseGEO.R#L569
    #############################
    vmd <- Columns(platformParsed)
    dat <- Table(platformParsed)
    ## GEO uses "TAG" instead of "ID" for SAGE GSE/GPL entries, but it is apparently
    ##     always the first column, so use dat[,1] instead of dat$ID
    ## The next line deals with the empty GSE
    tmpnames=character(0)
    if(ncol(dat)>0) {
        tmpnames=as.character(dat[,1])
    }
    ## Fixed bug caused by an ID being "NA" in GSE15197, for example
    tmpnames[is.na(tmpnames)]="NA"
    rownames(dat) <- make.unique(tmpnames)
    ## Apparently, NCBI GEO uses case-insensitive matching
    ## between platform IDs and series ID Refs ???
    dat <- dat[match(tolower(rownames(es)),tolower(rownames(dat))),]
    # Fix possibility of duplicate column names in the
    # GPL files; this is prevalent in the Annotation GPLs
    rownames(vmd) <- make.unique(colnames(Table(platformParsed)))
    colnames(dat) <- rownames(vmd)
    ##############################

    featureData(es) <- new('AnnotatedDataFrame',data=dat,varMetadata=vmd)
    es
}

context("Load GEO and its utils")
library(jsonlite)
library(Biobase)
library(data.table)
Sys.setenv(R_USER_CONFIG_DIR = system.file("/testdata/config", package = "phantasus"))
test_that("loadGEO finishes with result", {
    old_conf <- Sys.getenv("R_CONFIG_ACTIVE")
    Sys.setenv(R_CONFIG_ACTIVE = "default")
    x <- loadGEO("GSE27112")
    expect_is(x, "json")
    cacheDir <- getPhantasusConf("cache_root")
    binPath <- file.path(cacheDir, fromJSON(x))
    ess <- protolite::unserialize_pb(readBin(binPath, what="raw", n=100000000))$ess
    expect_equal(length(ess), 2)

    x <- loadGEO("GSE27112-GPL6885")
    expect_is(x, "json")

    binPath <- file.path(cacheDir, fromJSON(x))
    ess <- protolite::unserialize_pb(readBin(binPath, what="raw", n=100000000))$ess

    expect_equal(length(ess), 1)

    expect_is(loadGEO("GSE14308"), "json")
    expect_is(loadGEO("GDS4885"), "json")

    expect_error(loadGEO("WRONGNAME"))
    Sys.setenv(R_CONFIG_ACTIVE = old_conf)

})

test_that("getGDS adds id field for GDS datasets", {
    old_conf <- Sys.getenv("R_CONFIG_ACTIVE")
    Sys.setenv(R_CONFIG_ACTIVE = "default")
    a <- getGDS("GDS4885")[[1]]
    expect_true("id" %in% tolower(fvarLabels(a)))
    Sys.setenv(R_CONFIG_ACTIVE = old_conf)
})

test_that("filterPhenoAnnotations saves colnames", {
    old_conf <- Sys.getenv("R_CONFIG_ACTIVE")
    Sys.setenv(R_CONFIG_ACTIVE = "test_real_geo")
    es <- getES("GSE53986")[[1]]
    expect_true(all(colnames(es) == colnames(exprs(es))))
    Sys.setenv(R_CONFIG_ACTIVE = old_conf)
})

test_that("reparseCachedGSEs works", {
    old_conf <- Sys.getenv("R_CONFIG_ACTIVE")
    Sys.setenv(R_CONFIG_ACTIVE = "test_real_geo")
    getES("GSE14308")
    expect_true("GSE14308" %in% reparseCachedESs(destdir = getPhantasusConf("cache_folders")$geo_path))
    Sys.setenv(R_CONFIG_ACTIVE = old_conf)
})

test_that("checkGPLs counts gpls correctly", {
    old_conf <- Sys.getenv("R_CONFIG_ACTIVE")
    Sys.setenv(R_CONFIG_ACTIVE = "default")
    expect_equal(fromJSON(checkGPLs("GSE14308")), c("GSE14308"))
    expect_equal(fromJSON(checkGPLs("GDS4885")), c("GDS4885"))
    expect_length(fromJSON(checkGPLs("GSE27112")), 2)
    expect_length(fromJSON(checkGPLs("GSE10000")), 2)
    expect_warning(checkGPLs("GSE0"))
    Sys.setenv(R_CONFIG_ACTIVE = old_conf)

})

test_that("checkGPLs works with fully specified name", {
    old_conf <- Sys.getenv("R_CONFIG_ACTIVE")
    Sys.setenv(R_CONFIG_ACTIVE = "default")
    expect_equal(fromJSON(checkGPLs("GSE27112-GPL6885")), c("GSE27112-GPL6885"))
    Sys.setenv(R_CONFIG_ACTIVE = old_conf)
})

# TODO: adapt to new checkGPLs
#test_that("checkGPLs counts existing files correctly without connection", {
    #options(phantasusMirrorPath = "https://alserglab.wustl.edu/files/phantasus")
    #options(phantasusCacheDir = tempfile())
    #expect_length(fromJSON(checkGPLs("GSE27112")), 2)
    #options(phantasusMirrorPath = "https://notworkingdomain")

    #expect_message(checkGPLs("GSE14308"), regexp = "Problems establishing connection")
    #expect_length(fromJSON(checkGPLs("GSE27112")), 2)

    #options(phantasusCacheDir = NULL,
            #phantasusMirrorPath = NULL)
#})

test_that("getGSE works with ARCHS4", {
    ess <- getGSE("GSE99709", destdir=system.file("testdata", package="phantasus"), mirrorPath = "https://ftp.ncbi.nlm.nih.gov")
    expect_gt(nrow(ess[[1]]), 0)
    expect_gt(ncol(ess[[1]]), 0)
})

test_that("InferConditionImpl  works correctly", {
    tests <- fread(system.file("testdata/dts.tsv", package="phantasus"))
    test_ds <- data.table(title=tests$Title, series=tests$Series, accession=tests$Accession, rep=tests$Replicate, inferCondition=tests$InferCondition)
    cond <- split(test_ds$title, test_ds$series)
    inf_cond_test <- split(test_ds$inferCondition, test_ds$series)
    rep_test <- split(test_ds$rep, test_ds$series)
    new_cond <- lapply(cond, inferConditionImpl)
    expect_equal(new_cond$GSE100221, list()) # text in all titles is unique
    expect_equal(new_cond$GSE10380, list())  # long dataset
    expect_equal(new_cond$GSE10382, list())  # number-only titles
    expect_equal(new_cond$GSE10383, list())  # two-color datasets
    expect_equal(new_cond$GSE10385, list())  # the same text and replicate number in all titles
    expect_equal(new_cond$GSE10039, list())  # ambiguous replicate number "High_Mo_seg_pool_Ler_col_F2" "Low_Mo_seg_pool_Ler_col_F2"  "Col-0 3"
    expect_equal(new_cond$GSE101508$condition, inf_cond_test$GSE101508) #"IFNγ+LPS rep2" -> "IFNγ+LPS" + "2"
    expect_equal(new_cond$GSE101508$replicate, as.character(rep_test$GSE101508))
    expect_equal(new_cond$GSE10392$condition, inf_cond_test$GSE10392) # "MPA 1" - > "MPA" + "1"
    expect_equal(new_cond$GSE10392$replicate, as.character(rep_test$GSE10392))
    expect_equal(new_cond$GSE10123$condition, inf_cond_test$GSE10123) # "WT-GFP-lamin A Induction: Day 0 Replicate A" -> "WT-GFP-lamin A Induction: Day 0" + "A"
    expect_equal(new_cond$GSE10123$replicate, as.character(rep_test$GSE10123))


})

test_that("getGPLAnnotation works with errorneous empty annotation files", {
    destdir <- file.path(tempdir(), "cache_bad")
    dir.create(destdir, showWarnings = FALSE)

    GPL <- "GPL17021"
    stub = gsub('\\d{1,3}$','nnn',GPL,perl=TRUE)
    GPLDirPath <- '%s/geo/platforms/%s/%s/annot'
    fullGPLDirPath <- file.path(sprintf(GPLDirPath, destdir, stub, GPL))

    dir.create(fullGPLDirPath, showWarnings = FALSE, recursive = TRUE)
    file.create(file.path(fullGPLDirPath, paste0(GPL, ".annot.gz")))

    gpl <- getGPLAnnotation(GPL, destdir)
    expect_true(!is.null(gpl))
})

test_that("getGSEType works", {
    expect_true(checkGSEType('GSE53986', tempdir()))
    expect_true(checkGSEType('GSE99709', tempdir()))
    expect_true(checkGSEType('GSE33356', tempdir()))
    expect_identical(checkGSEType('GSE33356', tempdir(), combine=identity),
                     c(GPL570=TRUE, GPL6801=FALSE))
    expect_false(checkGSEType('GSE33356-GPL6801', tempdir()))
})

test_that("filterPhenoAnnotations does not split title", {
    old_conf <- Sys.getenv("R_CONFIG_ACTIVE")
    Sys.setenv(R_CONFIG_ACTIVE = "test_real_geo")
    ess <- getGSE("GSE120978")
    expect_true("title" %in% varLabels(ess[[1]]))
    Sys.setenv(R_CONFIG_ACTIVE = old_conf)
})

test_that("filterPhenoAnnotations parses characteristics", {
    load(file = system.file("testdata/GSE120978_unfiltered_pData.rda", package="phantasus"))
    es <- ess[[1]]
    empty_strings <- sample(1:ncol(es), 3, replace=FALSE)
    pData(es)[empty_strings,"characteristics_ch1" ] <- ""
    old_values <-  pData(es)$"characteristics_ch1"[-empty_strings]
    es <- filterPhenoAnnotations(es)
    expect_true(all(c("gender", "strain", "tissue") %in% varLabels(es)))
    expect_true(all(is.na(pData(es)$gender[empty_strings])))
    new_values <- pData(es)$gender[-empty_strings]
    expect_true(all(old_values == paste0("gender: ", new_values)))
})

#' Load GEO Dataset.
#'
#' \code{loadPreloaded} returns the file with serialized ExpressionSets using
#'     ProtoBuf, that were preloaded on server.
#'
#' @param name String, containing filename. Assuming
#'     that in the directory with preloaded files \code{preloadedDir}
#'     exists file \code{filename.rda} with list of ExpressionSets \code{ess}.
#'
#'
#' @return File with ProtoBuf-serialized ExpressionSet-s
#'     that were loaded from specified file.
#'
#' @import Biobase
#' @keywords internal
loadPreloaded <- function(name) {
    preloadedDir <- getOption("phantasusPreloadedDir", default = getPhantasusConf("preloaded_dir"))
    cacheDir <- getPhantasusConf("cache_root")
    if (is.null(preloadedDir)) {
        stop("Specify the directory with presaved files")
    } else if (!dir.exists(preloadedDir)) {
        stop("No such directory")
    }

    fileToLoad <- file.path(preloadedDir, paste0(name, '.rda'))
    if (!file.exists(fileToLoad)) {
        stop('No such dataset')
    }

    x <- load(fileToLoad) # must return the object ess
    loaded <- get(x)

    wrongFormat <- paste("Wrong format.",
                            "File must contain either ExpressionSet",
                            "or list of ExpressionSets")

    ess <- NULL

    if (class(loaded) == "ExpressionSet") {
        ess <- list()
        ess[[name]] <- loaded
    } else if (class(loaded) == "list") {
        ess <- loaded
        if (!all(unlist(lapply(loaded, class)) == "ExpressionSet")) {
            stop(wrongFormat)
        }
    } else {
        stop(wrongFormat)
    }

    files <- list()
    assign("es", ess[[1]], envir = parent.frame())
    files[[name]] <- writeToList(ess[[1]])
    path <- NULL

    if (is.null(cacheDir)) {
        cacheDir <- tempdir()
    }

    preloadedCacheDir <- file.path(cacheDir, 'preloaded')
    binaryName <- paste0(name, '.bin')
    dir.create(preloadedCacheDir, showWarnings = FALSE, recursive = TRUE)



    cachedFile <- file.path(preloadedCacheDir, binaryName)
    dir.create(dirname(cachedFile), showWarnings = FALSE, recursive = TRUE)
    bin_is_valid <- checkBinValidity(cachedFile, file.info(fileToLoad)$ctime)
    if (!bin_is_valid) {
        writeBin(protolite::serialize_pb(list(layout_version = as.raw(PROTOBUF_LAYOUT_VERSION), ess = files)), cachedFile)
    }

    path <- paste0('/preloaded/', binaryName)

    jsonlite::toJSON(path)

}
#' Generate files for preloaded session from a session link.
#'
#' @param sessionURL String with session link produced by phantasus.
#' @param preloadedName String with name that should be assigned to the session.
#' @param preloadedDir  Path to the directory with preloaded datasets and sessions.
#' @return Function produces two files (\code{preloadedName.rda} with ExpressionSet
#'         and \code{preloadedName.json} with session features) in \code{preloadedDir} folder.
#' @export
#' @examples
#' \dontrun{
#' sessionURL <- "https://ctlab.itmo.ru/phantasus/?session=x063c1b365b9211" # link from 'Get dataset link...' tool in phantasus
#' newName <- "my_session" # user defined name
#' preloadedDir <- "./preloaded" # directory where files will be stored. In order too get access through phantasus web-app should be preloadedDir
#' dir.create(preloadedDir, showWarnings = FALSE)
#' generatePreloadedSession(sessionURL= sessionURL,
#'                          preloadedName = newName,
#'                          preloadedDir = preloadedDir)
#'
#' servePhantasus(preloadedDir=preloadedDir, openInBrowser=FALSE)
#' # open browser manually at http://0.0.0.0:8000/phantasus/index.html?preloaded=my_session
#' }
generatePreloadedSession <- function (sessionURL, preloadedName, preloadedDir) {
    sessionName <-gsub(".+session=([^&]+).*","\\1", sessionURL)
    phantasusPath <- gsub("([^?])(index\\.html)?\\?.*","\\1", sessionURL)
    if (is.null(preloadedDir)) {
        stop("Specify the directory with presaved files")
    } else if (!dir.exists(preloadedDir)) {
        stop("Preloaded directory  does not exist")
    }
    RDataPath = tempfile()
    RDA_req =curl::curl_fetch_disk(url = paste0(phantasusPath,
                                                "ocpu/tmp/", sessionName, "/files/.RData"),
                                   path = RDataPath)

    if (!RDA_req$status_code==200) {
        stop('Invalid session')
    }
    env <- new.env()
    load(RDataPath, envir = env)
    file.remove(RDataPath)

    if (!is.null(env$es)) {
        assign("es", env$es, envir = parent.frame())
        save(es, file = file.path(preloadedDir, paste0(preloadedName, ".rda")))

        if (!is.null(env$heatmapJson)) {
            assign("heatmapJson", env$heatmapJson, envir = parent.frame())
            heatmapPath <- file.path(preloadedDir, paste0(preloadedName,'.json'))
            write(jsonlite::toJSON(heatmapJson, auto_unbox = TRUE, null = 'null'), heatmapPath)
        }
    }
}

context("Load preloaded datasets")
library(Biobase)

test_that("loadPreloaded throws errors correctly", {
    expect_error(loadPreloaded("aa"), regexp = "Specify the directory")

    options(phantasusPreloadedDir = "falseDirectory")
    expect_error(loadPreloaded("aa"), regexp = "No such directory")

    options(phantasusPreloadedDir = system.file("testdata", package="phantasus"))
    expect_error(loadPreloaded("falseFile"), regexp = "No such dataset")
    expect_error(loadPreloaded("wrongFormatFile"), regexp = "Wrong format")
    expect_error(loadPreloaded("wrapped_wrongFormat"), regexp = "Wrong format")

    options(phantasusPreloadedDir = NULL)
})

test_that("loadPreloaded loads file with just one ExpressionSet", {
    options(phantasusPreloadedDir = system.file("testdata", package="phantasus"))
    expect_is(loadPreloaded("aa"), "json")

    options(phantasusPreloadedDir = NULL)
})

#https://support.bioconductor.org/p/67600/#67612
#https://github.com/ctlab/sysbio-training/blob/master/masters-2019/rnaseq/mnt/scripts/do_deseq2.R
#' @import apeglm
deseqAnalysis <- function (es, fieldValues, version = "One-factor design", contrast =  list('Comparison', 'Target', 'Reference'), designFields = NULL, designData = NULL) {
    fieldValues <- replace(fieldValues, fieldValues == "", NA)
    de <- NULL
    contrast <- unlist(contrast)
    if (version == "One-factor design" ){
        de <- deseqAnalysisSimpleImpl(es, fieldValues, contrast)
    }
    if (version == "Advanced design"){
        de <- deseqAnalysisAdvancedImpl(es, designData, designFields, contrast)
    }
    deDf <- as.data.frame(de)
    toRemove <- intersect(colnames(fData(es)), colnames(deDf))
    fData(es)[, toRemove] <- NULL

    es$Comparison <- fieldValues
    fData(es) <- cbind(fData(es), deDf)
    assign("es", es, envir = parent.frame())

    f <- tempfile(pattern = "de", tmpdir = getwd(), fileext = ".bin")
    writeBin(protolite::serialize_pb(as.list(de)), f)
    jsonlite::toJSON(f)
}

deseqAnalysisSimpleImpl <- function(es, fieldValues, contrast){
    fieldValues <- replace(fieldValues, fieldValues == "", NA)

    es.copy <- es
    es.copy$Comparison <- fieldValues
    pData(es.copy)[,"Comparison"] <- as.factor(pData(es.copy)[,"Comparison"])
    pData(es.copy)[,"Comparison"] <- relevel(pData(es.copy)[,"Comparison"], ref = "Reference")
    es.copy <- es.copy[, !is.na(fieldValues)]
    dds <- DESeq2::DESeqDataSetFromMatrix(exprs(es.copy), pData(es.copy), design=~Comparison)

    populatedDds <- DESeq2::DESeq(dds)
    de <- DESeq2::results(populatedDds, contrast = contrast, cooksCutoff = FALSE)
    shr_values <- DESeq2::lfcShrink(populatedDds, coef =  paste(contrast[1], contrast[2], "vs", contrast[3], sep = "_"), type = "apeglm")
    de$log2FoldChange <- shr_values$log2FoldChange
    de$lfcSE <- shr_values$lfcSE
    return(de)
}
deseqAnalysisAdvancedImpl <- function(es, designData, designFields, contrast){

    ux_designMatrix <- getDesignMatrix(designData)
    es.copy <- es
    if ("geo_accession" %in% colnames(pData(es.copy))){
        sampleNames(es.copy) <- pData(es.copy)$geo_accession
    }
    designFields <- unlist(designFields)
    pData(es.copy) <- subset(pData(es.copy), select = designFields)


    pData(es.copy)[,contrast[1]] <- as.factor(pData(es.copy)[,contrast[1]])
    pData(es.copy)[,contrast[1]] <- relevel(pData(es.copy)[,contrast[1]], ref = contrast[3])
    colnames(pData(es.copy)) <- make.names(colnames(pData(es.copy)))
    contrast[1] <- make.names(contrast[1])
    designFields <- c("~1", colnames(pData(es.copy)))
    designFormula <- formula(paste(designFields, collapse = "+"))


    dds <- DESeq2::DESeqDataSetFromMatrix(exprs(es.copy), pData(es.copy), design=designFormula)
    real_designMatrix <- model.matrix(dds@design, data = pData(es.copy))
    colnames(ux_designMatrix) <- make.names(colnames(ux_designMatrix))
    colnames(real_designMatrix) <- make.names(colnames(real_designMatrix))
    real_designMatrix <- real_designMatrix[, colnames(ux_designMatrix)]
    if (!isTRUE(all.equal(ux_designMatrix[,], real_designMatrix[,]))){
        stop("Internal error! Clientside and serverside designs are not equal!")
    }
    populatedDds <- DESeq2::DESeq(dds)
    de <- DESeq2::results(populatedDds, contrast = contrast, cooksCutoff = FALSE)
    coef_name <- paste(contrast[1], contrast[2], "vs", contrast[3], sep = "_")
    coef_name <- make.names(coef_name)
    shr_values <- DESeq2::lfcShrink(populatedDds, coef =coef_name, type = "apeglm")
    de$log2FoldChange <- shr_values$log2FoldChange
    de$lfcSE <- shr_values$lfcSE
    return(de)
}

#' Adjust dataset
#'
#' @param es Expression set to perform adjustment on
#'
#' @param scaleColumnSum perform sum scaling of columns (default FALSE)
#'
#' @param log2 perform logarithm2 adjustment (default FALSE)
#'
#' @param onePlusLog2 perform log2(1+x) adjustment (default FALSE)
#'
#' @param inverseLog2 perform 2^x adjustment (default FALSE)
#'
#' @param quantileNormalize perform quantile normalization (default FALSE)
#'
#' @param zScore perform zScore adjustment:
#'  subtract mean, divide by std (default FALSE)
#'
#' @param robustZScore perform robustZScore adjustment:
#'  subtract median, divide by MAD (default FALSE)
#'
#' @param sweep perform sweep adjustment on rows/columns (default FALSE)
#'
#' @return Nothing. Adjusted dataset will be assigned as ES in global environment
#'
#' @import stats
#'
#' @examples
#' \dontrun{
#' es <- gseGSE('GSE53986')[[1]]
#' adjustDataset(es, log2 = T, quantileNormalize = T)
#' }
#' @keywords internal

adjustDataset <- function (es, scaleColumnSum = NULL,
                           log2 = FALSE, onePlusLog2 = FALSE,
                           inverseLog2 = FALSE, quantileNormalize = FALSE,
                           zScore = FALSE, robustZScore = FALSE, sweep = NULL) {

    if (length(scaleColumnSum) > 0 && is.numeric(scaleColumnSum)) {
        sum <- apply(exprs(es), 2, sum, na.rm = TRUE)
        sum <- t(scaleColumnSum / sum)
        exprs(es) <- sweep(exprs(es), 2, sum, "*")
    }

    if (log2) {
        m <- exprs(es)
        # ignore warning for log x when x < 0
        exprs(es) <- ifelse(m < 0 | is.na(m), 0, suppressWarnings(log2(m)))
    }

    if (onePlusLog2) {
        m <- exprs(es)
        # ignore warning for log x when x < 0
        exprs(es) <- ifelse(m < 0 | is.na(m), 0, suppressWarnings(log2(m + 1)))
    }

    if (inverseLog2) {
        m <- exprs(es)
        exprs(es) <- 2^m
    }

    if (quantileNormalize) {
        exprs(es) <- normalizeBetweenArrays(exprs(es), method = 'quantile')
    }

    if (zScore) {
        exprs(es) <- t(scale(t(exprs(es))))
    }

    if (robustZScore) {
        medians <- apply(exprs(es), 1, median)
        exprs(es) <- sweep(exprs(es), 1, medians, "-")
        mads <- apply(exprs(es), 1, mad)
        exprs(es) <- sweep(exprs(es), 1, mads, "/")
    }

    if (length(sweep) > 0) {
        if (sweep$mode == 'row') {
            margin <- 1
            target <- fData(es)[[sweep$name]]
        } else {
            margin <- 2
            target <- phenoData(es)[[sweep$name]]
        }

        exprs(es) <- sweep(exprs(es), margin, target, sweep$op)
    }

    assign("es", es, envir = parent.frame())
}

#' Principal Component Analysis.
#'
#' \code{calcPCA} calculates PCA-matrix for the given ExpressionSet
#'     and returns this matrix encoded to JSON.
#'
#' @param es an ExpressionSet object, should be normalized
#'
#' @param replacena method for replacing NA values (mean by default)
#'
#' @return json with full description of the plot for plotly.js
#'
#' @import ggplot2
#' @import htmltools
#' @import jsonlite
#'
#' @examples
#' \dontrun{
#' data(es)
#' calcPCA(es)
#' }
#' @keywords internal
calcPCA <- function(es, replacena = "mean") {

    scaledExprs <- unname(exprs(es))

    naInd <- which(is.na(scaledExprs), arr.ind = TRUE)
    if (nrow(naInd) > 0) {
        replaceValues <- apply(scaledExprs, 1, replacena, na.rm=TRUE)
        scaledExprs[naInd] <- replaceValues[naInd[,1]]
    }

    scaledExprs <- t(scale(t(scaledExprs)))
    rowsToPca <- which(!apply(is.na(scaledExprs), 1, any))

    pca <- stats::prcomp(t(scaledExprs[rowsToPca, ]))
    explained <- (pca$sdev) ^ 2 / sum(pca$sdev ^ 2)

    xs <- sprintf("PC%s", seq_along(explained))
    xlabs <- sprintf("%s (%.1f%%)", xs, explained * 100)

    pca.res <- as.matrix(pca$x)
    colnames(pca.res) <- NULL
    row.names(pca.res) <- NULL
    return(jsonlite::toJSON(list(pca = t(pca.res), xlabs = xlabs)))
}

context("PCA Plot")

test_that("pcaPlot finishes with result", {
  load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))
  expect_is(calcPCA(es), "json")
})

test_that("pcaPlot results in columnsXcolumns matrix", {
  load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))
  expect_equal(nrow(jsonlite::fromJSON(calcPCA(es))$pca),
                length(sampleNames(es)))
})

test_that("pcaPlot drops NA rows", {
    es <- readGct(system.file("testdata/centers.gct", package="phantasus"))
    pca <- jsonlite::fromJSON(calcPCA(es))
    expect_equal(nrow(exprs(es)), 6)
    expect_equal(nrow(pca$pca), 5)
})

limmaAnalysisSimpleImpl <- function(es, fieldValues, contrast){
    fieldValues <- replace(fieldValues, fieldValues == "", NA)

    es.copy <- es
    es.copy$Comparison <- fieldValues
    pData(es.copy)[,"Comparison"] <- as.factor(pData(es.copy)[,"Comparison"])
    pData(es.copy)[,"Comparison"] <- relevel(pData(es.copy)[,"Comparison"], ref = "Reference")
    fData(es.copy) <- data.frame(row.names = rownames(es.copy))
    es.copy <- es.copy[, !is.na(fieldValues)]

    # Getting rid of check NOTEs
    Comparison=ComparisonA=ComparisonB=NULL

    es.design <- stats::model.matrix(~0 + Comparison, data = pData(es.copy))

    fit <- lmFit(es.copy, es.design)

    A <- NULL; B <- NULL
    fit2 <- contrasts.fit(fit, makeContrasts(ComparisonTarget - ComparisonReference, levels = es.design))
    fit2 <- eBayes(fit2)
    de <- topTable(fit2, adjust.method = "BH", number = Inf)
    de <- de[row.names(fData(es.copy)), ]
    return(de)
}
limmaAnalysisAdvancedImpl <- function(es, designData, contrast){
    ux_designMatrix <- getDesignMatrix(designData)

    es.copy <- es
    fData(es.copy) <- data.frame(row.names = rownames(es.copy))
    colnames(ux_designMatrix) <-  make.names(colnames(ux_designMatrix))
    target_level <- make.names(paste0(contrast[1],contrast[3]))
    reference_level <-  make.names(paste0(contrast[1],contrast[2]))


    fit <- lmFit(es.copy, ux_designMatrix)
    fit2 <- contrasts.fit(fit, makeContrasts(contrasts = paste(reference_level,target_level, sep = "-"), levels = ux_designMatrix))
    fit2 <- eBayes(fit2)
    de <- topTable(fit2, adjust.method = "BH", number = Inf)
    de <- de[row.names(fData(es.copy)), ]
    return(de)
}


#' Differential Expression analysis.
#'
#' \code{limmaAnalysis} performs differential expression analysis
#'     from limma package and returns a ProtoBuf-serialized resulting
#'     de-matrix.
#'
#' @param es ExpressionSet object. It should be normalized for
#'     more accurate analysis.
#'
#' @param fieldValues Vector of comparison values, mapping
#'     categories' names to columns/samples
#'
#' @param version name of the limma analysis implementation. Should be "One-factor design" or "Advanced design"
#'
#' @param contrast a character vector with exactly three elements: the name of a factor in the design formula, the name of the numerator level for the fold change, and the name of the denominator level for the fold change
#'
#' @param designData data.frame with design matrix
#'
#' @return Name of the file containing serialized de-matrix.
#'
#' @import Biobase
#' @import limma
#'
#' @examples
#' \dontrun{
#' data(es)
#' limmaAnalysis(es, fieldValues = c("A", "A", "A", "B", "B"))
#' }
#' @keywords internal
limmaAnalysis <- function (es, fieldValues, version = "One-factor design", contrast =  list('Comparison', 'Target', 'Reference'), designData = NULL) {
    fieldValues <- replace(fieldValues, fieldValues == "", NA)
    de <- NULL
    contrast <- unlist(contrast)
    if (version == "One-factor design" ){
        de <- limmaAnalysisSimpleImpl(es, fieldValues, contrast)
    }
    if (version == "Advanced design"){
        de <- limmaAnalysisAdvancedImpl(es, designData, contrast)
    }
        deDf <- as.data.frame(de)
    toRemove <- intersect(colnames(fData(es)), colnames(deDf))
    fData(es)[, toRemove] <- NULL

    es$Comparison <- fieldValues
    fData(es) <- cbind(fData(es), deDf)
    assign("es", es, envir = parent.frame())

    f <- tempfile(pattern = "de", tmpdir = getwd(), fileext = ".bin")
    writeBin(protolite::serialize_pb(as.list(de)), f)
    jsonlite::toJSON(f)
}

context("Limma Analysis")

test_that("limmaAnalysis finishes with result", {
  load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))
  expect_is(limmaAnalysis(es,
                          fieldValues = c(rep("Target", 3), rep("Reference", 2))), "json")
})

test_that("limmaAnalysis works when there is only one phenotype attribute", {
    load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))
    pData(es) <- pData(es)[, "title", drop=F]
    expect_is(limmaAnalysis(es,
                            fieldValues = c(rep("Target", 3), rep("Reference", 2))), "json")
})


test_that("limmaAnalysisSimpleImpl works", {
    load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))
    de <- limmaAnalysisSimpleImpl(es, fieldValues = c(rep("Target", 2), rep("Reference", 2), NA))

    expect_equal(de$logFC[1], mean(exprs(es)[1, 1:2]) - mean(exprs(es)[1, 3:4]))
})

# Limma works for full dataset only
#test_that("limmaAnalysisImpl works for subsamples", {
#    load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))
#    de1 <- limmaAnalysisImpl(es, rows=seq_len(nrow(es)), columns=seq_len(ncol(es)),
#                            fieldValues = c(rep("A", 2), NA, rep("B", 2)))
#
#    de2 <- limmaAnalysisImpl(es, rows=seq_len(nrow(es)), columns=c(1:2, 4:5),
#                             fieldValues = c(rep("A", 2), rep("B", 2)))
#    expect_equal(de1$t, de2$t)
#})

#' Create ExpressionSet.
#'
#' \code{createES} function produces an ExpressionSet object from given data,
#'     and exports it to global scope.
#'
#' @param data Gene expression matrix.
#'
#' @param pData Matrix with phenotypical data.
#'
#' @param varLabels Names of phenoData columns.
#'
#' @param fData Matrix with feature data.
#'
#' @param fvarLabels Names of featureData columns.
#'
#' @param eData List with experimentData
#'
#' @return produced ExpressionSet object
#'
#' @import Biobase
#' @importFrom methods new
#'
#' @examples
#' \dontrun{
#' data <- matrix(1:15, 5, 3)
#' pData <- c("A", "B", "C")
#' varLabels <- "cat"
#' fData <- c("p", "r", "s", "t", "u")
#' fvarLabels <- "id"
#' eData <- list(name="", lab="", contact="", title="", url="", other=list(), pubMedIds="")
#' createES(data, pData, varLabels, fData, fvarLabels, eData)
#' }
#' @keywords internal
createES <- function(data, pData, varLabels, fData, fvarLabels, eData) {
    phenoData <- AnnotatedDataFrame(data.frame(pData, stringsAsFactors = FALSE))
    varLabels(phenoData) <- varLabels

    featureData <- AnnotatedDataFrame(data.frame(fData, stringsAsFactors = FALSE))
    varLabels(featureData) <- fvarLabels

    ed <- new ("MIAME",
              name=eData$name,
              lab=eData$lab,
              title=eData$title,
              contact=eData$contact,
              pubMedIds=eData$pubMedIds,
              url=eData$url,
              other=eData$other)

    es <- ExpressionSet(assayData = data,
                        phenoData = phenoData,
                        featureData = featureData,
                        experimentData = ed)
    assign("es", es, envir = parent.frame())
    es
}

context("Create ExpressionSet")
library(Biobase)

test_that("createES finishes with result", {
  load(file = system.file("testdata/GSE27112-GPL6103.rda", package="phantasus"))
  expect_is(createES(data = exprs(es), pData = pData(es),
                     fData = fData(es),
                     eData = list(name="", lab="", contact="", title="", url="", other=list(), pubMedIds=""),
                     varLabels = varLabels(es),
                     fvarLabels = fvarLabels(es)), "ExpressionSet")
})

#!/bin/bash

PHANTASUS_JS_DIR="$1"

if [ -z "${PHANTASUS_JS_DIR}" ]; then
    echo "Please specify git dir with phantasus.js"
    exit 1
    # PHANTASUS_JS_DIR=inst-raw/phantasus.js
fi

PHANTASUS_JS_DIR=$(readlink -f "${PHANTASUS_JS_DIR}")

echo "Using phantasus.js from ${PHANTASUS_JS_DIR}"

grunt --base "${PHANTASUS_JS_DIR}" dist

git -C "${PHANTASUS_JS_DIR}" log -1 > ./inst/www/phantasus.js/commit

cp ${PHANTASUS_JS_DIR}/index.html ./inst/www/phantasus.js/
cp ${PHANTASUS_JS_DIR}/js/phantasus.js ./inst/www/phantasus.js/js/
cp ${PHANTASUS_JS_DIR}/js/phantasus-external-*.min.js ./inst/www/phantasus.js/js/
cp -r ${PHANTASUS_JS_DIR}/jasmine ./inst/www/phantasus.js/
cp ${PHANTASUS_JS_DIR}/my.conf.js ./inst/www/phantasus.js/
cp ${PHANTASUS_JS_DIR}/package.json ./inst/www/phantasus.js/
cp -r ${PHANTASUS_JS_DIR}/css/images ./inst/www/phantasus.js/css/
cp -r ${PHANTASUS_JS_DIR}/css/phantasus-latest.min.css ./inst/www/phantasus.js/css/

#!/bin/bash

RSCRIPT_PID=""

_term() {
  trap ERR
  echo "Caught signal!"
  nginx -s stop
  if [ -n "$RSCRIPT_PID" ]; then
    kill -2 "$RSCRIPT_PID" 2>/dev/null || true  # SIGINT → R calls q() → CleanupParallel kills children
  fi
}

_term_config(){
    echo "Failed to complete setup! Bad configuration file or user: $OCPU_USER doesn't have permissions to write cache or configuration file."
    exit 1
}

trap _term SIGINT SIGTERM

trap _term_config ERR
set -eE

chown -R $OCPU_USER /var/log/nginx
chown -R $OCPU_USER /var/lib/nginx
chown -R $OCPU_USER /var/cache/nginx

touch /run/nginx.pid
chown $OCPU_USER /run/nginx.pid

# Let's not change user files ownership?
# chown -R $OCPU_USER $R_USER_CONFIG_DIR/R/phantasus

# editing config under root
R -e "phantasus:::createDockerConf()"
# then setup under ocpu user
gosu $OCPU_USER R -e "phantasus::setupPhantasus()" #|| _term_config

# Generate nginx upstream block from internal_ports in user config.
# Falls back to ports 8001-8010 if the key is absent.
UPSTREAM_PORTS=$(gosu $OCPU_USER Rscript --vanilla -e "
  tryCatch({
    ports <- unlist(phantasus::getPhantasusConf('internal_ports'))
    if (is.null(ports) || length(ports) == 0) stop('internal_ports not configured')
    cat(paste(ports, collapse=' '), '\n')
  }, error = function(e) {
    cat('8001 8002 8003 8004 8005 8006 8007 8008 8009 8010\n')
  })
" 2>/dev/null)

{
  echo "upstream phantasus_backend {"
  echo "    least_conn;"
  for p in $UPSTREAM_PORTS; do
    echo "    server localhost:$p;"
  done
  echo "}"
} > /etc/nginx/conf.d/upstream.conf

gosu $OCPU_USER nginx

gosu $OCPU_USER Rscript /var/phantasus/serve-managed.R &
RSCRIPT_PID=$!


color() {
    STARTCOLOR="\e[38;5;$2";
    ENDCOLOR="\e[0m";
    export "$1"="$STARTCOLOR%s"
}
color darkred 196m
color pinkish 211m
color blue 27m
color lightblue 68m
color coral 203m
color lessblue 26m
color normal 0m

printf $darkred "#### "
printf $pinkish "++++ "
printf $blue'\n' "----"
printf $darkred "#### "
printf $pinkish "++++ "
printf $blue'\n' "----"
printf $darkred "#### "
printf $pinkish "++++ "
printf $blue'\n' "----"

printf $lightblue "====      "
printf $coral "&&&&"
printf "\e[0m%b"'\n' "      _____    _                       _                                   _                         "
printf $lightblue "====      "
printf $coral "&&&&"
printf "\e[0m%s\n" "     |  __ \  | |                     | |                                 (_)                        "
printf $lightblue   "====      "
printf $coral"&&&&     "
printf "\e[0m%s\n" "| |__) | | |__     __ _   _ __   | |_    __ _   ___   _   _   ___     _   ___     _   _   _ __  "
printf $coral "&&&& "
printf $darkred "#### "
printf $lessblue "----     "
printf "\e[0m%s\n" "|  ___/  | '_ \   / _\` | | '_ \  | __|  / _\` | / __| | | | | / __|   | | / __|   | | | | | '_ \ "
printf $coral "&&&& "
printf $darkred "#### "
printf $lessblue "----     "
printf "\e[0m%s\n" "| |      | | | | | (_| | | | | | | |_  | (_| | \__ \ | |_| | \__ \   | | \__ \   | |_| | | |_) |"
printf $coral "&&&& "
printf $darkred "#### "
printf $lessblue "----     "
printf "\e[0m%s\n" "|_|      |_| |_|  \__,_| |_| |_|  \__|  \__,_| |___/  \__,_| |___/   |_| |___/    \__,_| | .__/ "
printf $blue "----"
printf "\e[0m%s\n" "                                                                                                        | |    "
printf $blue "----"
printf "\e[0m%s\n" "                                                                                                        |_|    "
printf $blue "----"
printf "\e[0m%s\n" ""


wait $RSCRIPT_PID

#!/bin/bash
PHANTASUS_VERSION=$(grep -r 'Version:' ./DESCRIPTION | awk {'print $2'})
PHANTASUS_BUILD=$(git rev-parse HEAD)
printf "window.PHANTASUS_VERSION='$PHANTASUS_VERSION';\nwindow.PHANTASUS_BUILD='none-${PHANTASUS_BUILD::10}';\n" | tee inst/www/phantasus.js/RELEASE.js

#!/usr/bin/env bash
# test-serve-managed.sh
#
# Integration test for serve-managed.R + nginx.
#
# Tests the full stack: serve-managed.R workers + nginx upstream routing.
# Intended to run in the Vagrant VM or a Docker container on Linux.
# Not part of devtools::test().
#
# Exit code: 0 = all assertions passed, non-zero = first failure.
# Each step prints PASS/FAIL so failures are easy to locate.
#
# Usage:
#   bash phantasus/inst/test-serve-managed.sh

set -euo pipefail

# ── configuration ──────────────────────────────────────────────────────────────

N_WORKERS=3
BASE_PORT=8991
NGINX_PORT=8990
POLL_URL_PATH="phantasus/ocpu/library/base/R/Sys.time"
GC_URL_PATH="phantasus/ocpu/library/base/R/Sys.time"
NGINX_CONF_DIR=$(mktemp -d)
NGINX_LOG_DIR=$(mktemp -d)
NGINX_PID_FILE="$NGINX_CONF_DIR/nginx.pid"
CFG_DIR=$(mktemp -d)/R/phantasus
CACHE_ROOT=$(mktemp -d)
MANAGER_LOG=$(mktemp)
MANAGER_PID=""

PORTS=()
for i in $(seq 0 $((N_WORKERS - 1))); do
    PORTS+=($((BASE_PORT + i)))
done

# ── helpers ────────────────────────────────────────────────────────────────────

pass() { printf "PASS  %s\n" "$*"; }
fail() { printf "FAIL  %s\n" "$*"; cleanup; exit 1; }

cleanup() {
    set +e
    if [ -n "$MANAGER_PID" ] && kill -0 "$MANAGER_PID" 2>/dev/null; then
        kill -INT "$MANAGER_PID" 2>/dev/null
        sleep 2
        kill -TERM "$MANAGER_PID" 2>/dev/null
    fi
    nginx -s stop -c "$NGINX_CONF_DIR/nginx.conf" 2>/dev/null || true
    for p in "${PORTS[@]}"; do
        fuser -k "${p}/tcp" 2>/dev/null || true
    done
    fuser -k "${NGINX_PORT}/tcp" 2>/dev/null || true
}
trap cleanup EXIT

wait_http_200() {
    local url=$1 timeout_s=$2 label=${3:-"$url"}
    local deadline=$(( $(date +%s) + timeout_s ))
    while [ "$(date +%s)" -lt "$deadline" ]; do
        if curl -sf --max-time 3 "$url" >/dev/null 2>&1; then
            return 0
        fi
        sleep 1
    done
    fail "timeout waiting for HTTP 200: $label"
}

count_5xx() {
    local url=$1 n=$2
    local failures=0
    for i in $(seq 1 "$n"); do
        code=$(curl -s -o /dev/null -w "%{http_code}" --max-time 5 "$url" 2>/dev/null || echo "000")
        if [[ "$code" == 5* ]]; then
            failures=$((failures + 1))
        fi
    done
    echo "$failures"
}

# Return the pid of the R process listening on a port, or empty string.
pid_on_port() {
    local port=$1
    ss -tlnp "sport = :${port}" 2>/dev/null \
        | grep -oP 'pid=\K[0-9]+' \
        | head -1 || true
}

# ── setup ──────────────────────────────────────────────────────────────────────

echo "=== serve-managed integration test ==="
echo "workers=${N_WORKERS}  ports=${PORTS[*]}  nginx_port=${NGINX_PORT}"
echo ""

mkdir -p "$CFG_DIR"
mkdir -p "$CACHE_ROOT"/{geo,annotationdb,fgsea,counts}

{
    echo "default:"
    echo "  host: \"0.0.0.0\""
    echo "  preloaded_dir: NULL"
    echo "  static_root: \"\""
    echo "  cache_root: \"$CACHE_ROOT\""
    echo "  internal_ports:"
    for p in "${PORTS[@]}"; do
        echo "    - $p"
    done
    echo "  cache_folders:"
    echo "    geo_path: \"$CACHE_ROOT/geo\""
    echo "    annot_db: \"$CACHE_ROOT/annotationdb\""
    echo "    fgsea_pathways: \"$CACHE_ROOT/fgsea\""
    echo "    rnaseq_counts: \"$CACHE_ROOT/counts\""
} > "$CFG_DIR/user.conf"

SCRIPT=$(Rscript --vanilla -e \
    "cat(system.file('serve-managed.R', package='phantasus'))" 2>/dev/null)
if [ -z "$SCRIPT" ] || [ ! -f "$SCRIPT" ]; then
    fail "serve-managed.R not found via system.file() — is phantasus installed?"
fi

{
    echo "worker_processes 1;"
    echo "pid $NGINX_PID_FILE;"
    echo "error_log $NGINX_LOG_DIR/error.log warn;"
    echo "events { worker_connections 64; }"
    echo "http {"
    echo "    access_log $NGINX_LOG_DIR/access.log;"
    echo "    upstream phantasus_backend {"
    echo "        least_conn;"
    for p in "${PORTS[@]}"; do
        echo "        server localhost:$p;"
    done
    echo "    }"
    echo "    server {"
    echo "        listen $NGINX_PORT;"
    echo "        location / {"
    echo "            proxy_pass http://phantasus_backend;"
    echo "        }"
    echo "    }"
    echo "}"
} > "$NGINX_CONF_DIR/nginx.conf"

# ── step 1: start serve-managed.R ─────────────────────────────────────────────

echo "--- Step 1: start serve-managed.R ($N_WORKERS workers)"

R_USER_CONFIG_DIR=$(dirname "$(dirname "$CFG_DIR")") \
R_CONFIG_ACTIVE=default \
PHANTASUS_MAX_WORKER_AGE_S=10 \
    Rscript "$SCRIPT" > "$MANAGER_LOG" 2>&1 &
MANAGER_PID=$!

for p in "${PORTS[@]}"; do
    wait_http_200 "http://localhost:${p}/${POLL_URL_PATH}" 60 "port $p"
done
pass "all $N_WORKERS workers started and responding"

# ── step 2: start nginx ────────────────────────────────────────────────────────

echo "--- Step 2: start nginx"
nginx -c "$NGINX_CONF_DIR/nginx.conf"
wait_http_200 "http://localhost:${NGINX_PORT}/${POLL_URL_PATH}" 10 "nginx"
pass "nginx started and proxying to workers"

# ── step 3: 20 healthy requests through nginx ──────────────────────────────────

echo "--- Step 3: 20 requests through nginx (expect all 200)"
failures=$(count_5xx "http://localhost:${NGINX_PORT}/${POLL_URL_PATH}" 20)
if [ "$failures" -ne 0 ]; then
    fail "Step 3: $failures/20 requests returned 5xx before gc test"
fi
pass "Step 3: 0/20 requests failed (pre-gc baseline)"

# ── step 4: wait for max_age replacement on first worker port ─────────────────
#
# Workers are started with MAX_AGE=30 s (random factor 1-2x means exit
# within 30-60 s).  We poll the manager log for the replacement message
# and confirm the port comes back up.

echo "--- Step 4: wait for max_age replacement on port $BASE_PORT"
target_port=${PORTS[0]}
replaced=false
deadline=$(( $(date +%s) + 120 ))

while [ "$(date +%s)" -lt "$deadline" ]; do
    sleep 2
    if grep -q "port=${target_port}.*exited.*replacing\|exited.*replacing.*port=${target_port}" \
            "$MANAGER_LOG" 2>/dev/null; then
        replaced=true
        break
    fi
done

if [ "$replaced" = "false" ]; then
    fail "Step 4: no replacement seen on port $target_port within 120 s"
fi
pass "Step 4: max_age triggered replacement on port $target_port"

# ── step 5: wait for replacement worker ───────────────────────────────────────

echo "--- Step 5: wait for replacement worker on port $target_port"
wait_http_200 "http://localhost:${target_port}/${POLL_URL_PATH}" 60 \
    "replacement worker on port $target_port"
pass "Step 5: replacement worker on port $target_port is responding"

# ── step 6: 20 more requests confirm stable state ─────────────────────────────

echo "--- Step 6: 20 more requests post-replacement"
failures=$(count_5xx "http://localhost:${NGINX_PORT}/${POLL_URL_PATH}" 20)
if [ "$failures" -ne 0 ]; then
    fail "Step 6: $failures/20 requests failed after replacement"
fi
pass "Step 6: 0/20 requests failed post-replacement"

# ── step 7: SIGINT shutdown ────────────────────────────────────────────────────

echo "--- Step 7: SIGINT shutdown"
kill -INT "$MANAGER_PID"

deadline=$(( $(date +%s) + 15 ))
while [ "$(date +%s)" -lt "$deadline" ]; do
    if ! kill -0 "$MANAGER_PID" 2>/dev/null; then
        break
    fi
    sleep 1
done
if kill -0 "$MANAGER_PID" 2>/dev/null; then
    fail "Step 7: manager still alive 15s after SIGINT"
fi
pass "Step 7: manager exited after SIGINT"
MANAGER_PID=""

# Poll for up to 15 s — workers in httpuv::service() may take a few
# seconds to process SIGTERM and release their sockets.
ports_free=false
deadline=$(( $(date +%s) + 15 ))
while [ "$(date +%s)" -lt "$deadline" ]; do
    all_free=true
    for p in "${PORTS[@]}"; do
        if ss -tlnp "sport = :$p" | grep -q LISTEN; then
            all_free=false
            break
        fi
    done
    if [ "$all_free" = "true" ]; then
        ports_free=true
        break
    fi
    sleep 1
done
if [ "$ports_free" = "false" ]; then
    for p in "${PORTS[@]}"; do
        if ss -tlnp "sport = :$p" | grep -q LISTEN; then
            echo "  port $p still bound"
        fi
    done
    fail "Step 7: some worker ports still bound after shutdown"
fi
pass "Step 7: all worker ports freed after shutdown"

# ── summary ────────────────────────────────────────────────────────────────────

echo ""
echo "=== All steps passed ==="

---
title: "Using Phantasus application"
date: "`r Sys.Date()`"
output:  
    BiocStyle::html_document:
        toc_float: true
vignette: >
  %\VignetteIndexEntry{Using phantasus application}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

This tutorial describes Phantasus -- a web-application for visual and interactive gene expression analysis.
Phantasus is based on 
[Morpheus](https://software.broadinstitute.org/morpheus/) -- a web-based software
for heatmap visualisation and analysis, which was integrated with an R environment via [OpenCPU API](https://www.opencpu.org/). 

The main object in Phantasus is a gene expression matrix.
It can either be uploaded from a local text or Excel file 
or loaded from Gene Expression Omnibus (GEO) database by the series identifier
(both microarray and RNA-seq datasets are supported).
Aside from basic visualization and filtering methods as implemented in Morpheus,
R-based methods such as k-means clustering, principal component analysis, 
differential expression analysis with limma package are supported.  

In this vignette we show example usage of Phantasus for analysis of public gene
expression data from GEO database.
It starts from loading data, normalization and filtering outliers, 
to doing differential gene expression analysis and downstream analysis.

# Example workfow for analysing gene expression changes in macrophage activation

To illustrate the usage of Phantasus let us consider public dataset from Gene Expression Omnibus (GEO) database
[GSE53986](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE53986). 
This dataset contains data from experiments, where bone marrow derived macrophages were
treated with three stimuli: LPS, IFNg and combined LPS+INFg.

## Starting application

The simplest way to try Phantasus application is to go to web-site
https://alserglab.wustl.edu/phantasus
where the latest version is deployed.

Alternatively, Phantaus can be start locally using the corresponding R package:

```{r,eval=FALSE}
library(phantasus)
servePhantasus()
```

This command runs the application with the default parameters,
opens it in the default browser (from `browser` option) 
with address http://0.0.0.0:8000. 
The starting screen should appear:

```{r, out.width = "750px", echo = FALSE}
knitr::include_graphics("images/start_screen.jpg")
```

## Preparing the dataset for analysis

**Opening the dataset**

Let us open the dataset. To do this, select _GEO Datasets_ option
in _Choose a file..._ dropdown menu. There, a text field will appear
where `GSE53986` should be entered. Clicking the _Load_ button
(or pressing _Enter_ on the keyboard) will start the loading.
After a few seconds, the corresponding heatmap should appear. 

```{r, out.width = "750px", echo = FALSE}
knitr::include_graphics("images/dataset_loaded.jpg")
```

On the heatmap, the rows correspond to genes (or microarray probes).
The rows are annotated with _Gene symbol_ and _Gene ID_ annotaions
(as loaded from GEO database). 
Columns correspond to samples. 
They are annotated with titles, GEO sample accession identifiers and treatment field.
The annotations, such as treatment, are loaded from user-submitted GEO annotations 
(they can be seen, for example, in _Charateristics_
secion at https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSM1304836).
We note that not for all of the datasets in GEO such proper annotations are supplied.

**Adjusting expression values**

By hovering at heatmap cell, gene expression values can be viewed.
The large values there indicate that the data is not log-scaled,
which is important for most types of gene expression analysis.


```{r, out.width = "500px", echo = FALSE}
knitr::include_graphics("images/huge_value.jpg")
```

For the proper further analysis it is recommended to normalize the 
matrix. To normalize values go to _Tools/Adjust_ menu and
check _Log 2_ and _Quantile normalize_ adjustments.

```{r, out.width = "500px", echo = FALSE}
knitr::include_graphics("images/adjust_tool.jpg")
```

The new tab with adjusted values will appear. All operations that modify gene expression
matrix (such as adjustment, subsetting and several others) create a new tab. This allows 
to revert the operation by going back to one of the previous tabs.

**Removing duplicate genes**

Since the dataset is obtained with a mircroarray, a single gene can be represented by
several probes. This can be seen, for example, by sorting rows 
by _Gene symbol_ column (one click on column header), entering `Actb` in the search 
field and going to the first match by clicking down-arrow next to the field. 
There are five probes corresponding to Actb gene in the considered microarray.

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/duplicates.jpg")
```

To simplify the analysis it is better to have one row per gene 
in the gene expression matrix. 
One of the easiest ways is to chose only one row that has
the maximal median level of expression
across all samples. Such method removes the noise of lowly-expressed probes.
Go to _Tools/Collapse_ and choose _Maximum Median Probe_ as the method and 
_Gene ID_ as the collapse field. 

```{r, out.width = "500px", echo = FALSE}
knitr::include_graphics("images/collapse_tool.jpg")
```

The result will be shown in a new tab. 


**Filtering lowly-expressed genes**

Additionally, lowly-epxressed genes can be filtered explicitly. It helps to reduce noise and increase
power of downstream analysis methods. 

First, we calculate mean expression of each gene using 
_Tools/Create Calculated Annotation_ menu. 
Select `Mean` operation, optionally enter a name for the resulting column,
and click _OK_. The result will appear as an additional column in row annotations.

```{r, out.width = "500px", echo = FALSE}
knitr::include_graphics("images/calculated_annotation_tool.jpg")
```


```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/calculated_annotation_loaded.jpg")
```

Now this annotation can be used to filter genes.
Open _Tools/Filter_ menu. Click _Add_ to add a new filter. Choose `mean_expression`
as a _Field_ for filtering. Then press _Switch to top filter_ button
and input the number of genes to keep. A good choice for a typical mammalian dataset
is to keep around 10--12 thousand most expressed genes. Filter is applied automatically,
so after closing the dialog with _Close_ button only the genes passing the filter 
will be displayed.


```{r, out.width = "500px", echo = FALSE}
knitr::include_graphics("images/filter_tool.jpg")
```

```{r, out.width = "500px", echo = FALSE}
knitr::include_graphics("images/filter_tool_result.jpg")
```

It is more convenient to extract these genes into a new tab. 
For this, select all genes (click on any gene and press _Ctrl+A_) and 
use _Tools/New Heat Map_ menu (or press _Ctrl+X_).

Now you have the tab with a fully prepared dataset for the further analysis. 
To easily distinguish it from other tabs, you can rename it
by right click on the tab and choosing _Rename_ option. Let us rename it to `GSE53986_norm`.

It is also useful to save the current result to be able to return to it later.
In order to save it use _File/Save Dataset_ menu. Enter an appropriate file name (e.g. `GSE53986_norm`)
and press `OK`. A file in text [GCT format](https://software.broadinstitute.org/cancer/software/gsea/wiki/index.php/Data_formats#GCT:_Gene_Cluster_Text_file_format_.28.2A.gct.29)
will be downloaded.

## Exploring the dataset

**PCA Plot**

One of the ways to asses quality of the dataset is to use
principal component analysis (PCA) method. This can be done
using _Tools/Plots/PCA Plot_ menu.

```{r, out.width = "550px", echo = FALSE}
knitr::include_graphics("images/pcaplot_tool_clean.png")
```

You can customize color, size and labels of points on the chart 
using values from annotation. Here we set color to come from
_treatment_ annotation.


```{r, out.width = "550px", echo = FALSE}
knitr::include_graphics("images/pcaplot_tool_finished.png")
```

It can be seen that in this dataset the first replicates in each condition 
are outliers.

**K-means clustering**

Another useful dataset exploration tool is k-means clustering.
Use _Tools/Clustering/k-means_ to cluster genes into 16 clusters.

```{r, out.width = "500px", echo = FALSE}
knitr::include_graphics("images/kmeans_tool.jpg")
```

Afterwards, rows can be sorted by _clusters_ column. By using 
menu _View/Fit to window_ one can get a "bird's-eye view" on the dataset.
Here also one can clearly see outlying samples.

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/kmeans_result.jpg")
```

**Hierarchical clustering**

_Tool/Hierarchical clustering_ menu can be used to cluster 
samples and highlight outliers (and 
concordance of other samples) even further. 

```{r, out.width = "500px", echo = FALSE}
knitr::include_graphics("images/hierarchical_tool.jpg")
```

**Filtering outliers**

Now, when outliers are confirmed and easily viewed with the dendrogram 
from the previous step, you can select the good samples 
and extract them into another heatmap (by clicking _Tools/New Heat Map_ or pressing _Ctrl+X_).

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/good_samples.jpg")
```

## Differential gene expression

**Apllying _limma_ tool**

Differential gene expression analysis can be carried out
with _Tool/Diffential Expression/limma_ menu. 
Choose _treatment_ as a _Field_, with
_Untreated_ and _LPS_ as classes. Clicking _OK_
will call differential gene expression analysis method 
with [*limma*](https://bioconductor.org/packages/limma) R package.

```{r, out.width = "500px", echo = FALSE}
knitr::include_graphics("images/limma_tool.jpg")
```

The rows can be ordered by decreasing *t*-statistic column to see which genes are the most
up-regulated upon LPS treatment. 

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/limma_results.png")
```

**Pathway analysis with FGSEA**

**Note**: Self-hosted Phantasus requires additional steps to make **FGSEA** work properly:
see section \@ref(serving-phantasus).

The results of differential gene expression can be used for pathway enrichment analysis
with _FGSEA_ tool.

Open _Tools/Pathway Analysis/Perform FGSEA_, then select Pathway database, which corresponds specimen used in dataset
(Mus Musculus in this example), ranking column and column with ENTREZID or Gene IDs.

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/fgsea_tool.png")
```

Clicking OK will open new tab with pathways table.

```{r, out.width = "800px", echo = FALSE}
knitr::include_graphics("images/fgsea_result.png")
```

Clicking on table row will provide additional information on pathway: pathway name, genes in pathway, leading edge.
You can save result of analysis in `TSV` format.

# Advanced usage

## Gene identifiers conversion with AnnotationDB

**Note**: Self-hosted Phantasus requires additional steps to make *AnnotationDB* work properly:
see section \@ref(serving-phantasus).

AnnotationDB provides an opportunity to convert annotation columns in dataset. Open _Tools/Annotate/Annotate Rows/From Database_. Example: converting `ENTREZID` column in GSE53986 to `ENSEMBL`. Select _org.Mm.hg.sqlite_ specimen database with 
*Source column* -- `Gene ID`,
*Source column type* -- `ENTREZID`, and *Result column type* -- `ENSEMBL`. 

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/annotation_tool.png")
```

Clicking *OK* will append a new column with correspoding `ENSEMBL` identifier to the gene annotation.

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/annotation_result.png")
```

## Pathway analysis with Enrichr

The results of differential gene expression can be used, for example,
for pathway enrichment analysis with online tools such as
[MSigDB](software.broadinstitute.org/gsea/msigdb/compute_overlaps.jsp)
or [Enrichr](http://amp.pharm.mssm.edu/Enrichr/).

For ease of use Enrichr is integrated into Phantasus. 
Open _Tools/Pathway Analysis/Submit to Enrichr_ menu and select about top 200 genes up-regulated on LPS.
Also select _Gene symbol_ as the column with gene symbols.

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/enrichr.png")
```


Clicking _OK_ will result in a new browser tab with Enrichr being opened with results
of pathway enrichment analysis.

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/enrichr_result.png")
```


## Metabolic network analysis with Shiny GAM

Another analysis integrated into Phantasus is metabolic network analysis
with [Shiny GAM](https://artyomovlab.wustl.edu/shiny/gam/).
After differential expression you can submit 
the table with _limma_ results using _Tools/Pathway Analysis/Sumbit to Shiny GAM_ menu.

```{r, out.width = "300px", echo = FALSE}
knitr::include_graphics("images/shiny_gam.jpg")
```

After successful submission, a new browser tab will be opened with _Shiny GAM_ interface
and the submitted data.

```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/shiny_gam_result.jpg")
```

## GSEA enrichment plot

The results of pathway analysis with _FGSEA_ can be used for generating enrichment plot.
Clicking on table row will provide you a list of _Gene IDs_. 
```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/fgsea_genes.png")
```
Copy the list, switch back to dataset tab and paste to search field. 
```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/fgsea_search.png")
```
This will select the corresponding genes in dataset. Then proceed to _Tools/Plots/GSEA plot_
```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/GSEA_plot.png")
```
Then use _Export to SVG_ button to save plot in `.svg` format.
Also there are options to control orientation of the plot, the ranking column 
and annotation.

## Link sharing

Current progress on the dataset can be shared via link.
Go to _File/Get link to a dataset_. 
This will provide you a link, to a current dataset.
```{r, out.width = "600px", echo = FALSE}
knitr::include_graphics("images/link.png")
```
**Note**: The following warning about 30 days to live is not applicable to self-hosted Phantasus

## Loading dataset options

There are three ways to upload a dataset into application:

- As a file from:
    - computer;
    - URL;
    - Dropbox.
- By GEO identifier:
    - with _Open file_ interface;
    - with adding parameter `geo` to the link (_e.g._, http://localhost:8000/?geo=GSE27112).
- By an identifier of a dataset saved on the server (see section \@ref(preloaded-datasets))
    - with _Saved on server datasets_
    - with adding parameter `preloaded` to the link (_e.g._, http://localhost:8000/?preloaded=fileName)
    
You can either open the dataset from the main page, or if you are already looking at some datasets and don't want to lose your progress you can use _File/Open_ (_Ctrl+O_), choose _Open dataset in new tab_ and then select the open option.

```{r, out.width="500px", echo = FALSE}
knitr::include_graphics("images/open_file_tool.jpg")
```

# Serving Phantasus

Some of Phantasus features require additional set up.

## Options for `servePhantasus`

You can customise serving of the application by specifying following parameters:

- `host` and `port` (by default: `"0.0.0.0"` and `8000`);
- `cacheDir` (by default: `tempdir()`) -- directory where downloaded datasets will be saved and reused in later sessions;
- `preloadedDir` (by default: `NULL`) -- directory with `ExpressionSet` objects encoded in rda-files, that can be quickly loaded to application by name (see section \@ref(preloaded-datasets));
- `openInBrowser` (by default `TRUE`).


## Preloaded datasets 

Preloaded datasets is a feature that allows quick access to frequently-accessed datasets
or to share them inside the research group.

To store dataset on a server, on nead to save list `ess` of `ExpressionSet` objects
into an RData file with `.rda` extension into a directory as specified in `servePhantasus`.

Let us preprocess and save `GSE14308` dataset:

```{r, message=FALSE, cache=TRUE, eval=FALSE}
library(GEOquery)
library(limma)
gse14308 <- getGEO("GSE14308", AnnotGPL = TRUE)[[1]]
gse14308$condition <- sub("-.*$", "", gse14308$title)
pData(gse14308) <- pData(gse14308)[, c("title", "geo_accession", "condition")]
gse14308 <- gse14308[, order(gse14308$condition)]

fData(gse14308) <- fData(gse14308)[, c("Gene ID", "Gene symbol")]
exprs(gse14308) <- normalizeBetweenArrays(log2(exprs(gse14308)+1), method="quantile")

ess <- list(GSE14308_norm=gse14308)

preloadedDir <- tempdir()

save(ess, file=file.path(preloadedDir, "GSE14308_norm.rda"))
```

Next we can serve Phantasus with set `preloadedDir` option:

```{r, eval=F}
servePhantasus(preloadedDir=preloadedDir)
```

There you can either put `GSE14308_norm` name when using open option _Saved on server datasets_ or just 
open by specifying the name in URL: http://localhost:8000/?preloaded=GSE14308_norm.

```{r, out.width="650px", echo = FALSE}
knitr::include_graphics("images/gse14308_norm.png")
```

## Support for RNA-seq datasets

Phantasus supports loading RNA-seq datasets from GEO using 
gene expression counts as computed by [ARCHS4 project](http://amp.pharm.mssm.edu/archs4/index.html).
To make it work one need to download gene level expression from 
the [Download](http://amp.pharm.mssm.edu/archs4/download.html) section. The
downloaded files `human_matrix.h5` and `mouse_matrix.h5` should be placed 
into Phantasus cache under archs4 folder.
Or you can call
```{r, eval=F}
updateARCHS4(cacheDir=cacheDir)
```


## Pathway database for FGSEA

*FGSEA* requires pathway database in `.rds` files under `<cacheDir>/fgsea` folder.
Pathway database is an `.rds` file containing dataframe with columns: `geneID`, `pathName`, `geneSymbol`. You can see an example dataframe by entering:
```{r, eval=TRUE}
data("fgseaExample", package="phantasus")
head(fgseaExample)
```

## Annotation database for AnnotationDB tool

*AnnotationDB* tool requires annotation databases 
under `<cacheDir>/annotationdb` folder. For example you can get 
[Mus Musculus](https://bioconductor.org/packages/release/data/annotation/html/org.Mm.eg.db.html) database package from Bioconductor, extract `org.Mm.eg.sqlite`, and put it to `<cacheDir>/annotationdb` folder.


# Feedback

You can see known issues and submit yours at GitHub:
(https://github.com/ctlab/phantasus/issues)

# Acknowledgments

The authors are very thankful to Joshua Gould for developing Morpheus.

# Citation

Please cite us as: 

Kleverov M., Zenkova D., Kamenev V., Sablina R., Artyomov M., Sergushichev A. Phantasus:
web-application for visual and interactive gene expression analysis.
https://https://www.biorxiv.org/content/10.1101/2022.12.10.519861 doi:
10.1101/2022.12.10.519861

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calcPCA.R
\name{calcPCA}
\alias{calcPCA}
\title{Principal Component Analysis.}
\usage{
calcPCA(es, replacena = "mean")
}
\arguments{
\item{es}{an ExpressionSet object, should be normalized}

\item{replacena}{method for replacing NA values (mean by default)}
}
\value{
json with full description of the plot for plotly.js
}
\description{
\code{calcPCA} calculates PCA-matrix for the given ExpressionSet
    and returns this matrix encoded to JSON.
}
\examples{
\dontrun{
data(es)
calcPCA(es)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadPreloaded.R
\name{loadPreloaded}
\alias{loadPreloaded}
\title{Load GEO Dataset.}
\usage{
loadPreloaded(name)
}
\arguments{
\item{name}{String, containing filename. Assuming
that in the directory with preloaded files \code{preloadedDir}
exists file \code{filename.rda} with list of ExpressionSets \code{ess}.}
}
\value{
File with ProtoBuf-serialized ExpressionSet-s
    that were loaded from specified file.
}
\description{
\code{loadPreloaded} returns the file with serialized ExpressionSets using
    ProtoBuf, that were preloaded on server.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{read.gct}
\alias{read.gct}
\title{Reads ExpressionSet from a GCT file.
Function is deprecated, please use phantasusLite:::readGct() instead}
\usage{
read.gct(...)
}
\arguments{
\item{...}{parameters for phantasusLite:::readGct() call}
}
\value{
ExpressionSet object
}
\description{
Reads ExpressionSet from a GCT file.
Function is deprecated, please use phantasusLite:::readGct() instead
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{updateDEE2meta}
\alias{updateDEE2meta}
\title{Update DEE2 meta files}
\usage{
updateDEE2meta(
  destDir = file.path(getPhantasusConf("cache_folders")$rnaseq_counts, "dee2")
)
}
\arguments{
\item{destDir}{path to directory with DEE2 .h5 files.}
}
\description{
Creates \code{meta.txt} file, which describes typical dee2 files.
}
\details{
This function produces very specific "hardcoded" \code{meta.txt} file for dee2 counts colletction.
See \code{\link{validateCountsCollection}} for more common information and \code{meta.txt}  file structure
}
\seealso{
\code{\link{validateCountsCollection}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinyGAM.R
\name{shinyGAMAnalysis}
\alias{shinyGAMAnalysis}
\title{Constructs data frame with gene annotations and submits it into Shiny GAM web-server}
\usage{
shinyGAMAnalysis(es)
}
\arguments{
\item{es}{Expression set object}
}
\value{
URL for Shiny GAM
}
\description{
Constructs data frame with gene annotations and submits it into Shiny GAM web-server
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{loadGEO}
\alias{loadGEO}
\title{Load GEO Dataset.}
\usage{
loadGEO(name, type = NA)
}
\arguments{
\item{name}{String, containing GEO identifier of the dataset.
It should start with 'GSE' or 'GDS' and can include exact GPL
to annotate dataset, separated with dash ('-') from the identifier.}

\item{type}{Type of the dataset: 'GSE' or 'GDS'. If not specified,
the function will take first three letters
of \code{name} variable as type.}
}
\value{
File with ProtoBuf-serialized ExpressionSet-s
    that were downloaded by this identifier.
    For GSE-datasets there can be multiple annotations, so in file will be a
    list mapping name with GPL to ExpressionSet.
}
\description{
\code{loadGEO} returns the file with serialized ExpressionSet using
    ProtoBuf, parsed from data downloaded from GEO by identifier.
}
\examples{
\dontrun{
    loadGEO("GSE27112")
    loadGEO("GDS4922")
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{updateCountsMeta}
\alias{updateCountsMeta}
\title{Update meta-data for counts collections}
\usage{
updateCountsMeta(
  counts_dir = getPhantasusConf("cache_folders")$rnaseq_counts,
  force = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{counts_dir}{path to counts cache directory}

\item{force}{logical value wich lets function replace existing \code{meta.rda} file}

\item{verbose}{logical value which determines a content of  the output.}
}
\description{
Creates \code{meta.rda} file which contain information about all samples in all collections.
Also function checks \code{priority.txt} file. This file is used to manage collections with the same samples.
}
\details{
First of all function checks validity of \code{priority.txt} file. \bold{Every} Collection should have \bold{unique} priority.
If \code{priority.txt} is not valid function creates new one, setting priorities for each subdirectory(=collection) equal to order in \code{list.dir} output.

Function updates \code{meta.rda} if this file is older than at least one \code{.h5} file in counts files.
\code{meta.rda} is \code{data.table} which is a result of union \code{data.table}s produced by \code{\link{getCountsMetaPart}} for each collection
}
\seealso{
{\code{\link{validateCountsCollection}},\code{\link{updateCountsMeta}}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/es-data.R
\docType{data}
\name{es}
\alias{es}
\title{Example dataset}
\format{
An object of class \code{ExpressionSet} with 20 rows and 5 columns.
}
\usage{
data(es)
}
\description{
Small slice from GSE27112-GPL6103 for
    runnable examples.
}
\examples{
\dontrun{
data(es)
performKmeans(es, k = 2)
}

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{loadFromARCHS4}
\alias{loadFromARCHS4}
\title{Loads expression data from ARCHS4 count files.
Only sapmles with counted expression are kept.
If es already containts expression data it is returned as is.}
\usage{
loadFromARCHS4(es, archs4_files)
}
\arguments{
\item{es}{ExpressionSet from GEO to check for expression in ARCHS4}

\item{archs4_files}{list of available .h5 files from ARCHS4 project}
}
\value{
either original es or an ExpressionSet with loaded count data from ARCHS4
}
\description{
Loads expression data from ARCHS4 count files.
Only sapmles with counted expression are kept.
If es already containts expression data it is returned as is.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapseDataset.R
\name{collapseDataset}
\alias{collapseDataset}
\title{Collapse dataset}
\usage{
collapseDataset(
  es,
  isRows = TRUE,
  selectOne = FALSE,
  fn,
  fields,
  removeEmpty = TRUE
)
}
\arguments{
\item{es}{Expression set}

\item{isRows}{Work with rows. False if columns (default True - row mode)}

\item{selectOne}{select best match or merge duplicates}

\item{fn}{select/merge function}

\item{fields}{fields to unique on}

\item{removeEmpty}{remove unannotated genes}
}
\value{
Nothing. Collapsed dataset will be assigned to es in environment
}
\description{
\code{collapseDataset} performs a collapse action on expression set
}
\examples{
\dontrun{
es <- getGSE('GSE53986')[[1]]
collapseDataset(es, isRows = TRUE, selectOne = TRUE,
fn = mean, fields = c('Gene ID', 'Gene symbol'))
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reproduceInR.R
\name{reproduceInR}
\alias{reproduceInR}
\title{Reproduce session in R code}
\usage{
reproduceInR(sessionName, leaf = T, step = 0, savedEnv = new.env())
}
\arguments{
\item{sessionName}{String, OCPU session name}

\item{leaf}{Boolean, is it leaf (default = F)}

\item{step}{Integer, step of recursion (default = 0)}

\item{savedEnv}{Environment, where to store complex arguments (default = new.env())}
}
\value{
JSON with R code
}
\description{
Reproduce session in R code
}
\examples{
\dontrun{
  setwd(tempdir())
  reproduceInR('x039f1672026678');
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appConfig.R
\name{getPhantasusConf}
\alias{getPhantasusConf}
\title{Read Phantasus Config}
\usage{
getPhantasusConf(
  value = NULL,
  configName = Sys.getenv("R_CONFIG_ACTIVE"),
  file = file.path(tools::R_user_dir(package = "phantasus", which = "config"),
    "user.conf")
)
}
\arguments{
\item{value}{Value to retrieve from the config file.}

\item{configName}{R_CONFIG_ACTIVE value. If unset, "default".}

\item{file}{Location of the config file}
}
\description{
Read Phantasus Config
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convertByAnnotationDB.R
\name{annotationDBMeta}
\alias{annotationDBMeta}
\title{Create meta file for AnnotationDB}
\usage{
annotationDBMeta(annotDir)
}
\arguments{
\item{annotDir}{path to folder with annotationDB sqlite files}
}
\value{
nothing
}
\description{
\code{annotationDBMeta} function creates txt files containing meta information of
provided sqlite files for AnnotationDB.
}
\examples{
\dontrun{
annotationDBMeta('/var/phantasus/cache')
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appConfig.R
\name{setupPhantasus}
\alias{setupPhantasus}
\title{Setup phantasus.
Read user config file ( or create default one) and fill \code{cache_root} using sources in \code{file}.}
\usage{
setupPhantasus(setup_name = "default", file = confFile("setup.yml"))
}
\arguments{
\item{setup_name}{name of config from \code{file}. If unset or not existed, "default".}

\item{file}{Location of the setup.yml file with setup parameters. If not existed use file from package}
}
\description{
Setup phantasus.
Read user config file ( or create default one) and fill \code{cache_root} using sources in \code{file}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{getArchs4Files}
\alias{getArchs4Files}
\title{Returns list of ARCHS4 hdf5 files with expression data}
\usage{
getArchs4Files(cacheDir)
}
\arguments{
\item{cacheDir}{base directory for cache}
}
\value{
list of .h5 files
}
\description{
Returns list of ARCHS4 hdf5 files with expression data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{updateARCHS4}
\alias{updateARCHS4}
\title{Update archs4 files.}
\usage{
updateARCHS4(
  cacheDir = file.path(getPhantasusConf("cache_folders")$rnaseq_counts, "archs4"),
  organism = c("all"),
  force = FALSE
)
}
\arguments{
\item{cacheDir}{file path to \bold{archs4} cache directory}

\item{organism}{vector which determines organisms to download: human, mouse, zoo or all as default.
Also can be a genus. Possible genus: \enumerate{ \item drosophila \item gallus \item bos \item caenorhabditis
\item danio \item rattus \item saccharomyces \item arabidopsis}}

\item{force}{logical value which let function replace current files}
}
\description{
Download archs4 or archs4zoo counts in \code{cacheDir}. If directory does not exists function makes nothing and produce corresponding warnings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{reparseCachedESs}
\alias{reparseCachedESs}
\title{Reparse cached expression sets from GEO.}
\usage{
reparseCachedESs(destdir, mirrorPath = getPhantasusConf("geo_mirrors"))
}
\arguments{
\item{destdir}{Directory used for caching loaded Series files from GEO database.}

\item{mirrorPath}{URL string which specifies the source of matrices.}
}
\value{
vector of previously cached GSE IDs
}
\description{
The function should be used on phantasus version updates that change
behavior of loading datasets from GEO. It finds all the datasets
that were cached and runs `getES` for them again. The function
uses cached Series and other files from GEO.
}
\examples{
reparseCachedESs(destdir=tempdir(), "https://ftp.ncbi.nlm.nih.gov")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculatedAnnotation.R
\name{calculatedAnnotation}
\alias{calculatedAnnotation}
\title{Create calculated annotation}
\usage{
calculatedAnnotation(
  es,
  operation,
  rows = c(),
  columns = c(),
  isColumns = FALSE,
  name = NULL
)
}
\arguments{
\item{es}{ExpressionSet object.}

\item{operation}{Name of the operation to perform calculation}

\item{rows}{List of specified rows' indices (optional), indices start from 0}

\item{columns}{List of specified columns' indices (optional),
indices start from 0#'}

\item{isColumns}{Apply fn to columns}

\item{name}{Name of the new annotation}
}
\value{
Nothing. Annotated dataset will be assigned to es in environment
}
\description{
\code{calculatedAnnotation} adds a column calculated by operation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{getES}
\alias{getES}
\title{Load ExpressionSet by GEO identifier}
\usage{
getES(
  name,
  type = NA,
  destdir = getPhantasusConf("cache_folders")$geo_path,
  mirrorPath = getPhantasusConf("geo_mirrors")
)
}
\arguments{
\item{name}{String, containing GEO identifier of the dataset.
It should start with 'GSE' or 'GDS' and can include exact GPL
to annotate dataset, separated with dash ('-') from the identifier.}

\item{type}{Type of the dataset: 'GSE' or 'GDS'. If not specified,
the function will take first three letters
of \code{name} variable as type.}

\item{destdir}{Directory for caching loaded Series and GPL
files from GEO database.}

\item{mirrorPath}{URL string which specifies the source of matrices.}
}
\value{
List of ExpressionSet objects, that were available by given
    in \code{name} variable GEO identifier.
}
\description{
\code{getES} return the ExpressionSet object(s) corresponding
    to GEO identifier.
}
\examples{
\dontrun{
    getES('GSE14308', type = 'GSE', destdir = 'cache')
    getES('GSE27112')
    getES('GDS4922')
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appConfig.R
\name{createDockerConf}
\alias{createDockerConf}
\title{Creates default Docker conf file
Copies the package user.conf template to \code{user_conf_file} when the
file is absent. When it already exists, any keys present in the template
but missing from the existing file are added (e.g. \code{internal_ports}
on an upgrade from a pre-multiport installation).}
\usage{
createDockerConf(user_conf_file = confFile("user.conf"))
}
\arguments{
\item{user_conf_file}{Location of the user conf file to write.}
}
\description{
Creates default Docker conf file
Copies the package user.conf template to \code{user_conf_file} when the
file is absent. When it already exists, any keys present in the template
but missing from the existing file are added (e.g. \code{internal_ports}
on an upgrade from a pre-multiport installation).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadPreloaded.R
\name{generatePreloadedSession}
\alias{generatePreloadedSession}
\title{Generate files for preloaded session from a session link.}
\usage{
generatePreloadedSession(sessionURL, preloadedName, preloadedDir)
}
\arguments{
\item{sessionURL}{String with session link produced by phantasus.}

\item{preloadedName}{String with name that should be assigned to the session.}

\item{preloadedDir}{Path to the directory with preloaded datasets and sessions.}
}
\value{
Function produces two files (\code{preloadedName.rda} with ExpressionSet
        and \code{preloadedName.json} with session features) in \code{preloadedDir} folder.
}
\description{
Generate files for preloaded session from a session link.
}
\examples{
\dontrun{
sessionURL <- "https://ctlab.itmo.ru/phantasus/?session=x063c1b365b9211" # link from 'Get dataset link...' tool in phantasus
newName <- "my_session" # user defined name
preloadedDir <- "./preloaded" # directory where files will be stored. In order too get access through phantasus web-app should be preloadedDir
dir.create(preloadedDir, showWarnings = FALSE)
generatePreloadedSession(sessionURL= sessionURL,
                         preloadedName = newName,
                         preloadedDir = preloadedDir)

servePhantasus(preloadedDir=preloadedDir, openInBrowser=FALSE)
# open browser manually at http://0.0.0.0:8000/phantasus/index.html?preloaded=my_session
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subsetES.R
\name{subsetES}
\alias{subsetES}
\title{Subsets es, if rows or columns are not specified, all are retained}
\usage{
subsetES(es, columns = c(), rows = c())
}
\arguments{
\item{es}{ExpressionSet object.#'}

\item{columns}{List of specified columns' indices (optional), indices start from 0#'}

\item{rows}{List of specified rows' indices (optional), indices start from 0}
}
\value{
new expression set `es`
}
\description{
Subsets es, if rows or columns are not specified, all are retained
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{parseCharacteristics}
\alias{parseCharacteristics}
\title{Parse characteristics columns somewhat intelligently}
\usage{
parseCharacteristics(pdata)
}
\arguments{
\item{pdata}{AnnotatedDataFrame with characteristics columns}
}
\value{
named data.frame with parsed annotations, rows are in the same order as the input
}
\description{
Parse characteristics columns somewhat intelligently
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjustDataset.R
\name{adjustDataset}
\alias{adjustDataset}
\title{Adjust dataset}
\usage{
adjustDataset(
  es,
  scaleColumnSum = NULL,
  log2 = FALSE,
  onePlusLog2 = FALSE,
  inverseLog2 = FALSE,
  quantileNormalize = FALSE,
  zScore = FALSE,
  robustZScore = FALSE,
  sweep = NULL
)
}
\arguments{
\item{es}{Expression set to perform adjustment on}

\item{scaleColumnSum}{perform sum scaling of columns (default FALSE)}

\item{log2}{perform logarithm2 adjustment (default FALSE)}

\item{onePlusLog2}{perform log2(1+x) adjustment (default FALSE)}

\item{inverseLog2}{perform 2^x adjustment (default FALSE)}

\item{quantileNormalize}{perform quantile normalization (default FALSE)}

\item{zScore}{perform zScore adjustment:
subtract mean, divide by std (default FALSE)}

\item{robustZScore}{perform robustZScore adjustment:
subtract median, divide by MAD (default FALSE)}

\item{sweep}{perform sweep adjustment on rows/columns (default FALSE)}
}
\value{
Nothing. Adjusted dataset will be assigned as ES in global environment
}
\description{
Adjust dataset
}
\examples{
\dontrun{
es <- gseGSE('GSE53986')[[1]]
adjustDataset(es, log2 = T, quantileNormalize = T)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{write.gct}
\alias{write.gct}
\title{Saves ExpressionSet to a GCT file (version 1.3).
Function is deprecated, please use phantasusLite:::writeGct() instead}
\usage{
write.gct(...)
}
\arguments{
\item{...}{parameters for phantasusLite:::writeGct() call}
}
\value{
Result of the closing file (as in `close()` function`)
}
\description{
Saves ExpressionSet to a GCT file (version 1.3).
Function is deprecated, please use phantasusLite:::writeGct() instead
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/performKmeans.R
\name{performKmeans}
\alias{performKmeans}
\title{K-means clusterisation.}
\usage{
performKmeans(es, k, replacena = "mean")
}
\arguments{
\item{es}{ExpressionSet object.}

\item{k}{Expected number of clusters.}

\item{replacena}{Method for replacing NA values
in series matrix (mean by default)}
}
\value{
Vector of corresponding clusters, serialized to JSON.
}
\description{
\code{performKmeans} returns a vector of corresponding clusters for
    each gene from a given ExpressionSet.
}
\examples{
\dontrun{
data(es)
performKmeans(es, k = 2)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convertByAnnotationDB.R
\name{convertByAnnotationDB}
\alias{convertByAnnotationDB}
\title{Map indexes using Annotation DB}
\usage{
convertByAnnotationDB(
  es,
  dbName,
  columnName,
  columnType,
  keyType,
  otherOptions
)
}
\arguments{
\item{es}{source ExpressionSet}

\item{dbName}{name of AnnotationDB file}

\item{columnName}{name of column in featureData of source ExpressionSet}

\item{columnType}{Type of indexes in columnName}

\item{keyType}{Type of mapped indexes}

\item{otherOptions}{additional parameters for conversion. Currently only named boolean deleteDotVersion is not ignored.}
}
\value{
JSON object with a vector of converted IDs
}
\description{
\code{convertByAnnotationDB} function returns \code{keyType} ids from \code{dbName} mapped to \code{columnName} in \code{es}.
}
\examples{
\dontrun{
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{getGSE}
\alias{getGSE}
\title{Load ExpressionSet from GEO Series}
\usage{
getGSE(
  name,
  destdir = getPhantasusConf("cache_folders")$geo_path,
  mirrorPath = getPhantasusConf("geo_mirrors")
)
}
\arguments{
\item{name}{String, containing GEO identifier of the dataset.
It should start with 'GSE' or 'GDS' and can include exact GPL
to annotate dataset, separated with dash ('-') from the identifier.}

\item{destdir}{Directory for caching loaded Series and GPL
files from GEO database.}

\item{mirrorPath}{URL string which specifies the source of matrices.}
}
\value{
List of ExpressionSet objects, that were available by given
    in \code{name} variable GEO identifier.
}
\description{
\code{getGSE} return the ExpressionSet object(s) corresponding
    to GEO Series Identifier.
}
\examples{
\dontrun{
    getGSE('GSE14308', destdir = 'cache')
    getGSE('GSE27112')
    getGSE('GSE53986')
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{checkGPLsFallback}
\alias{checkGPLsFallback}
\title{Check possible annotations for GEO Dataset.}
\usage{
checkGPLsFallback(name)
}
\arguments{
\item{name}{String, containing GEO identifier of the dataset.}
}
\value{
Vector of filenames serialized in JSON format.
    If there is only one GPL for that dataset, the function will
    return \code{name}.
}
\description{
\code{checkGPLs} returns GPL-names for
    the specified GEO identifier.
}
\examples{
\dontrun{
checkGPLs('GSE27112')
checkGPLs('GSE14308')
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{getCountsMetaPart}
\alias{getCountsMetaPart}
\title{Create meta-data for single counts collection}
\usage{
getCountsMetaPart(counts_dir, collection_name, verbose)
}
\arguments{
\item{counts_dir}{path to directory with count collections}

\item{collection_name}{name of collection and collection's directory}

\item{verbose}{logical value which determines a content of  the output.}
}
\value{
\code{data.table} with meta-data or nothing if \code{destdir} does not exist or does not contain files.
}
\description{
Creates a part of counts collections meta-data
}
\details{
Function assumes that \code{collection_name} contains \code{meta.txt} which is valid (in sence of \code{\link{validateCountsCollection}}).
For each row in \code{meta.txt} function reads specified \code{sample_id} dataset and writes every sample id to the resulting \code{data.table} with source file name and collection name.
}
\examples{
\dontrun{
    collDir <- "/path/to/my/collection"
    valid_collection = validateCountsCollection(collectionDir = collDir, verbose = TRUE)
    if (valid_collection){
        metaPart = getCountsMetaPart(destdir = collDir, verbose = TRUE)
     }
 }


}
\seealso{
{\code{\link{validateCountsCollection}},\code{\link{getCountsMetaPart}}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serve.R
\name{servePhantasus}
\alias{servePhantasus}
\title{Serve phantasus.}
\usage{
servePhantasus(
  host = getPhantasusConf("host"),
  port = getPhantasusConf("port"),
  staticRoot = getPhantasusConf("static_root"),
  preloadedDir = getPhantasusConf("preloaded_dir"),
  openInBrowser = TRUE,
  quiet = TRUE,
  background = FALSE
)
}
\arguments{
\item{host}{Host to listen.}

\item{port}{Single integer port to listen on.}

\item{staticRoot}{Path to static files with phantasus.js
(on local file system).}

\item{preloadedDir}{Full path to directory with preloaded files.}

\item{openInBrowser}{Boolean value which states if application will
be automatically loaded in default browser.}

\item{quiet}{Boolean value which states whether the connection log
should be hidden (default: TRUE)}

\item{background}{Boolean value which states whether the server
should be started in background (default: FALSE). Returns an
httpuv server handle when \code{TRUE}.}
}
\value{
In \code{background=TRUE} mode, the httpuv server handle.
    Otherwise does not return (blocks until interrupted).
}
\description{
\code{servePhantasus} starts http server handling phantasus static
files and opencpu server.
}
\examples{
\dontrun{
# Returns handle immediately
s <- servePhantasus(background=TRUE)
s$stop()
}

httpuv::stopAllServers() # can be used if handle is lost
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/limmaAnalysis.R
\name{limmaAnalysis}
\alias{limmaAnalysis}
\title{Differential Expression analysis.}
\usage{
limmaAnalysis(
  es,
  fieldValues,
  version = "One-factor design",
  contrast = list("Comparison", "Target", "Reference"),
  designData = NULL
)
}
\arguments{
\item{es}{ExpressionSet object. It should be normalized for
more accurate analysis.}

\item{fieldValues}{Vector of comparison values, mapping
categories' names to columns/samples}

\item{version}{name of the limma analysis implementation. Should be "One-factor design" or "Advanced design"}

\item{contrast}{a character vector with exactly three elements: the name of a factor in the design formula, the name of the numerator level for the fold change, and the name of the denominator level for the fold change}

\item{designData}{data.frame with design matrix}
}
\value{
Name of the file containing serialized de-matrix.
}
\description{
\code{limmaAnalysis} performs differential expression analysis
    from limma package and returns a ProtoBuf-serialized resulting
    de-matrix.
}
\examples{
\dontrun{
data(es)
limmaAnalysis(es, fieldValues = c("A", "A", "A", "B", "B"))
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phantasus-package.R
\docType{data}
\name{fgseaExample}
\alias{fgseaExample}
\title{Example pathway data.frame for fgsea tool}
\description{
Example pathway data.frame for fgsea tool
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gseaPlot.R
\name{colMeansByGroups}
\alias{colMeansByGroups}
\title{Calculate column averages in row groups}
\usage{
colMeansByGroups(m, groups)
}
\arguments{
\item{m}{matrix n x m}

\item{groups}{vector of size n of numbers from 1 to k}
}
\value{
matrix k*m of column averages  by groups
}
\description{
Calculate column averages in row groups
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{loadCounts}
\alias{loadCounts}
\title{Loads expression data from .h5 count files.
Only samples with counted expression are kept.
If es already containts expression data it is returned as is.}
\usage{
loadCounts(es, counts_dir)
}
\arguments{
\item{es}{ExpressionSet from GEO to check for expression in ARCHS4/dee2 or other h5 files}

\item{counts_dir}{directory with  .h5 files  collections. There must be meta.rda file
in counts_dir and each collection's sub directory must have meta.txt file with description.
Also \code{counts_dir} must contain \code{counts_priority.txt} file.}
}
\value{
either original es or an ExpressionSet with loaded count data from ARCHS4
}
\description{
Loads expression data from .h5 count files.
Only samples with counted expression are kept.
If es already containts expression data it is returned as is.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validateCountsCollection}
\alias{validateCountsCollection}
\title{Check a counts collection}
\usage{
validateCountsCollection(collectionDir, verbose = FALSE)
}
\arguments{
\item{collectionDir}{path to directory with collection}

\item{verbose}{logical value which determines a content of  the output.}
}
\description{
Function checks existing  and  structure of \code{meta.txt} file in specified counts folder.Also it checks accessibility of specified datasets in  corresponding \code{.h5} files.
}
\details{
\code{collectionDir} should contain a bunch of \code{.h5} files
 and a single \code{meta.txt}.  \code{meta.txt} is \code{.tsv}-like file where for each \code{.h5} exists a row wit columns:
\describe{ \item{file_name}{ name of \code{.h5} file in \code{collectionDir}.}
           \item{sample_id}{name of dataset in \code{file_name} which contains sample IDs (sample_geo_accession for example).}
           \item{sample_dim}{which dimension of the expression matrix in \code{file_name} corresponds to samples. Should be one of \code{c("rows", "columns")} }
           \item{gene_id}{name of dataset in \code{file_name} which contains ids for genes and the "meaning" for that ids( column name in result ES). For correct work this dataset should contain unique values. Example: ENSEMBLID:/meta/genes/ensembl_gene_id}
           \item{genes_annot}{Names of datasets and their meanings to extract gene-related metadata from \code{file_name}. Can be empty or \code{gene_id}-like values separated with semicolon(;).}
           }
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gseaPlot.R
\name{gseaPlot}
\alias{gseaPlot}
\title{Returns path to an svg file with enrichment plot}
\usage{
gseaPlot(
  es,
  rankBy,
  selectedGenes,
  width,
  height,
  vertical = FALSE,
  addHeatmap = FALSE,
  showAnnotation = NULL,
  annotationColors = NULL,
  pallete = c("blue", "white", "red")
)
}
\arguments{
\item{es}{ExpressionSet object.}

\item{rankBy}{name of the numeric column used for gene ranking}

\item{selectedGenes}{indexes of selected genes (starting from one, in the order of fData)}

\item{width}{width of the image (in inches)}

\item{height}{height of the image (in inches)}

\item{vertical}{whether to use vertical orientation (default: FALSE)}

\item{addHeatmap}{whether to add an expression heatmap, sorted by rankBy (default: FALSE)}

\item{showAnnotation}{a name of column annotation to add to the heatmap, default: NULL (no annotation)}

\item{annotationColors}{a list of colors to use in annotation}

\item{pallete}{a vector of colors to draw heatmap}
}
\value{
path to an svg file
}
\description{
Returns path to an svg file with enrichment plot
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{checkGSEType}
\alias{checkGSEType}
\title{Checks GSE to be supported}
\usage{
checkGSEType(name, destDir, combine = any)
}
\arguments{
\item{name}{GSE id, with optional GPL specification}

\item{destDir}{path to cache directory}

\item{combine}{function on how to combine results, when multiple platforms are present}
}
\value{
logical vector if the dataset is supported or not
}
\description{
Checks GSE to be supported
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadGEO.R
\name{getGDS}
\alias{getGDS}
\title{Load ExpressionSet from GEO Datasets}
\usage{
getGDS(
  name,
  destdir = getPhantasusConf("cache_folders")$geo_path,
  mirrorPath = getPhantasusConf("geo_mirrors")
)
}
\arguments{
\item{name}{String, containing GEO identifier of the dataset.
It should start with 'GSE' or 'GDS' and can include exact GPL
to annotate dataset, separated with dash ('-') from the identifier.}

\item{destdir}{Directory for caching loaded Series and GPL
files from GEO database.}

\item{mirrorPath}{URL string which specifies the source of matrices.}
}
\value{
ExpressionSet object wrapped in list, that was available by given
    in \code{name} variable GEO identifier.
}
\description{
\code{getGDS} return the ExpressionSet object corresponding
    to GEO Dataset identifier.
}
\examples{
\dontrun{
   getGDS('GDS4922', destdir = tempdir(), mirrorPath = "https://ftp.ncbi.nlm.nih.gov")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{updateARCHS4meta}
\alias{updateARCHS4meta}
\title{Update ARCHS4 meta files}
\usage{
updateARCHS4meta(
  archDir = file.path(getPhantasusConf("cache_folders")$rnaseq_counts, "archs4")
)
}
\arguments{
\item{archDir}{path to directory with arch4 .h5 files.}
}
\description{
Creates \code{meta.txt} file, which describes typical archs4 and archs4Zoo files.
}
\details{
This function produces very specific "hardcoded" \code{meta.txt} file for arch4 and archs4ZOO counts collections.
See \code{\link{validateCountsCollection}} for more common information and \code{meta.txt}  file structure
}
\seealso{
\code{\link{validateCountsCollection}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convertByAnnotationDB.R
\name{queryAnnotationDBMeta}
\alias{queryAnnotationDBMeta}
\title{Get meta list for annotationDB files}
\usage{
queryAnnotationDBMeta()
}
\value{
meta info in JSON
}
\description{
\code{queryAnnotationDBMeta} Function reads txt meta files for provided
sqlite annotation databases.
}
\examples{
\dontrun{
queryAnnotationDBMeta()
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{isHSDS}
\alias{isHSDS}
\title{check if url  responding as HSDS server
TRUE - hsds
FALSE - web link but not working
NULL - not web link}
\usage{
isHSDS(url)
}
\arguments{
\item{url}{URL to check}
}
\description{
check if url  responding as HSDS server
TRUE - hsds
FALSE - web link but not working
NULL - not web link
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createES.R
\name{createES}
\alias{createES}
\title{Create ExpressionSet.}
\usage{
createES(data, pData, varLabels, fData, fvarLabels, eData)
}
\arguments{
\item{data}{Gene expression matrix.}

\item{pData}{Matrix with phenotypical data.}

\item{varLabels}{Names of phenoData columns.}

\item{fData}{Matrix with feature data.}

\item{fvarLabels}{Names of featureData columns.}

\item{eData}{List with experimentData}
}
\value{
produced ExpressionSet object
}
\description{
\code{createES} function produces an ExpressionSet object from given data,
    and exports it to global scope.
}
\examples{
\dontrun{
data <- matrix(1:15, 5, 3)
pData <- c("A", "B", "C")
varLabels <- "cat"
fData <- c("p", "r", "s", "t", "u")
fvarLabels <- "id"
eData <- list(name="", lab="", contact="", title="", url="", other=list(), pubMedIds="")
createES(data, pData, varLabels, fData, fvarLabels, eData)
}
}
\keyword{internal}
